#!/usr/bin/env python
# -*- coding:utf-8 -*-
#############################################################################
#Copyright (c): 2020, Huawei Tech. Co., Ltd.
#FileName     : GSroachBackup.py
#Version      : V1.0.0
#Date         : 2020-01-20
#Description  : GSroachBackup.py is the main process module for performing backup.
#############################################################################

try:
    import time
    import sys
    import os
    import commands
    import subprocess
    import signal
    import json
    import re
    import copy
    import socket
    from datetime import datetime

    import util.GSroachRestore

    sys.path.append(sys.path[0] + "/../")
    from gspylib.common.Common import DefaultValue, ClusterCommand
    from gspylib.common.GaussLog import GaussLog
    from gspylib.common.DbClusterInfo import dbClusterInfo
    from gspylib.common.ErrorCode import ErrorCode
    from gspylib.threads.SshTool import SshTool
    from multiprocessing.dummy import Pool as ThreadPool
    from gspylib.os.gsfile import g_file
    from util.GSroachValidate import startValidate
    from util.GSroachCommonParameter import validateRoachParams
    from util.GSroachCommon import sigint_handler, checkPrivateTable, mark_completion_progress, \
        final_ip_param_dict, getRoachOpResult, setRoachOpResult, final_ip_param_dict, \
        makeRaochOptions, master_host, isClusterBackup, remove_pattern, \
        isMiniaturizedDeployment, checkFile, getCoHostPort, check_gs_roach_running, \
        create_roachPath, checkStartingPitrArchive, checkPitrStopInProgress, \
        getParentBkptype, getMetaDest, getParentBackupKey, \
        checkClusterNodeCount, gauss_home_loc, proc_id_dict, \
        parseMaster_generated_file, sendSignal, pollSubprocess, create_roachPath, \
        create_command, createLogPath, reform_tablename, pssh_tout, \
        getBackupDetails, getMaster_ip, VERSION_FILE, STATIC_CONFIG_FILE, PITR_MARK_FILE, \
        parseRoachConf, validate_logparams, updateCN_hosts, getSingleDnHostPort, postExecCleanup, \
        rmMediaBackupDirs, gsinitLogger, getClusterVersion, signalAndExit, \
        compareVersionFile, hasStopProcess, loadMetaFileAsJson, LOCAL_ROACH, \
        getClusterBalanceState, checkBalanceStateSame, cleanMetaMediaPath, unsetRoachCron, \
        updateBackupIniFailures, checkStartingPitrArchive, \
        ETCD_TERM_FILE, ETCD_TERM_FILE_TMP, TIMEOUT_SCP, agentErrorReport, \
        deliveryOneItemToAgents
except ImportError, e:
    sys.exit("[GAUSS-52200] : Unable to import module: %s." % str(e))


#global variables
ACTION_CONFIG = "config"
userProfile = "~/.bashrc"
osProfile = "/etc/profile"
currPath = os.path.dirname(os.path.realpath(__file__))
localRoach = os.path.normpath(currPath + "/../local/LocalRoach.py")
cluster_node_details = None
master_ip = ""
table_count = "100001"
cluster_version = 0
g_clusterInfo = None
g_logger = None
backup_key = None
g_backup_is_stopped = False #backup operation is aborted by stop command
ACTION_RESET_FULLPAGEWRITES = "reset_full_page_writes"
ACTION_CHECK_UPGRADE_VERSION = "check_upgrade_version"
TIMEOUT_PSSH_PRECHECK = 300          #5min to check python process is runing


# This value simply identifies that the backup-key has been successfully restored, and has no other meaning
TLI_HAS_RESTORED = 255

signal.signal(signal.SIGINT, sigint_handler)


def getMulttalbeBackupkey():
    """
    Get backup key for logial mtable backup
    """
    cmd = "tail -n1 $GAUSSHOME/bin/roach_key.txt | awk -F ' ' '{print $2}' | cut -d '_' -f 1,2"
    g_logger.debug("Get logical backupkey cmd: %s" % (cmd))
    (status, output) = commands.getstatusoutput(cmd)
    if (status != 0):
        g_logger.debug("File does not exist. Error: \n%s." % output)
    if (len(output) > 1):
        return output
    else:
        g_logger.log("Warning: Get logical backupkey failed.")

def upload_metadata_in_all_node(g_opts):
    if g_opts.mediaType.upper() not in ["NBU", "OBS"] :
        return
    pyPath = os.path.dirname(currPath)
    g_logger.log("Uploading metadata in all nodes.")
    l_clusterInfo = dbClusterInfo()
    l_clusterInfo.initFromStaticConfig(g_opts.user)
    l_sshTool = SshTool(l_clusterInfo.getClusterNodeNames(), g_logger.logFile, timeout = 60 * 10)

    if g_opts.mediaType.upper() == "NBU":
        cmd = "source %s && export PYTHONPATH=%s:$PYTHONPATH && python $GPHOME/script/util/GSroachLocal.py -t" \
            "uploadmetadata --backupKey %s --metaData '%s' --mediaType 'NBU'  --remoteMetaData '%s' --sourceFile '%s' -U %s " \
            % (g_opts.sourceFile, pyPath, backup_key, g_opts.metadataDestination, g_opts.mediaDest, g_opts.sourceFile, g_opts.user)
    elif g_opts.mediaType.upper() == "OBS":
        cmd = 'source %s && export PYTHONPATH=%s:$PYTHONPATH && python $GPHOME/script/util/GSroachLocal.py -t uploadmetadata' \
            ' --metaData \'%s\' --mediaType \'OBS\' --sourceFile \'%s\' -U %s' \
            ' --backupKey \'%s\' --obs-server-ip \'%s\' --bucket-name \'%s\' --clusterUniqueID \'%s\'' \
        % (g_opts.sourceFile, pyPath, g_opts.metadataDestination, g_opts.sourceFile, g_opts.user,
        backup_key, g_opts.OBSAddres, g_opts.bucketName, g_opts.clusterUniqueID)
    g_logger.debug("Uploading metadata cmd: %s" % cmd)
    (status, output) = l_sshTool.getSshStatusOutput(cmd, g_opts.nodeList)
    g_logger.debug("The status.values() is %s" % str(status.values()))
    g_logger.debug("The output is %s" % output)
    if 'Failure' in status.values():
        g_logger.debug("SshTools Execute cmd failed: %s" % cmd)
        g_logger.log("ERROR: Upload metadata to net backup failed, we will rollback current backup.")
        return
    g_logger.log("Uploading metadata in all nodes finished.")

def checkUpgradeVersion(g_opts):
    # check upgrade version in the end of backup, avoid an upgrade occurred during the backup
    if DefaultValue.isUnderGreyUpgrade(g_opts.user) and not DefaultValue.isGreyPreCommit(g_opts.user):
        cmd = "source %s && python %s -t %s -U %s -l %s" \
            % (g_opts.sourceFile, LOCAL_ROACH, ACTION_CHECK_UPGRADE_VERSION, g_opts.user, g_opts.logFile)
        g_logger.debug("GaussRoach.py calling LocalRoach.py to check upgrade version, cmd=[%s]" % (cmd))
        try:
            sshTool = SshTool(g_clusterInfo.getClusterNodeNames(), g_logger.logFile, TIMEOUT_PSSH_PRECHECK)
            sshTool.executeCommand(cmd, "check upgrade version")
            g_logger.debug("Successfully check upgrade version.")
        except Exception, e:
            g_logger.log("Error: An upgrade occurred during the backup!!")
            return False
    return True


def doBackup(g_opts, command_dict, cnHost, table_list = []):
    """
    Do backup operation
    """
    global g_backup_is_stopped
    bkp_key = None
    if not g_opts.single:
        cnPort = getCoHostPort(cnHost)
        cnPort = cnPort[1]

    g_logger.debug("start backup operation...")
    if g_opts.logical:
        if len(table_list) < int(table_count):
            if g_opts.dbName != "":
                db = g_opts.dbName
            else:
                db = 'postgres'
            checkDatabase(g_opts, cnHost, cnPort, db)
            for table in table_list:
                checkTablesForLogical(g_opts, cnHost, table)
            backupOnNodes(g_opts, command_dict, cnHost, cnPort)
            if g_backup_is_stopped or getRoachOpResult():
                g_logger.debug("Backup failed...")
                return
            else:
                final_ip_param_dict.clear()
                makeRaochOptions(g_opts, master_ip, master_host, command_dict)
        else:
            os.system("clear")
            g_logger.log("Error : Table List should not contain Tables Greater than %s" \
                % table_count)
            raise SystemExit
        os.system('clear')
        bkp_key = getMulttalbeBackupkey()
        g_logger.log("Logical Multiple table backup COMPLETED")
        g_logger.log("Backup operation SUCCESSFUL. Backup key: %s\n" % bkp_key)
    else:
        #for normal cluster
        if not g_opts.single: 
            backupOnNodes(g_opts, command_dict, cnHost, cnPort)
        #for single mode
        else:
            singleDnPort = getSingleDnHostPort(cnHost)
            backupOnNodes(g_opts, command_dict, cnHost, singleDnPort)

        if g_backup_is_stopped or getRoachOpResult():
            g_logger.debug("Backup failed...")
            if ((not g_opts.resumeBkp) and (g_opts.remoteMasterIp != "")):
                cleanRemoteBackup(g_opts, backup_key)
            return

        #backup c-udf, javaudf and postgis libs
        if (isClusterBackup(g_opts, command_dict)):
            backupAuxFiles(g_opts, backup_key)

    g_logger.debug("end backup operation...")

def cleanRemoteBackup(g_opts, backupKey, connectTimeout = 60):
    """
    Clean backupset of remote cluster
    """
    remoteBackupMetaPath = '%s/roach/%s' % (g_opts.remoteMetaData, backupKey)
    remoteBackupMediaPath = '%s/roach/%s' % (g_opts.remoteMediaData, backupKey)
    remoteHostGaussHome = g_opts.remoteHostGaussHome
    remoteHostIp = g_opts.remoteMasterIp
    srcFile = "%s/bin/cluster_static_config" % remoteHostGaussHome
    gaussHome = DefaultValue.getEnv("GAUSSHOME")
    currentTime = int(round(time.time() * 1000))
    destFile = "%s/bin/cluster_static_config_%s_%d" % (gaussHome, remoteHostIp, currentTime)

    # get remote static config file
    cmdPscp = "pscp -t %d -O StrictHostKeyChecking=no -H %s %s %s"\
              % (connectTimeout, master_host, srcFile, destFile)
    cmd = "pssh -t %d -O StrictHostKeyChecking=no -s -H %s \"%s\""\
          % (connectTimeout, remoteHostIp, cmdPscp)
    g_logger.debug("Command is %s" % cmd)
    status, output = commands.getstatusoutput(cmd)
    if status != 0:
        g_logger.debug("ERROR: get remote static config file failed after backup failed with status(%d), output(%s)" \
            % (status, output))
        g_logger.error("get Remote static config file failed after backup failed.")
        return
    remoteClusterInfo = dbClusterInfo()
    remoteClusterInfo.initFromStaticConfig(g_opts.remoteHostUserName, destFile)
    remoteNodeNames = remoteClusterInfo.getClusterNodeNames()

    # remove remote static config file
    cmd = "rm -rf %s" % destFile
    g_logger.debug("Command is %s" % cmd)
    status, output = commands.getstatusoutput(cmd)
    if status != 0:
        g_logger.debug("Warning: remove remote static config file failed after backup failed with status(%d), output(%s)" \
            % (status, output))

    # remove backup files in remote cluster
    pNodeList = " -H ".join(remoteNodeNames)
    cmd = "pssh -s -H %s -t %d -O StrictHostKeyChecking=no \"rm -rf %s %s\"" \
        % (pNodeList, connectTimeout, remoteBackupMetaPath, remoteBackupMediaPath)
    g_logger.debug("Command is %s" % cmd)
    status, output = commands.getstatusoutput(cmd)
    if status != 0:
        g_logger.debug("ERROR: remove backup files in remote cluster failed after backup failed with status(%d), output(%s)" \
            % (status, output))
        g_logger.error("remove backup files in remote cluster failed after backup failed.")
        return

def logOptions():
    """
    writing to logfile
    """
    for key in final_ip_param_dict.keys():
        g_logger.debug("options for %s" %key)
        for key, val in final_ip_param_dict[key].items():
            if key == "-W":
                g_logger.debug("%s : %s" %(key.translate(remove_pattern, "--"), "***"))
            else:
                g_logger.debug("%s : %s" %(key.translate(remove_pattern, "--"), val))

def createMasterOption():
    """
    create option for master
    """
    global final_ip_param_dict
    new_value = None
    final_value = ''
    for k in final_ip_param_dict.keys():
        for key, val in final_ip_param_dict[k].iteritems():
            if val != None:
                if key == '--tablename':
                    val = reform_tablename(val)
                    #If tablename and/or schemaname starting with
                    #"digit", need to protect them with quotes...
                    if '.' in val:
                        lval = val.split('.')
                    else:
                        lval = [val]
                    tname = ''
                    for elem in lval:
                        if elem.strip()[0].isdigit() or '$' in elem:
                            #Add quoting...
                            tname += '\'"%s"\'.' % elem.strip()
                        else:
                            tname += '%s.' % elem.strip()
                    tname = tname[:-1]
                    new_value = ''.join("%s %s " %(key, tname))
                else:
                    new_value = ''.join("%s %r " %(key, val))
            else:
                new_value = ''.join("%s " %(key))
            final_value += new_value
    
    if isMiniaturizedDeployment(cluster_version):
        final_value += ''.join("--miniaturized-deployment ")

    return final_value


def checkClusterNormal(g_opts, g_logger):
    """
    function: Check cluster status, if NORMAL, return 0, else return 1
        for backup, we must make sure cluster_state is Normal.
        Do not care about the balanced status.
    input : NA
    output: 0  successfully
            1  failed
    """
    # if pitr-xlog backup in progress, never check Normal for failover
    if checkStartingPitrArchive(g_opts):
        return 0
    
    # according to the implementation of the results to determine whether the implementation of success
    cmd = "source %s; cm_ctl query" % g_opts.sourceFile
    (status, output) = commands.getstatusoutput(cmd)
    if (status != 0 or output.find("Normal") < 0):
        errorString = ErrorCode.GAUSS_516["GAUSS_51602"] + "Output:\n%s" % output
        agentErrorReport("roach_backup", errorString)
        g_logger.logExit(errorString)

    cluster_state_check = False
    for line in output.split('\n'):
        if(len(line.split(":")) != 2):
            continue
        (key, value) = line.split(":")
        if(key.strip() == "cluster_state" and value.strip() == "Normal"):
            cluster_state_check = True

            
    if cluster_state_check == True:
        return 0
    else:
        return 1

def checkNodeGroup(g_opts, tablename, db, coordPort, cnHost):
    """
    If logical backup for a table on a subset of DN's - return false
    """
    sqlstring = "select count(*) from pgxc_node where node_type='D'"
    (status, output) = ClusterCommand.remoteSQLCommand(sqlstring, 
        g_opts.user, cnHost, coordPort, False, db)
    g_logger.debug("checkNodeGroup: cmd to get dnCount is: [%s], result is %d, output is (%s)." \
        % (sqlstring, status, output))
    if status != 0:
        g_logger.logExit("Error: Not able to check for node group before logical table backup...")
    dnCount = int(output)

    sqlstring = "select nodeoids from pgxc_class where pcrelid = '%s'::regclass::oid" % (tablename)
    (status, output) = ClusterCommand.remoteSQLCommand(sqlstring, 
        g_opts.user, cnHost, coordPort, False, db)
    g_logger.debug("checkNodeGroup: cmd to get nodeCountTbl is: [%s], result is %d, output is (%s)." \
        % (sqlstring, status, output))
    if status != 0:
        g_logger.logExit("Error: Not able to check for node group before logical table backup...")
    nodeCountTbl = len(output.strip().split())

    g_logger.debug("checkNodeGroup: dnCount=%d, nodeCountTbl=%d" % (dnCount, nodeCountTbl))
    if dnCount != nodeCountTbl:
        g_logger.logExit("Error: Logical backup not supported for a table created on a node group having limited DN's.")

def createBackupLibDir(g_opts, dirpath):
    """
    create backup dir for libs
    """
    create_cmd = 'source %s && if [ ! -e %s ]; then mkdir -p %s -m 700; fi' \
        % (g_opts.sourceFile, dirpath, dirpath)
    status, output = commands.getstatusoutput(create_cmd)
    g_logger.debug("create dir commond when backup libs: %s" % (create_cmd))
    if status != 0:
        g_logger.logExit("Error: Unable to create path \"%s\" when backup libs! %s" \
            % (dirpath, output))

def backupLibs(g_opts, filetype, srcpath, destpath):
    """
    backup libs
    """
    if (not os.path.exists(srcpath)):
        g_logger.logExit("ERROR: source path \"%s\" does not exist when backup libs."
            % (srcpath))
    if (not os.path.exists(destpath)):
        g_logger.logExit("ERROR: destination path \"%s\" does not exist when backup libs."
            % (destpath))
    bkcmd = "source %s && cp -frp %s/%s %s" % (g_opts.sourceFile, srcpath, filetype, destpath)
    g_logger.debug("config file or verison file backup command: %s" % (bkcmd))
    status, output = commands.getstatusoutput(bkcmd)
    if status != 0:
        g_logger.logExit("Error: backup config file or verison file file \"%s\" failed when backup libs! %s" \
            % (filetype, output))

def backupLibsToRemote(g_opts, filetype, backuptype, srcpath, key, timeout = 60):
    """
    backup libs to remote cluster
    """
    if (g_opts.remoteMetaData == "" or g_opts.remoteMasterIp == ""):
        return
    if (not os.path.exists(srcpath)):
        g_logger.log("Error: source path \"%s\" does not exist" % (srcpath))
        return

    if ((VERSION_FILE == filetype) or (STATIC_CONFIG_FILE == filetype) or (ETCD_TERM_FILE == filetype)):
        destPath = "%s/roach/%s" % (g_opts.remoteMetaData, key)
    else:    
        destPath = "%s/roach/%s/binaries/%s" % (g_opts.remoteMetaData, key, backuptype)

    cmd = "pssh -s -H %s -t %d -O StrictHostKeyChecking=no \"if [ ! -d %s ]; then mkdir -p %s -m 700; fi \"" \
        % (g_opts.remoteMasterIp, timeout, destPath, destPath)
    g_logger.debug("Creating metadata path, cmd: %s" % cmd)
    (status, output) = commands.getstatusoutput(cmd)
    if (status != 0):
        g_logger.logExit("Create %s in %s failed with output: %s" % (destPath, g_opts.remoteMasterIp, output))
    cmd = "pscp -t %d -O StrictHostKeyChecking=no -H %s -r %s/%s %s" \
          % (timeout, g_opts.remoteMasterIp, srcpath, filetype, destPath)
    g_logger.debug("remote backup command:%s" % (cmd))
    status, output = commands.getstatusoutput(cmd)
    if status != 0:
        g_logger.logExit("Error: backup file \"%s\" failed! %s" % (filetype, output))

def deliveryClusterConfigToAgents(g_opts, sshTool, timeout = 60):
    """
    copy cluster config file from master to the same directory of other agent nodes
    """
    pathName = DefaultValue.getEnv("GAUSSHOME")
    itemName  = "cluster_static_config_remote"
    cmdPscp = "pscp -t %d -O StrictHostKeyChecking=no -H %s %s/bin/cluster_static_config %s/bin/%s"\
              % (timeout, master_host, g_opts.remoteHostGaussHome, pathName, itemName)
    cmd = "if [ ! -f %s/bin/%s ]; then pssh -t %d -O StrictHostKeyChecking=no -s -H %s \"%s\"; fi"\
          % (pathName, itemName, timeout, g_opts.remoteMasterIp, cmdPscp)
    status, output = commands.getstatusoutput(cmd)
    if status != 0:
        raise Exception("Get remote static config file failed with output: %s" % output)

    pNodeList = " -H ".join(g_opts.nodeList)
    cmdPscp = "pscp -H %s %s/bin/%s %s/bin/" % (pNodeList, pathName, itemName, pathName)
    status, output = commands.getstatusoutput(cmdPscp)
    if status != 0:
        raise Exception("Delivery %s/%s failed with output: %s" % (pathName, itemName, output))
    else :
        g_logger.debug("Copy remote static config file successfully.")

def getClusterNodeIps(ClusterInfo):
    """
    Get node IP in cluster
    """
    return [dbNode.sshIps[0] for dbNode in ClusterInfo.dbNodes]

def getCurrentNodeCount(g_opts, clusterInfo=None):
    if clusterInfo == None:
        clusterInfo = dbClusterInfo()
        clusterInfo.initFromStaticConfig(g_opts.user)
    nodeList = clusterInfo.getClusterNodeNames()
    nodeCount = len(nodeList)
    dataNodeCount = 0
    for count in range(nodeCount):
        dataNodeCount += clusterInfo.dbNodes[count].dataNum
    g_logger.debug("Current node count is %d, datanode count is %d" % (nodeCount, dataNodeCount))
    return nodeCount, dataNodeCount

def backupAuxFiles(g_opts, backup_key):
    """
    backup auxiliary non-data files
    """
    if g_opts.pitr and g_opts.resumeBkp:
        g_logger.debug("When pitr in resume-backup phase, no need to backup!")
        return
    backupVersionFile(g_opts, backup_key)
    backupClusterConfigFile(g_opts, backup_key)
    dealCFuncFiles(g_opts, backup_key)
    dealTsDict(g_opts, backup_key)
    if (g_opts.priorBkpKey == ""):
        backupLibFiles(g_opts, backup_key)
    backupPitrMarkFile(g_opts)
    backupETCDTerm(g_opts, backup_key)


def dealCFuncFiles(g_opts, backupkey):
    pyPath = os.path.dirname(currPath)
    g_logger.debug("Backing up C function so file.")
    l_clusterInfo = dbClusterInfo()
    l_clusterInfo.initFromStaticConfig(g_opts.user)
    l_sshTool = SshTool(l_clusterInfo.getClusterNodeNames(), g_logger.logFile, timeout=30)

    if g_opts.remoteMasterIp != "":
        deliveryClusterConfigToAgents(g_opts, l_sshTool)

        cmd = "source %s && export PYTHONPATH=%s:$PYTHONPATH && python $GPHOME/script/util/GSroachLocal.py -t \
        backupcfuncfiles --backupKey %s --remoteMetaData '%s' --remoteMasterIp %s   --sourceFile '%s' -U %s -l %s  " \
        % (g_opts.sourceFile, pyPath, backupkey, g_opts.remoteMetaData, g_opts.remoteMasterIp, g_opts.sourceFile, \
        g_opts.user, g_opts.localLog)
    else:
        if (g_opts.mediaType.upper() == "OBS"):
            cmd = "source %s && export PYTHONPATH=%s:$PYTHONPATH && python $GPHOME/script/util/GSroachLocal.py -t \
            backupcfuncfiles --backupKey %s --metaData '%s' --clusterUniqueID %s --mediaType %s \
            --sourceFile '%s' -U %s -l %s  " \
            % (g_opts.sourceFile, pyPath, backupkey, g_opts.metadataDestination, str(g_opts.clusterUniqueID), \
            g_opts.mediaType, g_opts.sourceFile, g_opts.user, g_opts.localLog)
        else:
            cmd = "source %s && export PYTHONPATH=%s:$PYTHONPATH && python $GPHOME/script/util/GSroachLocal.py -t \
            backupcfuncfiles --backupKey %s --metaData '%s' --sourceFile '%s' \
            -U %s -l %s  " \
            % (g_opts.sourceFile, pyPath, backupkey, g_opts.metadataDestination, g_opts.sourceFile, \
            g_opts.user, g_opts.localLog)

    (status, output) = l_sshTool.getSshStatusOutput(cmd, g_opts.nodeList)
    g_logger.debug("The status.values() is %s" % str(status.values()))
    g_logger.debug("The output is %s" % output)
    if 'Failure' in status.values():
        g_logger.logExit("Copy CFunc backup files failed!")
    else:
        g_logger.debug("Copy CFunc backup files successfully.")

def dealTsDict(g_opts, backupkey):
    pyPath = os.path.dirname(currPath)
    g_logger.debug("Backing up ts dict so file.")

    l_clusterInfo = dbClusterInfo()
    l_clusterInfo.initFromStaticConfig(g_opts.user)
    l_sshTool = SshTool(l_clusterInfo.getClusterNodeNames(), g_logger.logFile, timeout=30)

    if g_opts.remoteMasterIp != "":
        deliveryClusterConfigToAgents(g_opts, l_sshTool)

        cmd = "source %s && export PYTHONPATH=%s:$PYTHONPATH && python $GPHOME/script/util/GSroachLocal.py -t \
        backuptsdictfiles --backupKey %s --remoteMetaData '%s' --remoteMasterIp %s   --sourceFile '%s' -U %s -l %s  " \
        % (g_opts.sourceFile, pyPath, backupkey, g_opts.remoteMetaData, g_opts.remoteMasterIp, g_opts.sourceFile, \
        g_opts.user, g_opts.localLog)
    else:
        if (g_opts.mediaType.upper() == "OBS"):
            cmd = "source %s && export PYTHONPATH=%s:$PYTHONPATH && python $GPHOME/script/util/GSroachLocal.py -t \
            backuptsdictfiles --backupKey %s --metaData '%s' --clusterUniqueID %s --mediaType %s \
            --sourceFile '%s' -U %s -l %s  " \
            % (g_opts.sourceFile, pyPath, backupkey, g_opts.metadataDestination, str(g_opts.clusterUniqueID), \
            g_opts.mediaType, g_opts.sourceFile, g_opts.user, g_opts.localLog)
        else:
            cmd = "source %s && export PYTHONPATH=%s:$PYTHONPATH && python $GPHOME/script/util/GSroachLocal.py -t \
            backuptsdictfiles --backupKey %s --metaData '%s' --sourceFile '%s' \
            -U %s -l %s  " \
            % (g_opts.sourceFile, pyPath, backupkey, g_opts.metadataDestination, g_opts.sourceFile, \
            g_opts.user, g_opts.localLog)
    (status, output) = l_sshTool.getSshStatusOutput(cmd, g_opts.nodeList)
    g_logger.debug("The status.values() is %s" % str(status.values()))
    g_logger.debug("The output is %s" % output)
    if 'Failure' in status.values():
        g_logger.logExit("Copy ts dict backup files failed!")
    else:
        g_logger.debug("Copy ts dict backup files successfully.")


def backupPitrMarkFile(g_opts):
    """
    backup pitr_mark_file
    """ 
    if not checkStartingPitrArchive(g_opts):
        g_logger.debug("pitr_mark_file not exist, cannot distribute")
        return
    
    g_logger.debug("start distribute pitr_mark_file")
    filePath = "%s/bin" % DefaultValue.getInstallDir(g_opts.user)
    hostname = commands.getstatusoutput('hostname')[1]
    nodeList = " -H ".join(g_clusterInfo.getClusterNodeNames())
    cmdPscp = 'pscp -t 600000 -H %s %s/%s %s' % (nodeList, filePath, PITR_MARK_FILE, filePath)
    g_logger.debug(cmdPscp)

    status, output = commands.getstatusoutput(cmdPscp)
    if status != 0:
        g_logger.debug("Error when distribute pitr_mark_files: %s" % output)
        g_logger.exit("Error: failed to distribute pitr_mark_file!")
        
    else:
        g_logger.log("Successfully distribute pitr_mark_files.")

def backupLibFiles(g_opts, backupkey):
    """
    backup libs for c-udf, postgis and javaudf
    """

    g_logger.debug("start backup libs for cluster backup.")
    backupdir = os.path.join(g_opts.metadataDestination, \
        "roach", backupkey, "binaries")
    createBackupLibDir(g_opts, backupdir)

    #backuplib
    srcpath_lib = "%s/lib" % (DefaultValue.getInstallDir(g_opts.user))
    destpath_lib = "%s/lib" % (backupdir)
    libTypeList = ["libgeos_c.so.*", "libproj.so.*", "libjson-c.so.*", \
        "libgeos-*.*.*so", "libgcc_s.so.*", "libstdc++.so.*"]
    for libType in libTypeList:
        if (checkFile(srcpath_lib, libType)):
            createBackupLibDir(g_opts, destpath_lib)
            backupLibs(g_opts, libType, srcpath_lib, destpath_lib)
            backupLibsToRemote(g_opts, libType, 'lib', srcpath_lib, backupkey)

    #bacukupPostgis
    srcpath_postgis = "%s/postgresql" % (srcpath_lib)
    destpath_postgis = "%s/postgresql" % (backupdir)
    postgisType = "postgis-*.*.so"
    if (checkFile(srcpath_postgis, postgisType)):
        createBackupLibDir(g_opts, destpath_postgis)
        backupLibs(g_opts, postgisType, srcpath_postgis, destpath_postgis)
        backupLibsToRemote(g_opts, postgisType, 'postgresql', srcpath_postgis, backupkey)

    #backupJavaudf
    srcpath_javaudf = "%s/java" % (srcpath_postgis)
    destpath_javaudf = "%s/java" % (backupdir)
    javaudfType = "*"
    if (checkFile(srcpath_javaudf, javaudfType)):
        createBackupLibDir(g_opts, destpath_javaudf)
        backupLibs(g_opts, javaudfType, srcpath_javaudf, destpath_javaudf)
        backupLibsToRemote(g_opts, javaudfType, 'java', srcpath_javaudf, backupkey)

    #backupshare
    srcpath_share = "%s/share/postgresql/extension/" % (DefaultValue.getInstallDir(g_opts.user))
    destpath_share = "%s/share/postgresql/extension" % (backupdir)
    shareTypeList = ["postgis--*.*.*.sql","postgis.control"]
    for shareType in shareTypeList:
        if (checkFile(srcpath_share, shareType)):
            createBackupLibDir(g_opts, destpath_share)
            backupLibs(g_opts, shareType, srcpath_share, destpath_share)
            backupLibsToRemote(g_opts, shareType, 'share', srcpath_share, backupkey)

    #backup bin and tde files
    srcpath_bin = "%s/bin" % (DefaultValue.getInstallDir(g_opts.user))
    destpath_bin = "%s/bin" % (backupdir)
    binTypeList = ["pgsql2shp","shp2pgsql", "gs_tde_keys.cipher"]
    for binType in binTypeList:
        if (checkFile(srcpath_bin, binType)):
            createBackupLibDir(g_opts, destpath_bin)
            backupLibs(g_opts, binType, srcpath_bin, destpath_bin)
            backupLibsToRemote(g_opts, binType, 'bin', srcpath_bin, backupkey)

    #backupsecbox
    srcpath_secbox = "%s/secbox" % (DefaultValue.getInstallDir(g_opts.user))
    destpath_secbox = "%s/secbox" % (backupdir)
    secboxConf = "*"
    if (checkFile(srcpath_secbox, secboxConf)):
        createBackupLibDir(g_opts, destpath_secbox)
        backupLibs(g_opts, secboxConf, srcpath_secbox, destpath_secbox)
        backupLibsToRemote(g_opts, secboxConf, 'secbox', srcpath_secbox, backupkey)

    g_logger.debug("end backup libs for cluster backup.")

def backupVersionFile(g_opts, backupkey):
    """
    backup upgrade version file
    """
    g_logger.debug("Backing up upgrade version file.")
    srcpath  = "%s/bin" % DefaultValue.getInstallDir(g_opts.user)
    destpath = os.path.join(g_opts.metadataDestination, \
        "roach", backupkey)
    versionFile = srcpath + "/" + VERSION_FILE
    if (os.path.exists(versionFile)):
        backupLibs(g_opts, VERSION_FILE, srcpath, destpath)
        backupLibsToRemote(g_opts, VERSION_FILE, VERSION_FILE, srcpath, backupkey)
    g_logger.debug("End backup upgrade version file.")

def backupETCDTerm(g_opts, backupkey):
    """
    backup etcd term into designated file
    """
    #just return if no etcd is deployed or it is master-standby cluster
    if (DefaultValue.hasEtcd(g_clusterInfo) == False or g_clusterInfo.isMasterStandbyCluster()):
        g_logger.debug("No need to backup etcd term file due to no etcd installed or master-standby cluster.")
        return

    g_logger.debug("Backing up etcd term.")

    etcd_term_key = "/%s/CMServer/status_key/term" % (g_opts.user)
    try:
        etcdTermDict = DefaultValue.getOldEtcdValue(g_logger, g_opts.user, g_clusterInfo, [etcd_term_key])
    except Exception, e:
        errorString = ErrorCode.GAUSS_534["GAUSS_53400"] + \
            "failed to get etcd term value for key %s, error: %s" % (etcd_term_key, str(e))
        agentErrorReport("roach_backup", errorString)
        g_logger.logExit(errorString)
    g_logger.debug("get etcd term result: key %s, value %s" % (etcd_term_key, etcdTermDict[etcd_term_key]))
    if (etcdTermDict[etcd_term_key].isdigit() is False):
        errorString = ErrorCode.GAUSS_534["GAUSS_53400"] + \
            "failed to get valid etcd term in backup etcd term"
        agentErrorReport("roach_backup", errorString)
        g_logger.logExit(errorString)

    filepath = os.path.join(g_opts.metadataDestination, \
        "roach", backupkey)
    etcdTermFile = os.path.join(filepath, ETCD_TERM_FILE)
    etcdTermFileTmp = os.path.join(filepath, ETCD_TERM_FILE_TMP)

    try:
        g_file.createFileInSafeMode(etcdTermFile)
        with open(etcdTermFileTmp, 'w') as fp:
            fp.write(etcdTermDict[etcd_term_key])
    except Exception, e:
        errorString = ErrorCode.GAUSS_534["GAUSS_53400"] + \
            "write etcd term temp file %s, error: %s" % (etcdTermFileTmp, str(e))
        agentErrorReport("roach_backup", errorString)
        g_logger.logExit(errorString)

    pNodeList = " -H ".join(g_clusterInfo.getClusterNodeNames())
    cmdPscp = "pscp -t %d -O StrictHostKeyChecking=no -H %s %s %s" % \
                    (TIMEOUT_SCP, pNodeList, etcdTermFileTmp, etcdTermFile)
    g_logger.debug("pscp etcd term file cmd: %s." % cmdPscp)
    status, output = commands.getstatusoutput(cmdPscp)
    if status != 0:
        errorString = ErrorCode.GAUSS_534["GAUSS_53400"] + \
            "Failed to distribute etcd term file to other nodes, output: %s" % (output)
        agentErrorReport("roach_backup", errorString)
        g_logger.logExit(errorString)

    backupLibsToRemote(g_opts, ETCD_TERM_FILE, ETCD_TERM_FILE, filepath, backupkey)

    rmCmd = "rm -f %s" % etcdTermFileTmp
    g_logger.debug("remove temporary etcd term file cmd: %s." % rmCmd)
    status, output = commands.getstatusoutput(rmCmd)
    if status != 0:
        g_logger.debug("Warning: failed to remove temporary etcd term file %s, output: %s" % (etcdTermFileTmp, output))

    g_logger.debug("End backup etcd term file.")

def backupClusterConfigFile(g_opts, backupkey):
    """
    backup cluster static config file to backup-key dir of metadata
    DWS will use this file to do restore after cluster resize
    """
    g_logger.debug("Backing up cluster static config file.")
    srcpath = "%s/bin" % DefaultValue.getInstallDir(g_opts.user)
    destpath = os.path.join(g_opts.metadataDestination, \
        "roach", backupkey)
    staticConfigFile = srcpath + "/" + STATIC_CONFIG_FILE
    if (os.path.exists(staticConfigFile)):
        backupLibs(g_opts, STATIC_CONFIG_FILE, srcpath, destpath)
        backupLibsToRemote(g_opts, STATIC_CONFIG_FILE, STATIC_CONFIG_FILE, srcpath, backupkey)
    g_logger.debug("End backup cluster static config file.")

def checkDatabase(g_opts, cnHost, coordPort, db):
    """
    Check database exist
    """
    sqlstring = "SELECT count(datname) FROM pg_database  WHERE DATNAME ='%s'" % db
    g_logger.debug("Check database cmd : \n[%s]" % (sqlstring))
    (status, output) = ClusterCommand.remoteSQLCommand(sqlstring, g_opts.user, \
        cnHost, coordPort, True, db)
    if status != 0:
        g_logger.logExit("Error: database \"%s\" does not exist!" %(db))
    if (0 == int(output)) :
        g_logger.logExit("Error: database \"%s\" does not exist!" %(db))


def checkSynonymTblExist(g_opts, cnHost, coordPort, db, schemaname, tablename):
    """
    Check if the synonyms table exists
    """
    sqlstring = "select a.synobjschema, a.synobjname " \
                "from (select synnamespace, synobjschema, synobjname " \
                "from pg_catalog.pg_synonym where synname = '%s') a " \
                "inner join " \
                "(select oid from pg_namespace where nspname = '%s') b " \
                "on a.synnamespace = b.oid;" % (tablename, schemaname)

    g_logger.debug("The cmd of check synonym table is : \n[%s]" % sqlstring)
    (status, output) = \
        ClusterCommand.remoteSQLCommand(sqlstring, g_opts.user, cnHost,
                                        coordPort, True, db)

    if status != 0 or output == "":
        g_logger.logExit("Error: table \"%s.%s\" does not exist "
                         "in database \"%s\"" % (schemaname, tablename, db))
    else:
        g_logger.debug("The synonym table information is: %s" % output)
        synonym_info = output.split("|")
        if len(synonym_info) != 2:
            g_logger.logExit("Error: table \"%s.%s\" does not exist in "
                             "database \"%s\"" % (schemaname, tablename, db))
        else:
            if g_opts.logical:
                g_logger.logExit("Error: roach does not support synonyms table,"
                                 " please modify \"%s.%s\" in file \"%s\" "
                                 "to \"%s.%s\""
                                 % (schemaname, tablename, g_opts.tablelist,
                                    synonym_info[0], synonym_info[1]))
            else:
                g_logger.logExit("Error: roach does not support synonyms table "
                                 "\"%s.%s\", please modify the value of "
                                 "parameter --tablename to \"%s.%s\""
                                 % (schemaname, tablename,
                                    synonym_info[0], synonym_info[1]))


def checkTable(g_opts, cnHost, coordPort, db, schema_table):
    """
    Check table exist
    """
    tablename = None
    schemaname = None

    if(schema_table.find('.')>=0):
        schema_table = schema_table.split(".")
        schemaname = schema_table[0].lower()
        tablename  = schema_table[1].lower()
    else:
        schemaname = "public"
        tablename  = schema_table.lower()
    sqlstring = "SELECT count(*) from PG_TABLES WHERE SCHEMANAME='%s' and TABLENAME='%s'" \
        % (schemaname, tablename)
    g_logger.debug("Check table cmd : \n[%s]" % (sqlstring))
    (status, output) = ClusterCommand.remoteSQLCommand(sqlstring, g_opts.user, \
        cnHost, coordPort, True, db)

    if status != 0:
        g_logger.logExit("Error: table \"%s.%s\" does not exist in database \"%s\"" \
            % (schemaname, tablename, db))
    if (0 == int(output)):
        checkSynonymTblExist(g_opts, cnHost, coordPort,
                             db, schemaname, tablename)


def checkTableHasTriggerOrSequence(g_opts, tableName, db, coordPort, cnHost):

    #Check trigger tables
    sql = "select relhastriggers from pg_class where oid ='%s'::regclass" % (tableName)
    g_logger.debug("excutor %s" % sql)
    status, output = ClusterCommand.remoteSQLCommand(sql, g_opts.user, cnHost, coordPort, False, db)
    if (status != 0):
        g_logger.logExit("Error: Failed to execute sql: %s, output: %s" % (sql, output))

    g_logger.debug("Output for check trigger table: [%s]" % output)
    output = (output.strip("\n")).strip()
    if (output == "t"):
        g_logger.logExit("Error: This table %s does not support backup/restore, bacause the table has triggers." % tableName)
    elif (output == ""):
        g_logger.logExit("Error: Failed to query database for trigger table %s." % tableName)

    #Check sequence tables
    sql = "select adrelid from pg_attrdef att, pg_class c where c.oid ='%s'::regclass and c.oid=att.adrelid and adbin like '%%seqName%%'" % (tableName)
    g_logger.debug("excutor %s" % sql)
    status, output = ClusterCommand.remoteSQLCommand(sql, g_opts.user, cnHost, coordPort, False, db)
    if (status != 0):
        g_logger.logExit("Error: Failed to execute sql: %s, output: %s" % (sql, output))

    g_logger.debug("Output for check sequence table: [%s]" % output)
    output = (output.strip("\n")).strip()
    if (output != ""): #it has sequence
        g_logger.logExit('Error: This table %s does not support '
                         'backup/restore, because the '
                         'table has sequence.' % tableName)

def checkReplicationTable(g_opts, tablename, db, coordPort, cnHost):
    """
    Check replication tables exist
    """
    sqlstring = "SELECT pclocatortype FROM pg_catalog.pgxc_class WHERE pcrelid =\'%s\'::regclass AND pclocatortype=\'R\';" % (tablename)
    g_logger.debug("Check replication table, cmd : \n[%s]" % sqlstring)

    (status, output) = ClusterCommand.remoteSQLCommand(sqlstring, \
        g_opts.user, cnHost, coordPort, False, db)
    if (status != 0):
        g_logger.logExit("Error: Failed to execute SQL command. %s\n" % output)
    if (output != ""):
        g_logger.debug("Check replication table, output : \n[%s]" % output)
        g_logger.logExit("Error:Replication type Table Backup not supported")

def checkTablesForLogical(g_opts, cnHost, schema_table):
    """
    Check for logical multi-table backups
    """
    if g_opts.dbName != "":
        db = g_opts.dbName
    else:
        db = 'postgres'
    coordPort = getCoHostPort(cnHost)
    coordPort = coordPort[1]

    #Check table exist
    checkTable(g_opts, cnHost, coordPort, db, schema_table)

    #Check for private tables
    checkPrivateTable(g_opts, cnHost, schema_table)

    #Check for tables created on user defined Node group with subset of DN's
    checkNodeGroup(g_opts, schema_table, db, coordPort, cnHost)

    #Check for Replication tables
    checkReplicationTable(g_opts, schema_table, db, coordPort, cnHost)

    #Check for trigger/sequence tables
    checkTableHasTriggerOrSequence(g_opts, schema_table, db, coordPort, cnHost)


def removeFile(master_generated_file):
    """
    remove intermediate file
    """
    cmd = 'rm -f %s' %(master_generated_file)
    if os.path.isfile(master_generated_file):
        output = commands.getstatusoutput(cmd)
        del_file_status = output[0]


def backupOnNodes(g_opts, command_dict, cnHost, cnPort, bkp_key = None):
    """
    backup on each nodes
    """
    global cluster_node_details
    global g_backup_is_stopped

    try:
        proc_list_dict = dict()
        validateRoachParams(g_opts)
        option = createMasterOption()
        create_roachPath(master_ip, g_opts)

        for key in final_ip_param_dict.keys():
            if '--tablename' in final_ip_param_dict[key].keys():
                if g_opts.dbName != "":
                    db = g_opts.dbName
                else:
                    db = 'postgres'

                tablename = g_opts.tableName
                if re.match(r'^"[^"]+"\."[^"]+"$', tablename):
                    tablename = tablename.replace('"', '\\\\\\"')
                if '`' in tablename:
                    tablename = tablename.replace('`', '\\\\\\`')

                #Check DB exist
                checkDatabase(g_opts, cnHost, cnPort, db)

                #Check table exist
                checkTable(g_opts, cnHost, cnPort, db, g_opts.tableName)

                #Check for private tables
                checkPrivateTable(g_opts, cnHost, g_opts.tableName)

                #Check for tables created on user defined Node group with subset of DN's
                checkNodeGroup(g_opts, tablename, db, cnPort, cnHost)

                #Check for Replication tables
                checkReplicationTable(g_opts, tablename, db, cnPort, cnHost)

                #Check for trigger/sequence tables
                checkTableHasTriggerOrSequence(g_opts, tablename, db, cnPort, cnHost)


            metaDest = ''
            if '--metadata-destination' in final_ip_param_dict[key].keys():
                metaDest = final_ip_param_dict[key]['--metadata-destination']
            if '--prior-backup-key' in final_ip_param_dict[key].keys():
                metaDest = getMetaDest(final_ip_param_dict)
                if not metaDest:
                    g_logger.logExit("Error: '--metadata-destination' not provided.")
                if not checkPriorBackup(metaDest, final_ip_param_dict[key]["--prior-backup-key"], command_dict):
                    g_logger.logExit("Validation of --prior-backup-key failed.")
                #cluster cluster topology matched
                bkpKeyList = getParentBackupKey(command_dict['--prior-backup-key'], \
                    final_ip_param_dict, command_dict)
                if (checkClusterNodeCount(g_opts, bkpKeyList[-1], metaDest) == False):
                    g_logger.logExit("Error: Node count not matched!")

            if ('--tablename' in final_ip_param_dict[key].keys()) \
              or ('--logical' in final_ip_param_dict[key].keys()):
                table_backup = True

            cmd = 'source %s&&%s/bin/gs_roach --master backup --master-ip \"%s\" %s' \
                %(g_opts.sourceFile, gauss_home_loc, master_ip, option)

            if (g_opts.clusterMigration == True):
                cmd += " --cluster-migration"

            if (g_opts.skipDelayDDL == False):
                cmd += " --exec-delay-ddl"

            g_opts.metadataDestination = metaDest #Save in object

        # removing old intermediate file before master creates same
        master_generated_file = '%s/bin/roach_nodes.done' %(gauss_home_loc)
        removeFile(master_generated_file)

        #if there is a stop process running, then quit backup
        if hasStopProcess(g_logger, g_opts, g_clusterInfo, cluster_version):
            g_logger.log("Warning: Quit backup due to user cancel or another backup operation is stopping!")    
            g_backup_is_stopped = True

        #If single Mode, no Cn can be used for querying pgxc_node 
        if not g_opts.single:
            initDNHaStatus = getDatanodeHAStatus(g_opts, cnHost, cnPort)
            if (len(initDNHaStatus) == 0):
                g_logger.log("ERROR: Failed to get datanode ha status before start back up!")
                setRoachOpResult(True)
                return

        #####################################################################
        #                  !! IMPORTANT SECURITY NOTE !!                    #
        # Regarding usage of subprocess.Popen([cmd], shell=True, bufsize=-1)#
        #  1.All the inputs from user are captured as individual options    #
        #  2.Direct inputs from the user are not passed in subprocess to    #
        #    gs_roach.                                                      #
        #  3.All the options are validated for the type as well as range    #
        #  4.The commands to be launched through subprocess are formed in   #
        #    code and passed as string to shell                             #
        #####################################################################

        g_logger.debug("python call gs_roach backup in backupOnNodes, cmd: %s" % cmd)
        master_proc = subprocess.Popen([cmd], shell=True, bufsize=-1)
        proc_list_dict.update({master_ip:master_proc})
        master_pid_cmd = 'ps -aef | grep %s | grep \"%s/bin/gs_roach --master backup --master-ip %s\" | grep -v grep | awk {\'print $2\'}' \
            %(g_opts.user, gauss_home_loc, master_ip)
        # wait time to get the master process id
        master_pid = ''
        master_pid = getMasterPID(master_pid_cmd, master_pid)
        if master_pid:
            proc_id_dict.update({master_ip:master_pid})
        tmp_val = 0
        while not os.path.isfile(master_generated_file) \
          and tmp_val < g_opts.metadataFileWaittime and master_proc.poll() == None:
            time.sleep(1)
            tmp_val = tmp_val + 1
        if os.path.isfile(master_generated_file):
            tmp_val = tuple()
            tmp_val = parseMaster_generated_file(master_generated_file)
            cluster_node_details = tmp_val[1]
            global backup_key
            backup_key = tmp_val[0]
        else:
            raise Exception("Error : Could not fetch cluster node details from metadata, roach_nodes.done file not found")

        #start agents
        startAgents(g_opts, command_dict, proc_list_dict)
    except Exception, e:
        g_logger.debug("Warning: Some exception happened in backupOnNodes, msg: %s" % str(e))
        signalAndExit(str(e))

    # make sure we do not miss concurrent stop backup operation
    if hasStopProcess(g_logger, g_opts, g_clusterInfo, cluster_version):
        g_logger.log("Warning: Send singal to kill gs_roach due to user cancel or another operation is stopping!")
        sendSignal()

    # poll all gs_roach processes in the following loop
    count = 0
    faileRetryTimes = 0
    pghost = DefaultValue.getEnv("PGHOST")
    while len(proc_list_dict) > 0:
        time.sleep(1)
        count = count + 1
        if (count >= 30):
            #check if switchover or failover has ocurred
            count = 0

            #when pitr basebackup is completed, no need to stop continuous backup
            if not g_opts.single:
                curDNHaStatus = getDatanodeHAStatus(g_opts, cnHost, cnPort)
                # check if there has been an expand during backup
                currentDatanodeCount = len(curDNHaStatus)
                _, oldDatanodeCount = getCurrentNodeCount(g_opts, g_clusterInfo)
                if (currentDatanodeCount != 0) and (oldDatanodeCount != currentDatanodeCount):
                    g_logger.log("Warning: The current datanode count %d is not match the datanode count %d that start backup." \
                                 "There may be expanded during the backup! Now stop this backup and mark it failed." \
                                    % (currentDatanodeCount, oldDatanodeCount))
                    sendSignal()
                for nodeI in initDNHaStatus:
                    for nodeC in curDNHaStatus:
                        if ((nodeI['nodename'] == nodeC['nodename']) and (nodeI['hostisprimary'] != nodeC['hostisprimary'])):
                            if (checkStartingPitrArchive(g_opts) or checkPitrStopInProgress(g_opts)) and g_opts.pitr:
                                continue
                            g_logger.log("Warning: Switchover or failover has happened!\nInitial datanode ha status:\n%s\nCurrent datanode ha status:\n%s" \
                                            % (str(initDNHaStatus), str(curDNHaStatus)))

                            if g_opts.pitr:
                                g_logger.log("unset Pitr Crontab for %s." % g_opts.action)
                                gaussHome = DefaultValue.getEnv("GAUSSHOME")
                                mpprcFile = g_opts.sourceFile
                                hostName = socket.gethostname()

                                status = unsetRoachCron("backup", gaussHome, hostName, mpprcFile, "GaussRoach.py", g_opts.logFile)
                                if (status != 0):
                                    g_logger.logExit("Stop backup failed failed.")
                                g_logger.log("unset Pitr Crontab successfully.")
                                
                            sendSignal()
                            
            #check cluster balance state for single cluster backup and resume backup
            faileRetryTimes = checkBalanceStateSame(g_opts, faileRetryTimes)
        if (count % 5 == 0) :
            deliveryOneItemToAgents(pghost, "/.roach_task_progress.dat", g_opts)
        pollSubprocess(proc_list_dict, g_opts)

    mark_completion_progress(g_opts)

def getMasterPID(master_pid_cmd, master_pid):
    """
    function to fetch master process id
    """
    pid_wait = 0
    while master_pid == '' and pid_wait < 5:
        time.sleep(1)
        master_pid_output = commands.getstatusoutput(master_pid_cmd)
        master_pid = master_pid_output[1]
        pid_wait = pid_wait + 1
    return master_pid

def check_mastreIP(master_ip):
    """
    check for multiple ip associated with master ip
    """
    global cluster_node_details
    if ' ' in master_ip:
        mast_ips = master_ip.split(' ')
        for ip in mast_ips:
            if ip in cluster_node_details.keys():
                cluster_node_details.pop(ip)

def startAgents(g_opts, command_dict, proc_list_dict):
    """
    start backup on agents
    """
    check_mastreIP(master_ip)
    for node_ip, host in cluster_node_details.items():
        create_roachPath(node_ip, g_opts)
        if node_ip != master_ip:
            makeRaochOptions(g_opts, node_ip, host, command_dict)
    validateRoachParams(g_opts)
    logOptions()

    ## removing --tablename from final_ip_param_dict and unneccessary keys from the dictionary final_ip_param_dict
    for node_ip in final_ip_param_dict.keys():
        if node_ip != master_ip:
            if '--tablename' in final_ip_param_dict[node_ip].keys():
                del final_ip_param_dict[node_ip]['--tablename']

    tempFinalDict = copy.deepcopy(final_ip_param_dict)
    create_command(g_opts)
    for node_ip, option in final_ip_param_dict.items():
        if node_ip != master_ip:
            proc_pid = '-1'
            c_log = createLogPath(g_opts.loggingPath)
            cmd = '%s&&pssh -t %s -H \"%s\" -P \"%s&& %s&&%s/bin/gs_roach\" --master-ip \\"%s\\" %s "2>&1"' \
                % (g_opts.sourceCmd, pssh_tout, node_ip, c_log, \
                g_opts.sourceCmd, gauss_home_loc, master_ip, option)
            if (g_opts.clusterMigration == True):
                cmd += " --cluster-migration"
            if (g_opts.skipDelayDDL == False):
                cmd += " --exec-delay-ddl"
            pid_cmd = '%s&&pssh -t %s -H \"%s\" -P \"ps -aef | grep %s | grep \"%s/bin/gs_roach --master-ip %s\" | grep -v grep | awk {\'print $2\'}\"' \
                % (g_opts.sourceCmd, pssh_tout, node_ip, g_opts.user, \
                gauss_home_loc, master_ip)

            #####################################################################
            #                  !! IMPORTANT SECURITY NOTE !!                    #
            # Regarding usage of subprocess.Popen([cmd], shell=True, bufsize=-1)#
            #  1.All the inputs from user are captured as individual options    #
            #  2.Direct inputs from the user are not passed in subprocess to    #
            #    gs_roach.                                                      #
            #  3.All the options are validated for the type as well as range    #
            #  4.The commands to be launched through subprocess are formed in   #
            #    code and passed as string to shell                             #
            #####################################################################
            g_logger.debug("python call gs_roach backup in startAgents, cmd: %s" % cmd)
            agent_proc = subprocess.Popen([cmd], stdout=subprocess.PIPE, \
                         stderr=subprocess.PIPE, shell=True, bufsize=-1)
            proc_list_dict.update({node_ip:agent_proc})
            proc_for_pid = subprocess.Popen([pid_cmd], \
                           stdout=subprocess.PIPE, \
                           stderr=subprocess.PIPE, shell=True, bufsize=-1)
            proc_output = proc_for_pid.stdout.readlines()
            if proc_output:
                proc_pid = proc_output[0].strip()
            if proc_pid.isdigit() and (proc_pid != '-1'):
                proc_id_dict.update({node_ip:proc_pid})

    return proc_list_dict

def checkPriorBackup(metaDest, priorBkpKey, command_dict):
    """
    function: to validate the prior backup key
    input: metaDest, priorBkpKey, command_dict
    output: boolean
    """
    try:
        media_type = command_dict["--media-type"]
        metaDest = os.path.realpath(metaDest)
        if os.path.isdir(metaDest):
            if priorBkpKey:
                backupdir = os.path.join(metaDest, "roach", priorBkpKey)
                if os.path.isdir(backupdir):
                    backupfile = os.path.join(backupdir, "roach_master_updated.metadata")
                    chkprior = loadMetaFileAsJson(backupfile)
                    if chkprior["Nodes"][0].has_key("PriorBackupKey"):
                        if chkprior["BackupType"].upper() != "TABLE":
                            g_logger.debug("Successfully validated prior backup key.")
                        else:
                            g_logger.log("Error: Specified prior-backup-key is corresponding to table backup.\nIncremental/Differential backup on table backup is not supported.")
                            return False
                    else:
                        g_logger.log("Error: Incremental backup is not supported over specified backup-key [%s]. Please take full backup." 
                            % str(priorBkpKey))
                        return False 
                else:
                    errorString = ErrorCode.GAUSS_534["GAUSS_53404"] % priorBkpKey
                    errorString = "%s in metadata-destination [%s]." % (errorString, metaDest)
                    agentErrorReport("roach_backup", errorString)
                    g_logger.log("Error: %s" % errorString)
                    return False
            else:
                priorBkpKey = "NULL"
                errorString = ErrorCode.GAUSS_534["GAUSS_53404"] % priorBkpKey
                errorString = "%s in metadata-destination [%s]." % (errorString, metaDest)
                agentErrorReport("roach_backup", errorString)
                g_logger.log("Error: %s" % errorString)
                return False
        else:
            g_logger.log("Error: Specified metadata-destination [%s] does not exist." \
                % str(metaDest))
            return False
    except OSError as oserr:
        g_logger.log(str(oserr))
        return False

    return True


def checkPriorBackupIsRestoreSuccess(metaDest,
                                     command_dict, skipDependentBackup):
    """
    Check if the parent backup was restored successfully
    """
    backupKey = command_dict["--prior-backup-key"]
    bkpdetails = ""
    bkpkeylist = list()

    def searchBackupKey(bkpkey, skipDependentBackup):
        """
        function: recursive fn to find the backup key in the json
        input: backup key
        output: list of backup keys
        """
        for bkpdict in bkpdetails:
            if "BackupKey" in bkpdict and bkpdict["BackupKey"] == bkpkey:
                if bkpdict["MediaType"] == "DISK" and bkpdict["MediaDestination"]:
                    metapath = os.path.join(metaDest, "roach", bkpkey)
                    if not os.path.isdir(metapath):
                        g_logger.logExit("Error: Backup key [%s] does not exist "
                                         "in metadata-destination [%s], "
                                         "restore not possible!"
                                         % (bkpkey, metaDest))
                bkpkeylist.append(str(bkpkey))
                if skipDependentBackup:
                    break
                if "PriorBackupKey" in bkpdict and \
                        bkpdict["PriorBackupKey"][:4] == "FULL":
                    if bkpdict["PriorBackupKey"] == "FULL" and \
                            int(bkpdict["Timeline ID"]) == \
                            int(TLI_HAS_RESTORED):
                        g_logger.exitWithError("Error: Incremental backup not "
                                               "allowed if restore occurred on "
                                               "prior-backup-key ['%s']. "
                                               "Perform FULL backup." %
                                               command_dict['--prior-backup-key'])
                    break
                else:
                    searchBackupKey(bkpdict["PriorBackupKey"], skipDependentBackup)

    # reading backups.ini in restore operation
    backupfile = os.path.join(metaDest, "backups.ini")
    if os.path.isfile(backupfile):
        chkprior = loadMetaFileAsJson(backupfile)
        if "BackupDetails" in chkprior:
            bkpdetails = chkprior["BackupDetails"]
            keyExists = False
            for bkpdict in bkpdetails:
                if "BackupKey" in bkpdict and bkpdict["BackupKey"] == backupKey:
                    keyExists = True
                    if bkpdict["BackupType"] == "INCREMENTAL":
                        searchBackupKey(backupKey, skipDependentBackup)
                    else:
                        if bkpdict["PriorBackupKey"] == "FULL" and\
                                int(bkpdict["Timeline ID"]) == \
                                int(TLI_HAS_RESTORED):
                            g_logger.exitWithError("Error: Incremental backup "
                                                   "not allowed if restore "
                                                   "occurred on prior-backup-key"
                                                   " ['%s']. Perform FULL backup."
                                                   % command_dict['--prior-backup-key'])
                    break
            if not keyExists:
                g_logger.logExit("Error: "
                                 "Specified backup key ['%s'] does not exist or"
                                 " backup type is logical backup." % backupKey)
    else:
        g_logger.logExit("Error: "
                         "Specified metadata-destination [%s] is corrupted, "
                         "restore not possible!" % (metaDest))


def validatePreviousBackups(g_opts, command_dict):
    """
    validate incremental backup
    """
    if "--backup-key" in command_dict.keys():
        g_logger.exitWithError("Error: Invalid parameter '--backup-key'")

    if (isClusterBackup(g_opts, command_dict) and ("--prior-backup-key" in
                                                   command_dict)):
        checkPriorBackupIsRestoreSuccess(g_opts.metadataDestination,
                                         command_dict, g_opts.skipDependentBackup)

    if "--validate-prior-backups" in command_dict.keys():
        g_opts.origAction = "backup" #Change the action type before "validate"
        g_opts.action = "validate" #Change the action type before "validate"
        bkpKeyList = getParentBackupKey(command_dict['--prior-backup-key'], \
            final_ip_param_dict, command_dict)
        backupDetails = getBackupDetails(command_dict, final_ip_param_dict)
        tempStatus = dict()
        validatePreBkpStr = command_dict['--validate-prior-backups'].upper()
        for bkpkey in bkpKeyList:
            for bkpdet in backupDetails:
                if bkpdet["BackupKey"] == bkpkey:
                    if bkpdet["STATUS"].upper() == "CORRUPTED" and validatePreBkpStr == "OPTIMISTIC":
                        g_logger.logExit("Error: Backup for the backup-key ['%s'] is corrupted." % bkpkey)
                        break
                    else:
                        tempStatus[bkpkey] = bkpdet["STATUS"].upper()
        gs_roach_op_fail = ""
        if validatePreBkpStr == "FORCE":
            temp_cmd_dict = copy.deepcopy(command_dict)
            del temp_cmd_dict['--prior-backup-key']
            del temp_cmd_dict['--validate-prior-backups']
            gs_roach_op_fail = startValidate(g_opts, temp_cmd_dict, bkpKeyList[0])
            g_opts.action = "backup" #Change the action type after "validate" ends
        elif validatePreBkpStr == "OPTIMISTIC":
            bkpKeyList.reverse()
            for item in range(len(bkpKeyList)):
                bkpkey = bkpKeyList[item]
                if bkpkey in tempStatus.keys():
                    if tempStatus[bkpkey] == "VALIDATED":
                        print "--------------------------------------------------------------------------------"
                        print "In VALIDATED state already! Skip validation for backup key [%s/%s]: %s" \
                            % (str(item+1), str(len(bkpKeyList)), bkpkey)
                        continue
                    temp_cmd_dict = copy.deepcopy(command_dict)
                    del temp_cmd_dict['--prior-backup-key']
                    del temp_cmd_dict['--validate-prior-backups']
                    print "--------------------------------------------------------------------------------"
                    print "Validate operation in-progress for 'INCREMENTAL BACKUP' key[%s/%s]: %s" \
                        % (str(item+1), str(len(bkpKeyList)), bkpkey)
                    gs_roach_op_fail = startValidate(g_opts, temp_cmd_dict, bkpkey, True)
                    g_opts.action = "backup" #Change the action type after "validate" ends
                    if gs_roach_op_fail:
                        g_logger.log("Error: Validate operation failed for backup key: %s" % bkpkey)
                        return gs_roach_op_fail
                    
        
        ## clear and form new final_ip_param_dict
        final_ip_param_dict.clear()
        makeRaochOptions(g_opts, master_ip, master_host, command_dict)

        if gs_roach_op_fail:
            g_logger.logExit("Error: Validation failed for one of the backup-key.") 

def updateForBackupFailures(g_opts):
    """
    if backup has failed, update backups.ini status as 'FAILED'
    """
    if getRoachOpResult():
        if g_opts.pitr:
            if g_opts.resumeBkp:
                g_logger.debug("No need to update backup ini when in pitr resume backup.")
            else:
                updateBackupIniFailures(g_logger, g_opts, backup_key)
        else:
            updateBackupIniFailures(g_logger, g_opts, backup_key)


#set guc and write state
def resetFullPageWrites(g_opts, g_logger):
    cmd = "source %s && python %s -t %s -U %s -l %s" \
        % (g_opts.sourceFile, LOCAL_ROACH, ACTION_RESET_FULLPAGEWRITES, g_opts.user, g_opts.logFile)
    g_logger.debug("GaussRoach.py calling LocalRoach.py to reset guc FullPageWrite, cmd=[%s]" % (cmd))
    try:
        sshTool = SshTool(g_clusterInfo.getClusterNodeNames(), g_logger.logFile, TIMEOUT_PSSH_PRECHECK)
        sshTool.executeCommand(cmd, "reset guc FullPageWrite")
        g_logger.log("Successfully reset guc FullPageWrite.")
    except Exception, e:
        g_logger.error("Failed to reset guc FullPageWrite: %s. Please reset it manully." % str(e))

#get hostis_primay information for datanodes from pgxc_node
def getDatanodeHAStatus(g_opts, cnHost, cnPort):
    """
    get datanode ha status from pgxc_node
    """
    HaInfo = []
    sql = "select node_name, hostis_primary from pgxc_node where node_type='D' order by node_name;"
    (status, output) = ClusterCommand.remoteSQLCommand(sql, g_opts.user, cnHost, cnPort, False)
    if (status != 0):
        g_logger.debug("Warning: failed to get datanode ha status!Output: %s" % (output))
        return HaInfo

    resList = output.split('\n')
    for eachLine in resList:
        tmpInfo = {}
        (nodeName, hostIsPrimary) = eachLine.split('|')
        tmpInfo['nodename'] = nodeName.strip()
        tmpInfo['hostisprimary'] = hostIsPrimary.strip()
        HaInfo.append(tmpInfo)
    return HaInfo

def getCurrentVersion(g_opts, gaussHome):
    cmd = "cat %s/bin/upgrade_version" % gaussHome
    g_logger.debug("Getting current upgrade_version, cmd: %s" % cmd)
    status, currVersion = DefaultValue.retryGetstatusoutput(cmd, 60)
    if status != 0:
        g_logger.logExit("Error: Unable to get current version info, output: %s" % currVersion)
    g_logger.debug("Current upgrade_version is: %s" % currVersion)
    versionList = currVersion.split('\n')[:2]
    return versionList

def checkVersionAfterBackup(g_opts, gaussHome, oldVersion):
    if not checkUpgradeVersion(g_opts):
        setRoachOpResult(True)
    currVersion = getCurrentVersion(g_opts, gaussHome)
    if (currVersion != oldVersion):
        g_logger.log("Error: An upgrade occurred during the backup!!")
        setRoachOpResult(True)

def checkDatanodeCountAfterBackup(g_opts, oldNodeCount, oldDatanodeCount):
    currentNodeCount, currentDatanodeCount = getCurrentNodeCount(g_opts)
    if (oldNodeCount != currentNodeCount) or (oldDatanodeCount != currentDatanodeCount):
        g_logger.log("Error: An expend occurred during the backup!!")
        setRoachOpResult(True)

def startBackup(cnHost, g_opts, sshTool, command_dict, table_list=[]):
    """
    call for backup functionality
    """
    global g_backup_is_stopped
    global master_ip

    master_ip = getMaster_ip()
    if g_opts.roachConf:
        parseRoachConf(g_opts)
    makeRaochOptions(g_opts, master_ip, master_host, command_dict)
    validate_logparams()

    global g_logger
    g_logger = gsinitLogger(g_opts)

    #get clusterInfo
    global g_clusterInfo
    gaussHome = DefaultValue.getEnv("GAUSSHOME")
    staticConfig = "%s/bin/cluster_static_config" % gaussHome
    staticConfig_bak = "%s/bin/cluster_static_config_bak" % gaussHome
    if (not os.path.exists(staticConfig)):
        g_logger.debug("Warning: static config file not found, use its bak file!")
        staticConfig = staticConfig_bak
    oldVersion = getCurrentVersion(g_opts, gaussHome)
    g_clusterInfo = dbClusterInfo()
    g_clusterInfo.initFromStaticConfig(g_opts.user, staticConfig)

    # get current node count at the begin of backup
    oldNodeCount, oldDatanodeCount = getCurrentNodeCount(g_opts, g_clusterInfo)

    #get cluster version used for miniaturized deployment 
    global cluster_version
    cluster_version = getClusterVersion(g_opts.user)
    if g_opts.logical and g_opts.backupKey != "":
        g_logger.logExit("Error: Invalid parameter '--backup-key'")
    validateRoachParams(g_opts)

    #for incremental backup, need version is match
    if (isClusterBackup(g_opts, command_dict) and \
      command_dict.has_key('--prior-backup-key')):
        incBackupKey = command_dict['--prior-backup-key']
        g_logger.debug("startBackup: this is an incremental backup, "
            "need check version is matched, backup key is: %s" % incBackupKey)
        compareVersionFile(g_opts, incBackupKey)

    #if there is a stop process running, then quit backup
    if hasStopProcess(g_logger, g_opts, g_clusterInfo, cluster_version):
        g_logger.log("Warning: Quit backup due to user cancel or another backup operation is stopping!")    
        return (True, None)

    #do validate on previous backups
    validatePreviousBackups(g_opts, command_dict)
    updateCN_hosts(cnHost, g_opts, command_dict)
    if g_opts.logical:
        command_dict['--tablename'] = table_list[0]
    if (checkClusterNormal(g_opts, g_logger) != 0):
        errorString = ErrorCode.GAUSS_516["GAUSS_51602"] + "Please make sure cluster_state is Normal."
        agentErrorReport("roach_backup", errorString)
        g_logger.logExit(errorString)
    if (g_opts.action == "backup" and isClusterBackup(g_opts, command_dict) and not checkStartingPitrArchive(g_opts)):
        turnOnCBM(g_opts, sshTool)
    g_logger.debug("Checking flock of python in startBackup")
    check_gs_roach_running(g_opts)
    
    #start backup operation
    try:
        doBackup(g_opts, command_dict, cnHost, table_list)
        if (not g_opts.resumeBkp):
            postExecCleanup(g_opts, g_logger)
            rmMediaBackupDirs(g_opts, backup_key)
        elif (not getRoachOpResult()) and (not checkStartingPitrArchive(g_opts)) and (not checkPitrStopInProgress(g_opts)):
            resetFullPageWrites(g_opts, g_logger)
            if (not g_opts.clusterMigration):
                postExecCleanup(g_opts, g_logger)
            rmMediaBackupDirs(g_opts, backup_key)
    except Exception, e:
        g_logger.debug("Warning: doBackup may met some issue, catch exception: %s" % str(e))
        setRoachOpResult(True)

    # check if there has been an upgrade during backup
    checkVersionAfterBackup(g_opts, gaussHome, oldVersion)

    # check if there has been an expand during backup
    checkDatanodeCountAfterBackup(g_opts, oldNodeCount, oldDatanodeCount)

    updateForBackupFailures(g_opts)
    if (g_backup_is_stopped):
        return (True, None)
    return (getRoachOpResult(), backup_key)


def turnOnCBM(g_opts, sshTool):
    """
    For cluster backup, turn on CBM function
    """
    g_logger.debug("Turn on CBM function.")
    cmd = "source %s && python %s -t %s --enableCBM -U %s -l %s" \
        % (g_opts.sourceFile, LOCAL_ROACH, ACTION_CONFIG,
           g_opts.user, g_opts.logFile)
    g_logger.debug("Calling LocalRoach.py to set \enable_cbm_tracking param, cmd=%s." % cmd)
    if not isMiniaturizedDeployment(cluster_version):
        sshTool.executeCommand(cmd, "set enable_cbm_tracking parameter", DefaultValue.SUCCESS, g_opts.nodeList)
    else:
        status, output = commands.getstatusoutput(cmd)
        if (status != 0):
            g_logger.logExit("Turn on CBM failed with output: %s" % (output))
    g_logger.debug("Successfully turn on CBM function.")
    return (getRoachOpResult(), backup_key)
