try:
    import sys
    import os
    import getopt
    import commands
    import pwd
    import time
    import string
    import signal
    import subprocess
    import multiprocessing
    import json
    import getpass
    from threading import Timer
    from datetime import datetime, timedelta
    
    sys.path.append(sys.path[0] + "/../")
    from gspylib.common.DbClusterInfo import dbClusterInfo
    from gspylib.common.GaussLog import GaussLog
    from gspylib.threads.SshTool import SshTool
    from gspylib.common.Common import DefaultValue, ClusterCommand
    from gspylib.os.gsOSlib import g_OSlib, VersionInfo
    from gspylib.os.gsfile import g_file
    from gspylib.common.OMCommand import OMCommand
    from gspylib.common.ErrorCode import ErrorCode
    import util.GSroachUtils
    from util.GSroachCommon import *
except ImportError, e:
    sys.exit('Error:\nUnable to import module: ' + str(e))

g_logger = None
LOG_PATH = DefaultValue.getEnv("GAUSSLOG")
DEFAULT_LOG_PATH = "%s/roach/controller" % LOG_PATH
LOG_FILE_NAME = "roach-controller.log"
g_logger = GaussLog(util.GSroachUtils.createFilePath(DEFAULT_LOG_PATH, LOG_FILE_NAME), "ROACH")

def updateClusterInstanceID(g_opts, g_sshTool):
    """
    refresh instance ID in static config files
    """
    gaussHome = DefaultValue.getEnv("GAUSSHOME")
    staticConfigFile = "%s/bin/cluster_static_config" % gaussHome
    #save the refreshed static config file of one node
    templateStaticConfigFile = "%s/bin/cluster_static_config_ref" % gaussHome
    #primary cluster info
    backClusterInfo = dbClusterInfo()
    backClusterInfo.initFromStaticConfig(g_opts.user, g_opts.staticConfigPath)
    #standby cluster info
    currentClusterInfo = backClusterInfo
    newDatanodes = []
    oldDatanodes = []

    g_logger.log("Refreshing static config files...")

    #make a copy of static config file
    g_logger.log("Save a copy of static config file for all nodes.")
    if not gaussHome:
        g_logger.logExit("ERROR: GAUSSHOME is None or empty string!")
    saveStaticConfigFile(gaussHome, g_sshTool)

    #only append master DN instances
    for newNode in currentClusterInfo.dbNodes:
        for instance in newNode.datanodes:
            if instance.instanceType == 0:
                newDatanodes.append(instance)

    for oldNode in backClusterInfo.dbNodes:
        for instance in oldNode.datanodes:
            if instance.instanceType == 0:
                oldDatanodes.append(instance)

    #refresh instance ID
    updateDatanodeInstance(newDatanodes, oldDatanodes, currentClusterInfo, backClusterInfo)

    #delivery static config files
    g_logger.debug("Delivery static config files to other nodes.")
    for newNode in currentClusterInfo.dbNodes:
        currentClusterInfo.saveToStaticConfig(templateStaticConfigFile, newNode.id, currentClusterInfo.dbNodes)
        cmd = "pscp -H %s %s %s" \
            %(newNode.name, templateStaticConfigFile, staticConfigFile)
        status, output = commands.getstatusoutput(cmd)
        if (status != 0):
            g_logger.logExit("Copy static config file failed. output: %s" % output)
        else:
            g_logger.debug("%s distribute success." % newNode.name)

    #remove dynamic config files
    removeDynamicConfigFile(gaussHome, g_sshTool)

    #remove temp config file
    removeLocalFile(g_logger, templateStaticConfigFile)
    g_logger.log("Successfully refreshed static config files.")


def updatePostgresConf(g_opts, g_sshTool):
    """
    refresh postgresql.conf of each instance
    """
    g_logger.log("Refreshing postgresql.conf for instance name.")
    cmd = "source %s && python %s -t %s -U %s -l %s" % (
        g_opts.sourceFile, LOCAL_ROACH, ACTION_REFRESH_CONF, 
        g_opts.user, g_opts.logFile)
    g_logger.debug("GaussRoach.py calling LocalRoach.py to update postgresql.conf cmd=[%s]" % cmd)
    g_sshTool.executeCommand(cmd, "update postgresql.conf")
    g_logger.log("Successfully refreshed postgresql.conf for instance name.")


def cleanOldInstanceDirs(g_opts, g_sshTool):
    """
    clean some dirs with old instance name
    """
    g_logger.log("Clean dirs with old instance name started.")

    #clean pg_log
    if not LOG_PATH:
        g_logger.logExit("ERROR: GAUSSLOG is None or empty string!")
    cleanOneLogDir(LOG_PATH, "pg_log")

    #clean pg_audit
    cleanOneLogDir(LOG_PATH, "pg_audit")

    #clean gs_profile
    cleanOneLogDir(LOG_PATH, "gs_profile", g_sshTool)

    #clean pg_replslot
    cmd = "source %s && python %s -t %s -U %s -l %s" % (
        g_opts.sourceFile, LOCAL_ROACH, ACTION_CLEAN_REPLSLOT, 
        g_opts.user, g_opts.logFile)
    g_logger.debug("GaussRoach.py calling LocalRoach.py to clean pg_replslot cmd=[%s]" % cmd)
    g_sshTool.executeCommand(cmd, "clean pg_replslot")

    g_logger.log("Successfully cleaned dirs with old instance name.")


def updatePgxcNodeName(g_opts, g_sshTool):
    """
    update DN pgxc nodename for instance name
    """
    g_logger.log("Update pgxc nodename started.")
    cmd = "source %s && python %s -t %s -U %s -l %s" % (
        g_opts.sourceFile, LOCAL_ROACH, ACTION_REFRESH_NODENAME, 
        g_opts.user, g_opts.logFile)
    g_logger.debug("GaussRoach.py calling LocalRoach.py to update DN pgxc nodename cmd=[%s]" % cmd)
    g_sshTool.executeCommand(cmd, "update pgxc nodename")
    g_logger.log("Update pgxc nodename completed.")


def updateDatanodeInstance(newDataNodes, oldDataNodes, currClusterInfo, backClusterInfo):
    """
    replace new cluster instance ID with old cluster instance ID
    """
    if (len(newDataNodes) > 0) and (len(oldDataNodes) > 0):
        for new, old in zip(newDataNodes, oldDataNodes):
            if not new.instanceId == old.instanceId:
                new.instanceId = old.instanceId

            newInstances = currClusterInfo.getPeerInstance(new)
            oldInstances = backClusterInfo.getPeerInstance(old)
            for newIns in newInstances:
                for oldIns in oldInstances:
                    if newIns.instanceType == oldIns.instanceType:
                        newIns.instanceId = oldIns.instanceId


def removeDynamicConfigFile(gaussHome, g_sshTool):
    """
    delete all the dynamic config file
    """
    dynamic_file = "%s/bin/cluster_dynamic_config" % gaussHome
    cmd = "if [ -f %s ]; then rm -f %s; fi" % (dynamic_file, dynamic_file)
    g_logger.debug("cmd to remove dynamic config file is: %s" % cmd)
    g_sshTool.executeCommand(cmd, "remove dynamic config file")


def saveStaticConfigFile(gaussHome, g_sshTool):
    """
    make a copy of static config file before modify it.
    and it will also be used for generate map on local CN nodes.
    """
    confFile = "%s/bin/cluster_static_config" % gaussHome
    confCopy = "%s/bin/orig_cluster_static_config" % gaussHome
    cmd = "cp -f %s %s" % (confFile, confCopy)
    g_logger.debug("cmd to make copy of config file is: %s" % cmd)
    g_sshTool.executeCommand(cmd, "copy static config file")


def cleanOneLogDir(dirPath, dirType, g_sshTool):
    """
    clean one type of log dir for DN instance
    """
    cmd = "rm -rf %s/%s/dn_*" % (dirPath, dirType)
    g_logger.debug("cmd to clean %s is: %s" % (dirType, cmd))
    g_sshTool.executeCommand(cmd, "clean %s" % dirType)



