#!/usr/bin/env python
# -*- coding:utf-8 -*-
#############################################################################
#Copyright (c): 2020, Huawei Tech. Co., Ltd.
#FileName     : GSroachValidate.py
#Version      : V1.0.0
#Date         : 2020-01-20
#Description  : GSroachValidate.py is a utility for double cluster synchronization.
#############################################################################

try:
    import sys
    import time
    import os
    import commands
    import subprocess
    import signal
    import json
    import re
    import copy

    sys.path.append(sys.path[0] + "/../")
    from gspylib.common.GaussLog import GaussLog
    from gspylib.common.ErrorCode import ErrorCode
    from util.GSroachCommon import parseRoachConf,\
        makeRaochOptions, gsinitLogger, parseMaster_generated_file, master_ip, \
        master_host, remove_pattern, gauss_home_loc, pssh_tout, dropUnwantedParams, \
        check_gs_roach_running, getRoachOpResult, \
        final_ip_param_dict, proc_id_dict, create_command, sendSignal, pollSubprocess, \
        g_logger, sigint_handler, validate_logparams, create_roachPath,\
        reform_tablename, createLogPath, getMetaDest, postExecCleanup, \
        getBackupDetails, getArchiveBackupKeys, getMaster_ip, DEF_BKP_KEY_SIZE, \
        loadMetaFileAsJson, agentErrorReport
    from util.GSroachCommonParameter import validateRoachParams
except ImportError, e:
    sys.exit('Error: Unable to import module: ' + str(e))


#global variables
userProfile = "~/.bashrc"
osProfile = "/etc/profile"
g_mediatype = ""
g_mediadest = ""
g_backupkey = ""
g_bkpkeylist = None
g_isIncremental = False
master_ip = ""


signal.signal(signal.SIGINT, sigint_handler)


def doValidate(g_opts, command_dict, bkp_key):
    """
    do validate operation
    """
    g_logger.debug("start validate operation")
    validateOnNodes(g_opts, command_dict, bkp_key)
    g_logger.debug("end validate operation")

def logOptions():
    """
    writing to logfile
    """
    for key in final_ip_param_dict.keys():
        g_logger.debug("options for %s" %key)
        for key, val in final_ip_param_dict[key].items():
            g_logger.debug("%s : %s" %(key.translate(remove_pattern, "--"), val))

def createMasterOption(command_dict):
    """
    create option for master
    """
    #global dictionary
    global final_ip_param_dict
    new_value = ''
    final_value = ''

    for k in final_ip_param_dict.keys():
        for key, val in final_ip_param_dict[k].iteritems():
            if val != None:
                if key == '--tablename':
                    val = reform_tablename(val)
                    new_value = ''.join("%s %s " %(key, val))
                else:
                    new_value = ''.join("%s %r " %(key, val))
            else:
                new_value = ''.join("%s " %(key))
            final_value += new_value
    for key, val in command_dict.iteritems():
        if key == "--validation-type":
            if val != None:
                new_value = ''.join("%s %r " %(key, val))
            else:
                new_value = ''.join("%s " %(key))
            final_value += new_value
    return final_value


def validateOnNodes(g_opts, command_dict, bkp_key):
    """
    restore on each nodes
    """

    g_logger.debug("Checking flock of python in validateOnNodes")
    check_gs_roach_running(g_opts)
    try:
        proc_list_dict = dict()
        create_roachPath(master_ip, g_opts)
        dropUnwantedParams(g_opts)
        validateRoachParams(g_opts)
        for node_ip in final_ip_param_dict.keys():
            final_ip_param_dict[node_ip]["--media-destination"] = g_mediadest
            final_ip_param_dict[node_ip]["--media-type"] = g_mediatype
            final_ip_param_dict[node_ip]["--backup-key"] = g_backupkey
        option = createMasterOption(command_dict)
        create_roachPath(master_ip, g_opts)

        cmd = 'source %s&&%s/bin/gs_roach --master validate --master-ip \"%s\" %s' \
            % (g_opts.sourceFile, gauss_home_loc, master_ip, option)
        if g_opts.action == 'backup':
            cmd = cmd + '--backup-key %s' % (bkp_key)

        # removing old intermediate file before master creates same
        master_generated_file = '%s/bin/roach_nodes.done' %(gauss_home_loc)
        removeFile(master_generated_file)

        #####################################################################
        #                  !! IMPORTANT SECURITY NOTE !!                    #
        # Regarding usage of subprocess.Popen([cmd], shell=True, bufsize=-1)#
        #  1.All the inputs from user are captured as individual options    #
        #  2.Direct inputs from the user are not passed in subprocess to    #
        #    gs_roach.                                                      #
        #  3.All the options are validated for the type as well as range    #
        #  4.The commands to be launched through subprocess are formed in   #
        #    code and passed as string to shell                             #
        #####################################################################

        g_logger.debug("python call gs_roach validate in validateOnNodes, action: %s, cmd: %s" % (g_opts.action, cmd))
        master_proc = subprocess.Popen([cmd], shell=True, bufsize=-1)
        proc_list_dict.update({master_ip:master_proc})
        master_pid_cmd = 'ps -aef | grep %s | grep \"%s/bin/gs_roach --master validate --master-ip %s\" | grep -v grep | awk {\'print $2\'}' \
            %(g_opts.user, gauss_home_loc, master_ip)
        # wait time to get the master process id
        master_pid = ''
        master_pid = getMasterPID(master_pid_cmd, master_pid)
        if master_pid:
            proc_id_dict.update({master_ip:master_pid})
        cluster_node_details = ""
        tmp_val = 0
        while not os.path.isfile(master_generated_file) \
          and tmp_val < g_opts.metadataFileWaittime and master_proc.poll() == None:
            time.sleep(1)
            tmp_val = tmp_val + 1
        if os.path.isfile(master_generated_file):
            tmp_val = tuple()
            tmp_val = parseMaster_generated_file(master_generated_file)
            cluster_node_details = tmp_val[1]
        else:
            raise Exception("Error : Could not fetch cluster node details from metadata, roach_master.metadata or roach_nodes.done file not found")

        #start agents
        startAgents(cluster_node_details, g_opts, command_dict, proc_list_dict)
    except (IOError, OSError) as err:
        g_logger.debug("Some exception happened in validateOnNodes")
        sendSignal()
        GaussLog.exitWithError(str(err))
    while len(proc_list_dict) > 0:
        time.sleep(1)
        pollSubprocess(proc_list_dict, g_opts)

def getMasterPID(master_pid_cmd, master_pid):
    """
    function to fetch master process id
    """
    pid_wait = 0
    while master_pid == '' and pid_wait < 5:
        time.sleep(1)
        master_pid_output = commands.getstatusoutput(master_pid_cmd)
        master_pid = master_pid_output[1]
        pid_wait = pid_wait + 1
    return master_pid


def check_mastreIP(cluster_node_details):
    """
    check for multiple ip associated with master ip
    """
    if ' ' in master_ip:
        mast_ips = master_ip.split(' ')
        for ip in mast_ips:
            if ip in cluster_node_details.keys():
                cluster_node_details.pop(ip)

def startAgents(cluster_node_details, g_opts, command_dict, proc_list_dict):
    """
    start restore on agents
    """
    check_mastreIP(cluster_node_details)
    for node_ip, host in cluster_node_details.items():
        create_roachPath(node_ip, g_opts)
        makeRaochOptions(g_opts, node_ip, host, command_dict)
    dropUnwantedParams(g_opts)
    validateRoachParams(g_opts)
    logOptions()

    ##removing --tablename from final_ip_param_dict
    for node_ip in final_ip_param_dict.keys():
        if node_ip != master_ip:
            if '--tablename' in final_ip_param_dict[node_ip].keys():
                del final_ip_param_dict[node_ip]['--tablename']
            final_ip_param_dict[node_ip]["--media-destination"] = g_mediadest
            final_ip_param_dict[node_ip]["--media-type"] = g_mediatype
            final_ip_param_dict[node_ip]["--backup-key"] = g_backupkey

    ## creating deepcopy of the final ip param dictionary
    tempFinalDict = copy.deepcopy(final_ip_param_dict)

    create_command(g_opts)
    for node_ip, option in final_ip_param_dict.items():
        if node_ip != master_ip:
            proc_pid = '-1'
            c_log = createLogPath(g_opts.loggingPath)
            cmd = 'source %s&&pssh -t %s -H \"%s\" -P \"%s&& source %s&&%s/bin/gs_roach\" --master-ip \\"%s\\" %s "2>&1"' \
                % (g_opts.sourceFile, pssh_tout, node_ip, c_log, g_opts.sourceFile, gauss_home_loc, master_ip, option)
            pid_cmd = 'source %s&&pssh -t %s -H \"%s\" -P ps -aef | grep %s | grep \"%s/bin/gs_roach --master-ip %s\" | grep -v grep | awk {\'print $2\'}' \
                %(g_opts.sourceFile, pssh_tout, node_ip, g_opts.user, gauss_home_loc, master_ip)

            #####################################################################
            #                  !! IMPORTANT SECURITY NOTE !!                    #
            # Regarding usage of subprocess.Popen([cmd], shell=True, bufsize=-1)#
            #  1.All the inputs from user are captured as individual options    #
            #  2.Direct inputs from the user are not passed in subprocess to    #
            #    gs_roach.                                                      #
            #  3.All the options are validated for the type as well as range    #
            #  4.The commands to be launched through subprocess are formed in   #
            #    code and passed as string to shell                             #
            #####################################################################

            g_logger.debug("python call gs_roach validate in startAgents, cmd: %s" % cmd)
            agent_proc = subprocess.Popen([cmd], stdout=subprocess.PIPE, \
                         stderr=subprocess.PIPE, shell=True, bufsize=-1)
            proc_list_dict.update({node_ip:agent_proc})
            proc_for_pid = subprocess.Popen([pid_cmd], \
                           stdout=subprocess.PIPE, \
                           stderr=subprocess.PIPE, shell=True, bufsize=-1)
            proc_output = proc_for_pid.stdout.readlines()
            if proc_output:
                proc_pid = proc_output[0].strip()
            if proc_pid.isdigit() and (proc_pid != '-1'):
                proc_id_dict.update({node_ip:proc_pid})
    return proc_list_dict

def removeFile(master_generated_file):
    """
    remove intermediate file
    """
    cmd = 'rm -f %s' %(master_generated_file)
    g_logger.debug("deleting file, cmd: %s" % cmd)
    if os.path.isfile(master_generated_file):
        g_logger.debug("deleting file again, cmd: %s" % cmd)
        output = commands.getstatusoutput(cmd)
        del_file_status = output[0]
        if del_file_status != 0:
            g_logger.log("Warning :while deleting %s" % master_generated_file)

def getParentBkpKey(metaDest, backupKey, command_dict):
    """
    function: to get the parent backup key
    input: media destination, backup key
    output: list of backup keys
    """
    bkpdetails = ""

    # inner function
    def searchBackupKey(bkpkey):
        """
        function: recursive fn to find the backup key in the json
        input: backup key
        output: list of backup keys
        """
        for bkpdict in bkpdetails:
            if bkpdict.has_key("BackupKey") and bkpdict["BackupKey"] == bkpkey:
                ## Check if metadata-destination exist for
                ## the current backup key. Checked only for media type "DISK".
                if bkpdict["MediaType"] == "DISK" and bkpdict["MediaDestination"]:
                    metapath = os.path.join(metaDest, "roach", bkpkey)
                    if not os.path.isdir(metapath):
                        errorString = ErrorCode.GAUSS_534["GAUSS_53404"] % bkpkey
                        errorString = "%s in metadata-destination [%s]" % (errorString, metaDest)
                        agentErrorReport("roach_validate", errorString)
                        g_logger.logExit(errorString)

                g_bkpkeylist.append(str(bkpkey))
                if bkpdict.has_key("PriorBackupKey") and (bkpdict["PriorBackupKey"] == "FULL"):
                    return g_bkpkeylist
                else:
                    searchBackupKey(bkpdict["PriorBackupKey"])

    # reading backups.ini in validate operation
    backupfile = os.path.join(metaDest, "backups.ini")
    if os.path.isfile(backupfile):
        chkprior = loadMetaFileAsJson(backupfile)
        if chkprior.has_key("BackupDetails"):
            bkpdetails = chkprior["BackupDetails"]
            for bkpdict in bkpdetails:
                if bkpdict.has_key("BackupKey") and bkpdict["BackupKey"] == backupKey:
                    if bkpdict["BackupType"] == "INCREMENTAL":
                        searchBackupKey(backupKey)
                    else:
                        g_bkpkeylist.append(str(backupKey))
                    break
    else:
        g_logger.logExit("Error: Specified metadata-destination [%s] is corrupted, restore not possible!" % (metaDest))

def resetFinalIPParam(g_opts, command_dict):
    """
    reform the final_ip_param_dict
    """
    final_ip_param_dict.clear()
    if g_opts.roachConf:
        parseRoachConf(g_opts)
    makeRaochOptions(g_opts, master_ip, master_host, command_dict)

def formCommandDict(backupkey, org_backupkey):
    """
    construct the final_ip_param_dict
    """
    global g_mediatype
    global g_mediadest
    global g_backupkey

    for node_ip in final_ip_param_dict.keys():
        metaDest = getMetaDest(final_ip_param_dict)
        inifile = os.path.join(metaDest, "backups.ini")
        inidata = loadMetaFileAsJson(inifile)
        backuplist = inidata["BackupDetails"]
        for bkpitem in backuplist:
            if bkpitem.get("BackupKey") and bkpitem.get("BackupType")\
            and bkpitem["BackupKey"] == backupkey:
                if final_ip_param_dict[node_ip]["--media-type"].upper() not in ["NBU", "DISK", "OBS"]:
                    g_logger.logExit("Error: Invalid media type. Media type should be NBU, DISK or OBS.")
                metaMediaType = str(bkpitem["MediaType"]).strip()
                if backupkey == org_backupkey:
                    metaMediaType = metaMediaType.replace(' with CONTRIB','') #If the type from Meta is with Contrib then ignore that part
                    if final_ip_param_dict[node_ip]["--media-type"].upper() != metaMediaType.upper():
                        g_logger.log("Warning: Provided media type [%s] is incorrect, using correct media type [%s]." \
                            % (final_ip_param_dict[node_ip]["--media-type"], metaMediaType))
                    final_ip_param_dict[node_ip]["--media-type"] = metaMediaType
                    if final_ip_param_dict[node_ip]["--media-type"].upper()[:3] == "NBU":
                        #Validation is not supported for NBU. Special return code so that incremental validate skips this key and continues...
                        return False

                    if os.path.realpath(final_ip_param_dict[node_ip]["--media-destination"]) != os.path.realpath(bkpitem["MediaDestination"]):
                        g_logger.log("Warning: Provided media destination [%s] is incorrect, using correct media destination [%s]." \
                            % (final_ip_param_dict[node_ip]["--media-destination"], bkpitem["MediaDestination"]))
                final_ip_param_dict[node_ip]["--media-destination"] = str(bkpitem["MediaDestination"])
                g_mediatype = metaMediaType
                g_mediadest = str(bkpitem["MediaDestination"])
                g_backupkey = str(backupkey)
    return True

def checkBackupKeyForValidate(backupKey, backupDetails):
    """
    check BackupKey exists for Validate
    """
    bkpExist = True
    for bkpdict in backupDetails:
        if (bkpdict["BackupKey"][:DEF_BKP_KEY_SIZE] == backupKey):
            bkpExist = True
            if (len(bkpdict["PriorBackupKey"][:DEF_BKP_KEY_SIZE]) == DEF_BKP_KEY_SIZE):
                backupKey = bkpdict["PriorBackupKey"][:DEF_BKP_KEY_SIZE]
            break
        else:
            bkpExist = False
    if (not bkpExist):
        errorString = ErrorCode.GAUSS_534["GAUSS_53404"] % backupKey
        agentErrorReport("roach_validate", errorString)
        g_logger.logExit(errorString)

def startValidate(g_opts, command_dict, bkp_key='',validateSingleKey=False):
    """
    call for validate functionality
    """
    global master_ip
    global final_ip_param_dict

    master_ip = getMaster_ip()
    final_ip_param_dict.clear()
    if g_opts.roachConf:
        parseRoachConf(g_opts)
    makeRaochOptions(g_opts, master_ip, master_host, command_dict)
    validate_logparams()

    global g_logger
    g_logger = gsinitLogger(g_opts)
    g_logger.debug("enter startValidate, g_opts.origAction is %s." % g_opts.origAction)

    dropUnwantedParams(g_opts)
    validateRoachParams(g_opts)

    if '--prior-backup-key' in command_dict.keys():
        g_logger.logExit("Error: Invalid Parameter, '--prior-backup-key'")

    global g_bkpkeylist
    global g_isIncremental
    g_isIncremental = False
    g_bkpkeylist = list()
    metaDest = getMetaDest(final_ip_param_dict)
    if not metaDest:
        g_logger.logExit("Error: '--metadata-destination' not provided")

    org_backupkey = ''
    if not bkp_key and "--backup-key" in command_dict.keys():
        bkp_key = command_dict["--backup-key"]
        org_backupkey = bkp_key

    ## make list of backup keys for time-based restore
    if "--restore-target-time" in command_dict.keys():
        arckeysdict = getArchiveBackupKeys(command_dict)
        for item in sorted(arckeysdict.items(),reverse=True):
            g_bkpkeylist.append(str(item[1]))
        if not org_backupkey:
            #In case not directly specified, need to set from --restore-target-time processing
            org_backupkey = g_bkpkeylist[0]
            bkp_key = org_backupkey
    else:
        if not bkp_key:
            g_logger.logExit("Error: Backup key not provided for validate.")
        getParentBkpKey(metaDest, bkp_key, command_dict)
    g_bkpkeylist.reverse()
    backupDetails = getBackupDetails(command_dict, final_ip_param_dict)
    checkBackupKeyForValidate(bkp_key, backupDetails)

    if validateSingleKey :
        g_logger.debug("Validate single key.")
        final_ip_param_dict.clear()
        if g_opts.roachConf:
            parseRoachConf(g_opts)
        makeRaochOptions(g_opts, master_ip, master_host, command_dict)
        dropUnwantedParams(g_opts)
        if not formCommandDict(bkp_key, org_backupkey):
            g_logger.log("Warning: Validate not supported for NBU, skipping backup key: %s" % bkp_key)
            return False
        doValidate(g_opts, command_dict, bkp_key)
        if getRoachOpResult():
            g_logger.log("Error: Validate operation failed for backup key: %s" % bkp_key)
        postExecCleanup(g_opts, g_logger)
        return getRoachOpResult()

    #Added to support proper error message for Logical Multi-table
    if len(g_bkpkeylist) == 0 and bkp_key:
        g_bkpkeylist.append(bkp_key)

    for item in range(len(g_bkpkeylist)):
        ## calling function to modify values for various types of backup keys
        final_ip_param_dict.clear()
        if g_opts.roachConf:
            parseRoachConf(g_opts)
        makeRaochOptions(g_opts, master_ip, master_host, command_dict)
        dropUnwantedParams(g_opts)
        if not formCommandDict(g_bkpkeylist[item], org_backupkey):
            g_logger.log("Warning: Validate not supported for NBU, skipping backup key: %s" % g_bkpkeylist[item])
            continue

        if item != 0:
            print "--------------------------------------------------------------------------------"
            print "Validate operation in-progress for 'INCREMENTAL BACKUP' key[%s/%s]: %s" \
                % (str(item + 1), str(len(g_bkpkeylist)), g_bkpkeylist[item])
            resetFinalIPParam(g_opts, command_dict)
        elif len(g_bkpkeylist) > 1:
            print "\nValidate operation in-progress for 'FULL BACKUP' key[%s/%s]: %s" \
                % (str(item + 1), str(len(g_bkpkeylist)), g_bkpkeylist[item])

        for key in final_ip_param_dict.keys():
            if item == 0:
                final_ip_param_dict[key].update({"--backup-key": str(g_bkpkeylist[item])})
            else:
                g_isIncremental = True
                final_ip_param_dict[key].update({"--backup-key": str(g_bkpkeylist[item])})
                final_ip_param_dict[key].update({"--incremental": None})
        for bkpdict in backupDetails:
            if bkpdict["BackupKey"][:DEF_BKP_KEY_SIZE] == g_bkpkeylist[item][:DEF_BKP_KEY_SIZE]:
                if bkpdict["MediaType"].upper() == "NBU":
                    g_logger.log("Warning: Validate not supported for NBU, skipping backup key: %s" % g_bkpkeylist[item])
                elif len(bkpdict["BackupKey"]) > DEF_BKP_KEY_SIZE:
                    g_logger.logExit("Error: Validate operation not supported for Logical Multi-table backups")
                else:
                    doValidate(g_opts, command_dict, g_bkpkeylist[item])

        if getRoachOpResult():
            g_logger.log("Error: Validate operation failed for backup key: %s" % g_bkpkeylist[item])
            postExecCleanup(g_opts, g_logger)
            return getRoachOpResult()

    postExecCleanup(g_opts, g_logger)
    return getRoachOpResult()
