#!/usr/bin/env python
#-*- coding:utf-8 -*-
try:
    import sys
    import os

    sys.path.append(sys.path[0] + "/../../")
    from gspylib.common.GaussLog import GaussLog
    from gspylib.common.DbClusterInfo import dbClusterInfo
    from gspylib.common.Common import DefaultValue
    from gspylib.common.VersionInfo import VersionInfo
    from gspylib.os.gsOSlib import g_OSlib
    from gspylib.common.ErrorCode import ErrorCode
    from gspylib.component.CM.CM_OLAP.CM_OLAP import CM_OLAP
    from gspylib.component.ETCD.ETCD_OLAP.ETCD_OLAP import ETCD_OLAP
    from gspylib.component.GTM.GTM_OLAP.GTM_OLAP import GTM_OLAP
    from gspylib.component.Kernal.CN_OLAP.CN_OLAP import CN_OLAP
    from gspylib.component.Kernal.DN_OLAP.DN_OLAP import DN_OLAP

except ImportError, e:
    sys.exit("[GAUSS-52200] : Unable to import module: %s." % str(e))

class LocalBaseOM(object):
    """
    Base class for local command
    """
    def __init__(self, logFile = None, user = None, clusterConf = None, dwsMode = False, initParas = [], gtmInitParas = []):
        '''
        Constructor
        '''
        if (logFile is not None):
            self.logger = GaussLog(logFile, self.__class__.__name__)
        else:
            self.logger = None
        self.clusterInfo = None
        self.dbNodeInfo = None
        self.clusterConfig = clusterConf
        self.user = user
        self.group = ""
        self.dws_mode = dwsMode
        self.initParas = initParas
        self.gtmInitParas = gtmInitParas
        self.etcdCons = []
        self.cmCons = []
        self.gtmCons = []
        self.cnCons = []
        self.dnCons = []
        self.gtsCons = []
    
    def initComponent(self):
        """
        function: Init component
        input : NA
        output: NA
        """
        self.initCmComponent()
        self.initGtmComponent()
        self.initEtcdComponent()
        self.initKernelComponent()
    
    def initComponentAttributes(self, component):
        """
        function: Init  component attributes on current node
        input : Object component
        output: NA
        """
        component.logger = self.logger
        component.binPath = "%s/bin" % self.clusterInfo.appPath
        component.dwsMode = self.dws_mode
        
    def initCmComponent(self):
        """
        function: Init cm component on current node
        input : Object nodeInfo
        output: NA
        """
        for inst in self.dbNodeInfo.cmservers:
            component = CM_OLAP()
            #init component cluster type
            component.clusterType = self.clusterInfo.clusterType
            component.instInfo = inst
            self.initComponentAttributes(component)
            self.cmCons.append(component)
                
        for inst in self.dbNodeInfo.cmagents:
            component = CM_OLAP()
            #init component cluster type
            component.clusterType = self.clusterInfo.clusterType
            component.instInfo = inst
            self.initComponentAttributes(component)
            self.cmCons.append(component)
        
    def initEtcdComponent(self):
        """
        function: Init etcd component on current node
        input : Object nodeInfo
        output: NA
        """
        for inst in self.dbNodeInfo.etcds:
            component = ETCD_OLAP()
            #init component cluster type
            component.clusterType = self.clusterInfo.clusterType
            component.instInfo = inst
            self.initComponentAttributes(component)
            self.etcdCons.append(component)
    
    def initGtmComponent(self):
        """
        function: Init gtm component on current node
        input : Object nodeInfo
        output: NA
        """
        for inst in self.dbNodeInfo.gtms:
            component = GTM_OLAP()
            #init component cluster type
            component.clusterType = self.clusterInfo.clusterType
            component.instInfo = inst
            component.initParas = self.gtmInitParas
            self.initComponentAttributes(component)
            self.gtmCons.append(component)
    
    def initKernelComponent(self):
        """
        function: Init kernel component on current node
        input : Object nodeInfo
        output: NA
        """
        for inst in self.dbNodeInfo.coordinators:
            component = CN_OLAP()
            #init component cluster type
            component.clusterType = self.clusterInfo.clusterType
            component.instInfo = inst
            self.initComponentAttributes(component)
            component.initParas = self.initParas
            self.cnCons.append(component)

        for inst in self.dbNodeInfo.datanodes:
            component = DN_OLAP()
            #init component cluster type 
            component.clusterType = self.clusterInfo.clusterType
            component.instInfo = inst
            self.initComponentAttributes(component)
            component.initParas = self.initParas
            self.dnCons.append(component)
    
    def readConfigInfo(self):
        """
        function: Read config from static config file
        input : NA
        output: NA
        """
        try:
            self.clusterInfo = dbClusterInfo()
            self.clusterInfo.initFromStaticConfig(self.user)
            hostName = DefaultValue.GetHostIpOrName()
            self.dbNodeInfo = self.clusterInfo.getDbNodeByName(hostName)
            if (self.dbNodeInfo is None):
                self.logger.logExit(ErrorCode.GAUSS_516["GAUSS_51619"] % hostName)
        except Exception, e:
            self.logger.logExit(str(e))

        self.logger.debug("Instance information on local node:\n%s" % str(self.dbNodeInfo))

    def readConfigInfoByXML(self):
        """
        function: Read config from xml config file
        input : NA
        output: NA
        """
        try:
            if (self.clusterConfig is None):
                self.logger.logExit(ErrorCode.GAUSS_502["GAUSS_50201"] % "XML configuration file")
            static_config_file = "%s/bin/cluster_static_config" % DefaultValue.getInstallDir(self.user)
            self.clusterInfo = dbClusterInfo()
            self.clusterInfo.initFromXml(self.clusterConfig, static_config_file)
            hostName = DefaultValue.GetHostIpOrName()
            self.dbNodeInfo = self.clusterInfo.getDbNodeByName(hostName)
            if (self.dbNodeInfo is None):
                self.logger.logExit(ErrorCode.GAUSS_516["GAUSS_51619"] % hostName)
        except Exception, e:
            self.logger.logExit(str(e))
        self.logger.debug("Instance information on local node:\n%s" % str(self.dbNodeInfo))

    def getUserInfo(self):
        """
        Get user and group
        """
        if os.path.islink(self.clusterInfo.appPath):
            appPath = os.path.realpath(self.clusterInfo.appPath)
        elif os.path.exists(self.clusterInfo.appPath):
            appPath = self.clusterInfo.appPath
        else:
            commitid = VersionInfo.getCommitid()
            appPath = self.clusterInfo.appPath + "_" + commitid
        self.logger.debug("Get the install path %s user info." % appPath)
        (self.user, self.group) = g_OSlib.getPathOwner(appPath)
        if (self.user == "" or self.group == ""):
            self.logger.logExit(ErrorCode.GAUSS_503["GAUSS_50308"])

    def getGTMDict(self, peerGtm, gtmInst, user = None, configItemType = None, alarm_component = None):
        """
        function: Get GTM configuration
        input : peerGtm is empty means that this is a single cluster.
        output: NA
        """ 
        tmpGTMDict = {}
        tmpGTMDict["listen_addresses"] = "'localhost,%s'" % ",".join(gtmInst.listenIps)
        tmpGTMDict["port"] = str(gtmInst.port)
        tmpGTMDict["nodename"] = "'gtm_%s'" % str(gtmInst.instanceId)
        if (configItemType != "ChangeIPUtility"):
            tmpGTMDict["log_directory"] = "'%s/pg_log/gtm'" % (DefaultValue.getUserLogDirWithUser(user))
        if(peerGtm):
            tmpGTMDict["local_host"] = "'%s'" % ",".join(gtmInst.haIps)
            tmpGTMDict["local_port"] = str(gtmInst.haPort)
            if(not self.clusterInfo.isSinglePrimaryMultiStandbyCluster()):
                tmpGTMDict["active_host"] = "'%s'" % ",".join(peerGtm[0].haIps)
                tmpGTMDict["active_port"] = str(peerGtm[0].haPort)
        if (configItemType == "ConfigInstance"):
            tmpGTMDict["alarm_component"] = "'%s'" % alarm_component
        return tmpGTMDict
