#!/usr/bin/env python
#-*- coding:utf-8 -*-
try:
    import os
    import sys
    import time
    import signal
    import copy
    import commands
    import re
    import getpass
    from datetime import datetime, timedelta
    from multiprocessing.dummy import Pool as ThreadPool

    sys.path.append(sys.path[0] + "/../../")
    from gspylib.common.GaussLog import GaussLog
    from gspylib.common.DbClusterInfo import dbClusterInfo
    from gspylib.common.Common import DefaultValue, ClusterCommand, TempfileManagement
    from gspylib.common.DbClusterStatus import DbClusterStatus
    from gspylib.common.OMCommand import OMCommand
    from gspylib.os.gsfile import g_file
    from gspylib.os.gsplatform import g_Platform, findCmdInPath
    from gspylib.threads.SshTool import SshTool
    from gspylib.common.ErrorCode import ErrorCode
    from gspylib.component.CM.CM_OLAP.CM_OLAP import CM_OLAP
    from gspylib.component.ETCD.ETCD_OLAP.ETCD_OLAP import ETCD_OLAP
    from gspylib.component.GTM.GTM_OLAP.GTM_OLAP import GTM_OLAP
    from gspylib.component.Kernal.CN_OLAP.CN_OLAP import CN_OLAP
    from gspylib.component.Kernal.DN_OLAP.DN_OLAP import DN_OLAP

except ImportError, e:
    sys.exit("[GAUSS-52200] : Unable to import module: %s." % str(e))

SPACE_USAGE_DBUSER = 80

class ParallelBaseOM(object):
    """
    Base class of parallel command
    """
    ACTION_INSTALL = "install"
    ACTION_CONFIG = "config"
    ACTION_START = "start"
    ACTION_REDISTRIBUTE = "redistribute"
    ACTION_HEALTHCHECK = "healthcheck"
    
    HEALTH_CHECK_BEFORE = "before"
    HEALTH_CHECK_AFTER = "after"

    DWS_COPY_CERT_LIST = {"gds"  : DefaultValue.GDS_CERT_LIST,
                          "gsql" : DefaultValue.GSQL_SERVER_CERT_LIST}
    """
    Base class for parallel command
    """
    def __init__(self):
        '''
        Constructor
        '''   
        self.logger = None
        self.clusterInfo = None
        self.oldClusterInfo = None
        self.sshTool = None
        self.action = ""

        # Cluster config file.
        self.xmlFile = ""
        self.oldXmlFile = ""

        self.logType = DefaultValue.LOCAL_LOG_FILE
        self.logFile = ""
        self.localLog = ""
        self.user = ""
        self.group = ""
        self.mpprcFile = ""
        #Temporary catalog for install
        self.operateStepDir = TempfileManagement.getTempDir("%s_step" % self.__class__.__name__.lower())
        #Temporary files for install step
        self.operateStepFile = "%s/%s_step.dat" % (self.operateStepDir, self.__class__.__name__.lower())
        self.initStep = ""
        self.dws_mode = False
        self.rollbackCommands = []
        self.etcdCons = []
        self.cmCons = []
        self.gtmCons = []
        self.cnCons = []
        self.dnCons = []
        # localMode is same as isSingle in all OM script, expect for gs_preinstall.
        # in gs_preinstall, localMode means local mode for master-standby cluster.
        # in gs_preinstall, localMode also means local mode for single cluster(will not create os user).
        # in gs_preinstall, isSingle means single cluster, it will create os user.
        # not isSingle and not localMode : master-standby cluster global mode(will create os user).
        # not isSingle and localMode : master-standby cluster local mode(will not create os user).
        # isSingle and not localMode : single cluster(will create os user).
        # isSingle and localMode : single cluster(will not create os user).
        self.localMode = False
        self.isSingle = False
        #Indicates whether there is a logical cluster.
        #If elastic_group exists, the current cluster is a logical cluster.
        #Otherwise, it is a large physical cluster.
        self.isElasticGroup = False
        self.isAddElasticGroup = False
        self.lcGroup_name = ""
        #Lock the cluster mode, there are two modes: exclusive lock and wait lock mode, 
        #the default exclusive lock
        self.lockMode = "exclusiveLock"
                
        # SinglePrimaryMultiStandby support binary upgrade, inplace upgrade
        self.isSinglePrimaryMultiStandby = False

        # Adapt to 200 and 300
        self.productVersion = None

        # show father process thread for tracing back
        self.call_thread = ""



    def initComponent(self):
        """
        function: Init component
        input : NA
        output: NA
        """
        for nodeInfo in self.clusterInfo.dbNodes:
            self.initCmComponent(nodeInfo)
            self.initGtmComponent(nodeInfo)
            self.initEtcdComponent(nodeInfo)
            self.initKernelComponent(nodeInfo)
    
    def initComponentAttributes(self, component):
        """
        function: Init  component attributes on current node
        input : Object component
        output: NA
        """
        component.logger = self.logger
        component.binPath = "%s/bin" % self.clusterInfo.appPath
        component.dwsMode = self.dws_mode
    
    def initCmComponent(self, nodeInfo):
        """
        function: Init cm component
        input : Object nodeInfo
        output: NA
        """
        for inst in nodeInfo.cmservers:
            component = CM_OLAP()
            #init component cluster type
            component.clusterType = self.clusterInfo.clusterType
            component.instInfo = inst
            self.initComponentAttributes(component)
            self.cmCons.append(component)
        for inst in nodeInfo.cmagents:
            component = CM_OLAP()
            #init component cluster type
            component.clusterType = self.clusterInfo.clusterType
            component.instInfo = inst
            self.initComponentAttributes(component)
            self.cmCons.append(component)

    def initEtcdComponent(self, nodeInfo):
        """
        function: Init etcd component
        input : Object nodeInfo
        output: NA
        """
        for inst in nodeInfo.etcds:
            component = ETCD_OLAP()
            #init component cluster type
            component.clusterType = self.clusterInfo.clusterType
            component.instInfo = inst
            self.initComponentAttributes(component)
            self.etcdCons.append(component)
    
    def initGtmComponent(self, nodeInfo):
        """
        function: Init gtm component
        input : Object nodeInfo
        output: NA
        """
        for inst in nodeInfo.gtms:
            component = GTM_OLAP()
            #init component cluster type
            component.clusterType = self.clusterInfo.clusterType
            component.instInfo = inst
            self.initComponentAttributes(component)
            self.gtmCons.append(component)
    
    def initKernelComponent(self, nodeInfo):
        """
        function: Init kernel component
        input : Object nodeInfo
        output: NA
        """
        for inst in nodeInfo.coordinators:
            component = CN_OLAP()
            #init component cluster type
            component.clusterType = self.clusterInfo.clusterType
            component.instInfo = inst
            self.initComponentAttributes(component)
            self.cnCons.append(component)

        for inst in nodeInfo.datanodes:
            component = DN_OLAP()
            #init component cluster type
            component.clusterType = self.clusterInfo.clusterType
            component.instInfo = inst
            self.initComponentAttributes(component)
            self.dnCons.append(component)

    def initLogger(self, module=""):
        """
        function: Init logger
        input : module
        output: NA
        """
        # log level
        LOG_DEBUG = 1
        self.logger = GaussLog(self.logFile, module, LOG_DEBUG)
        
        dirName = os.path.dirname(self.logFile)
        self.localLog = os.path.join(dirName, DefaultValue.LOCAL_LOG_FILE)
        
    def initClusterInfo(self, refreshCN=True):
        """
        function: Init cluster info
        input : NA
        output: NA
        """
        try:
            self.clusterInfo = dbClusterInfo()
            if (refreshCN):
                static_config_file = "%s/bin/cluster_static_config" % DefaultValue.getInstallDir(self.user)
                self.clusterInfo.initFromXml(self.xmlFile, static_config_file)
            else:
                self.clusterInfo.initFromXml(self.xmlFile)
        except Exception, e:
            raise Exception(str(e))
        self.logger.debug("Instance information of cluster:\n%s." % str(self.clusterInfo))
        
    def initClusterInfoFromStaticFile(self, user, flag = True):
        """
        function: Function to init clusterInfo from static file
        input : user
        output: NA
        """
        try:
            self.clusterInfo = dbClusterInfo()
            self.clusterInfo.initFromStaticConfig(user)
        except Exception as e:
            raise Exception(str(e))
        if flag:
            self.logger.debug("Instance information of cluster:\n%s." % str(self.clusterInfo))

    def initSshTool(self, nodeNames, timeout = 0):
        """
        function: Init ssh tool
        input : nodeNames, timeout
        output: NA
        """
        self.sshTool = SshTool(nodeNames, self.logger.logFile, timeout)

    def check_cluster_version_consistency(self, clusterNodes, newNodes=[], paraDict={}):
        """
        """
        self.logger.log("Check cluster version consistency.")
        dic_version_info = {}
        # check version.cfg on every node.
        gp_home = DefaultValue.getEnv("GPHOME")
        gauss_home = DefaultValue.getEnv("GAUSSHOME")
        if "express" not in paraDict:
            if not (os.path.exists(gp_home) and os.path.exists(gauss_home)):
                GaussLog.exitWithError(ErrorCode.GAUSS_502["GAUSS_50201"] % ("%s or %s") % ("GPHOME", "GAUSSHOME"))
        else:
            if not os.path.exists(gp_home):
                GaussLog.exitWithError(ErrorCode.GAUSS_502["GAUSS_50201"] % "GPHOME")
        for ip in clusterNodes:
            if ip in newNodes:
                cmd = "pssh -s -H %s 'cat %s/version.cfg'" % (ip, DefaultValue.getEnv("GPHOME"))
            else:
                cmd = "pssh -s -H %s 'cat %s/bin/upgrade_version'" % (ip, DefaultValue.getEnv("GAUSSHOME"))
            status, output = commands.getstatusoutput(cmd)
            if (status != 0):
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % cmd + " Error:\n%s" % str(output))
            if len(output.strip().split()) < 3:
                raise Exception(ErrorCode.GAUSS_516["GAUSS_51623"])
            dic_version_info[ip] = ",".join(output.strip().split()[1:])

        self.logger.debug("The cluster version on every node.")
        for check_ip, version_info in dic_version_info.items():
            self.logger.debug("%s : %s" % (check_ip, version_info))
        if len(set(dic_version_info.values())) != 1:
            L_inconsistent = list(set(dic_version_info.values()))
            self.logger.debug("The package version on some nodes are inconsistent\n%s" % str(L_inconsistent))
            raise Exception(ErrorCode.GAUSS_516["GAUSS_51652"] + "Version list is: %s" % ", ".join(L_inconsistent))
        self.logger.log("Successfully checked cluster version.")

    def checkBaseFile(self, checkXml = True):
        """
        function: Check xml file and log file
        input : checkXml
        output: NA
        """
        if (checkXml):
            if(self.xmlFile == ""):
                raise Exception(ErrorCode.GAUSS_500["GAUSS_50001"] %'X' + ".")
                
            if (not os.path.exists(self.xmlFile)):
                raise Exception(ErrorCode.GAUSS_502["GAUSS_50201"] % ("configuration file [%s]" % self.xmlFile))
                
            if (not os.path.isabs(self.xmlFile)):
                raise Exception(ErrorCode.GAUSS_502["GAUSS_50213"] % ("configuration file [%s]" % self.xmlFile))
        else:
            self.xmlFile = ""
            
        if (self.logFile == ""):
            self.logFile = DefaultValue.getOMLogPath(self.logType, self.user, "", self.xmlFile)
            
        if (not os.path.isabs(self.logFile)):
            raise Exception(ErrorCode.GAUSS_502["GAUSS_50213"] % "log")
        
    def initSignalHandler(self):
        """
        function: Function to init signal handler
        input : NA
        output: NA
        """
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        signal.signal(signal.SIGQUIT, signal.SIG_IGN)
        signal.signal(signal.SIGTERM, signal.SIG_IGN)
        signal.signal(signal.SIGALRM, signal.SIG_IGN)
        signal.signal(signal.SIGHUP, signal.SIG_IGN)        
        signal.signal(signal.SIGUSR1, signal.SIG_IGN)
        signal.signal(signal.SIGUSR2, signal.SIG_IGN)
        
    def print_signal_stack(self, frame):
        """
        function: Function to print signal stack
        input : frame
        output: NA
        """
        if (self.logger is None):
            return
        try:
            import inspect
            stacks = inspect.getouterframes(frame)
            for curr in range(len(stacks)):
                stack = stacks[curr]
                self.logger.debug("Stack level: %d. File: %s. Function: %s. LineNo: %d." % (curr, stack[1], stack[3], stack[2]))
                self.logger.debug("Code: %s." % (stack[4][0].strip().strip("\n")))
        except Exception as e:
            self.logger.debug("Failed to print signal stack. Error: \n%s" % str(e))

    def raise_handler(self, signal_num, frame):
        """
        function: Function to raise handler
        input : signal_num, frame
        output: NA
        """
        if (self.logger is not None):
            self.logger.debug("Received signal[%d]." % (signal_num))
            self.print_signal_stack(frame)
        raise Exception(ErrorCode.GAUSS_516["GAUSS_51614"] % (signal_num))

    def setupTimeoutHandler(self):
        """
        function: Function to set up time out handler
        input : NA
        output: NA
        """
        signal.signal(signal.SIGALRM, self.timeout_handler)

    def setTimer(self, timeout):
        """
        function: Function to set timer
        input : timeout
        output: NA
        """
        self.logger.debug("Set timer. The timeout: %d." % timeout)
        signal.signal(signal.SIGALRM, self.timeout_handler)
        signal.alarm(timeout)

    def resetTimer(self):
        """
        function: Reset timer
        input : NA
        output: NA
        """
        signal.signal(signal.SIGALRM, signal.SIG_IGN)
        self.logger.debug("Reset timer. Left time: %d." % signal.alarm(0))

    def timeout_handler(self, signal_num, frame):
        """
        function: Received the timeout signal
        input : signal_num, frame
        output: NA
        """
        if (self.logger is not None):
            self.logger.debug("Received the timeout signal: [%d]." % (signal_num))
            self.print_signal_stack(frame)
        raise Timeout("Time out.")
    
    def setAuditSubItems(self, mode=True, coorInst=None):
        """
        function: Set audit sub items by guc tool.
                  Before the redistribution starts, record the current set of audit parameter values to the temporary file,
                  and then close all audit sub items, read the file value after the redistribution is finished, and restore the audit parameters.
        input : mode : True,False 
                     True: Restore audit sub items settings
                     False: Record the current set of audit parameter values, and close audit sub items settings
        output: NA
        """
        #Audit related parameters
        auditItems = ['audit_copy_exec','audit_database_process','audit_dml_state','audit_dml_state_select',\
    'audit_function_exec','audit_grant_revoke','audit_set_parameter','audit_system_object','audit_user_locked','audit_user_violation','audit_login_logout','enable_random_datanode']
        tmpDir = DefaultValue.getTmpDirFromEnv(self.user)
        #Redistribution of audit log records temporary files
        auditItemsRecordFile = "%s/audit_items_set.dat" % tmpDir
        gucStr = ""
        #close audit log parameter
        if (not mode):
            if (not os.path.isfile(auditItemsRecordFile)):
                #Before closing, the parameters of the current settings are recorded to a temporary file
                sql = "select name ,setting from pg_settings where name in ('audit_copy_exec','audit_database_process','audit_dml_state','audit_dml_state_select',\
    'audit_function_exec','audit_grant_revoke','audit_set_parameter','audit_system_object','audit_user_locked','audit_user_violation','audit_login_logout','enable_random_datanode', 'bypass_workload_manager');"
                (status, output) = ClusterCommand.remoteSQLCommand(sql, self.user, coorInst.hostname, coorInst.port, False, DefaultValue.DEFAULT_DB_NAME)
                if (status !=0 ):
                    raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))
                self.logger.debug("Record the audit sub items settings before redistribution:\n %s" % output)
                g_file.createFileInSafeMode(auditItemsRecordFile)
                fp = open(auditItemsRecordFile, "w")
                fp.write(output)
                fp.flush()
                fp.close()
                os.chmod(auditItemsRecordFile, DefaultValue.KEY_FILE_PERMISSION)
                #Send temporary files to all nodes
                self.sshTool.scpFiles(auditItemsRecordFile, tmpDir)
                for item in auditItems:
                    gucStr += " -c \" %s = 0\" " % item
                gucStr +=  " -c \" bypass_workload_manager = on\" "
            else: 
                self.logger.debug("The record file of the audit sub items settings before redistribution already exists, no need to set auditing items.")
                return
            
        #restore the audit log parameter ,Determine whether the file exists
        if (mode):
            if (os.path.isfile(auditItemsRecordFile)):
                #Read the contents of the file
                fp = open(auditItemsRecordFile, 'r')
                auditStrList = fp.readlines()
                fp.close()
                #Spelling the command set string of audit items
                for line in auditStrList:
                    line = line.strip()
                    itemList = line.split('|')
                    #File records 2 elements(parameter, value) per line
                    if (len(itemList) == 2):
                        gucStr += " -c \" %s = %s\" " % (itemList[0], itemList[1])
                    else:
                        continue
            else:
                self.logger.debug("Restore audit log %s file is not exists." % auditItemsRecordFile)
                return
        cmd = "gs_guc reload -Z coordinator -N all -I all %s && \
            gs_guc reload -Z datanode -N all -I all %s" % (gucStr, gucStr)
        self.logger.debug("Command to set GUC parameters: %s" % cmd)
        (status, output) = DefaultValue.retryGetstatusoutput(cmd)
        if (status != 0):
            raise Exception(ErrorCode.GAUSS_500["GAUSS_50007"] % "GUC" + " Command:%s. Error:\n%s" % (cmd , output))
        else:
            if (mode):
                #After restoring the audit parameter settings, delete the temporary file
                cmd = "rm -rf '%s'" % auditItemsRecordFile
                self.sshTool.executeCommand(cmd, "clean the record file of audit sub items settings")

    def switchNodeToStandby(self, nodeName):
        """
        function: Switch all instances on a node to standby
        input : nodeName
        output: NA
        """
        self.logger.log("Switching nodes to be standby.")
        
        dbNode = self.clusterInfo.getDbNodeByName(nodeName)
        if (dbNode is None):
            raise Exception(ErrorCode.GAUSS_516["GAUSS_51636"] % nodeName)

        instList = dbNode.datanodes + dbNode.gtms
        for inst in instList:
            peerInstList = self.clusterInfo.getPeerInstance(inst)
            if (len(peerInstList) != 2 and len(peerInstList) != 1):
                raise Exception(ErrorCode.GAUSS_516["GAUSS_51603"] % inst.datadir)
            
            peerInst = peerInstList[0]
            peerNode = self.clusterInfo.getDbNodeByName(peerInst.hostname)
            if (peerNode is None):
                raise Exception(ErrorCode.GAUSS_516["GAUSS_51603"] % inst.datadir)
            cmd = ClusterCommand.getSwitchOverCmd(self.user, peerNode.id, peerInst.datadir)
            
            self.logger.debug("Switch over command: %s" % cmd)
            (status, output) = commands.getstatusoutput(cmd)
            if (status != 0):
                self.logger.debug(output)
                raise Exception(ErrorCode.GAUSS_530["GAUSS_53007"] % "instance to standby" + " Data directory: %s." % inst.datadir)

        self.logger.log("Successfully switched the node over to the standby node.")

    def stopNode(self, user="", nodeName="", stopForce=False, timeout=DefaultValue.TIMEOUT_CLUSTER_STOP):
        """
        function: Stop all Instances on the node
        input : user, nodeName, stopForce, timeout
        output: NA
        """
        self.logger.log("Stopping node.")
        if(user == ""):
            user = self.user
        nodeId = 0
        if(nodeName != ""):
            clusterInfo = dbClusterInfo()
            clusterInfo.initFromStaticConfig(user)
            dbNode = clusterInfo.getDbNodeByName(nodeName)
            if not dbNode:
                raise Exception(ErrorCode.GAUSS_516["GAUSS_51619"] % nodeName)
            nodeId = dbNode.id
        
        stopMode = "i" if stopForce else "f"
        cmd = ClusterCommand.getStopCmd(user, nodeId, stopMode, timeout)
        (status, output) = commands.getstatusoutput(cmd)
        if (status != 0):
            self.logger.debug("Stopping failed for the first time. Error: \n%s" % output)

            app_path = DefaultValue.getInstallDir(self.user)
            if app_path != "":
                cmd = "ps ux |grep %s/bin |grep -v grep" % app_path
                (status, output) = self.sshTool.getSshStatusOutput(cmd)
                self.logger.debug("Status: %s.\nOutput:\n%s" % (str(status), output))
            
            self.logger.log("Try forcibly stop.")
            cmd = ClusterCommand.getStopCmd(user, nodeId, "i")
            (status, output) = commands.getstatusoutput(cmd)
            if (status != 0):
                self.logger.debug(output)
                raise Exception(ErrorCode.GAUSS_516["GAUSS_51610"] % "the cluster")
        cmd = ClusterCommand.getQueryStatusCmd(user, 0)
        (status, output) = commands.getstatusoutput(cmd)
        if (status != 0):
            self.logger.debug("Cannot check cluster status.")
        else:
            self.logger.debug("The cluster status is %s" % output)
        self.logger.log("Successfully stopped node.")
        
    def startNode(self, user="", nodeName="", doCheck=True, timeout=DefaultValue.TIMEOUT_CLUSTER_START, retry=False):
        """
        function: Start all instance on the node
        input : user, nodeName, doCheck, timeout
        output: NA
        """
        if(user == ""):
            user = self.user
        nodeId = 0
        startType = "cluster"
        if (nodeName != ""):
            startType = "node"
            clusterInfo = dbClusterInfo()
            clusterInfo.initFromStaticConfig(user)
            dbNode = clusterInfo.getDbNodeByName(nodeName)
            if not dbNode:
                raise Exception(ErrorCode.GAUSS_516["GAUSS_51619"] % nodeName)
            nodeId = dbNode.id

        endTime = None
        if (timeout > 0):
            endTime = datetime.now() + timedelta(seconds=timeout)

        self.logger.log("Starting %s." % startType)
        cmd = ClusterCommand.getStartCmd(user, nodeId, timeout)
        (status, output) = commands.getstatusoutput(cmd)
        if status != 0 and retry and \
                output.find("Maybe the cluster is continuely being stopped in the background") >= 0:
            self.stopNode(user, nodeName)
            time.sleep(1)
            self.logger.log("Starting %s." % startType)
            (status, output) = commands.getstatusoutput(cmd)

        if (status != 0):
            self.logger.debug(output)

            app_path = DefaultValue.getInstallDir(self.user)
            if app_path != "":
                cmd = "ps ux |grep %s/bin |grep -v grep" % app_path
                (status, output) = self.sshTool.getSshStatusOutput(cmd)
                self.logger.debug("Status: %s.\nOutput:\n%s" % (str(status), output))

            raise Exception(ErrorCode.GAUSS_516["GAUSS_51607"] % startType + " Error: \n%s" % output)
        
        if (doCheck):
            #The fourth parameter ignore check the status of redistribution
            self.waitForNormal(user, nodeId, endTime, "")


    def waitForNormal(self, user="", nodeId=0, endTime=None, expected_redistributing = "No"):
        """
        function: Wait the node become Normal
        input : user, nodeId, endTime, expected_redistributing
        output: NA
        """
        checkStatus = 0
        checkResult = ""
        startType = "cluster" if nodeId == 0 else "node"
        if (expected_redistributing == "Yes"):
            self.logger.log("Waiting for the redistributing status to become yes.")
        else:
            self.logger.log("Waiting for the %s status to become normal." % startType )
        dotCount = 0
        startWaitTime = datetime.now()
        waitagain = 0
        while True:
            time.sleep(5)
            sys.stdout.write(".")
            dotCount += 1
            if (dotCount >= 12):
                dotCount = 0
                sys.stdout.write("\n")

            checkStatus = 0
            checkResult = ""
            (checkStatus, checkResult) = OMCommand.doCheckStaus(user, nodeId, None, expected_redistributing)
            if (checkStatus == 0):
                if (dotCount != 0): sys.stdout.write("\n")
                if (expected_redistributing == "Yes"):
                    self.logger.log("The cluster redistributing status is yes.")
                else:
                    self.logger.log("The %s status is normal." % startType )
                break
            if (datetime.now() > startWaitTime + timedelta(seconds=DefaultValue.TIMEOUT_CLUSTER_START)):
                MATCH_KEYWORD = "^cluster_state[ ]+: Degraded\n"
                if (expected_redistributing in ("No", "Yes")):
                    MATCH_KEYWORD += "redistributing[ ]+: %s\n" % expected_redistributing
                pattern = re.compile(MATCH_KEYWORD)
                if (pattern.match(checkResult) is None):
                    break
                cmd = "cm_ctl query -Cv | grep Building | grep -v grep"
                (status, output) = DefaultValue.retryGetstatusoutput(cmd)
                if(status == 0):
                    continue
                else:
                    if (waitagain == 0):
                        time.sleep(120)
                        waitagain = 1
                        continue
                    else:
                        break

            if (endTime is not None and datetime.now() >= endTime):
                if (dotCount != 0): sys.stdout.write("\n")
                self.logger.debug(checkResult)

                if (expected_redistributing == "Yes"):
                    raise Exception("Timeout." + "\n" + ErrorCode.GAUSS_516["GAUSS_51637"] % ("redistributing status", "Yes"))
                else:
                    raise Exception("Timeout." + "\n" + ErrorCode.GAUSS_516["GAUSS_51602"].replace("cluster", startType))

        if (checkStatus != 0):
            self.logger.debug(checkResult)
            raise Exception(ErrorCode.GAUSS_516["GAUSS_51607"] % startType + " The %s status is: \n%s" % (startType, checkResult))

    def waitForAvailable(self, user, timeout=DefaultValue.TIMEOUT_CLUSTER_START, isExpandScene=False):
        """
        function: Wait the node become available(Normal or Degraded)
        input : user, timeout, isExpandScene
        output: NA
        """
        self.logger.log("Waiting for the cluster status to become available.")

        endTime = datetime.now() + timedelta(seconds=timeout)
        dotCount = 0

        while True:
            time.sleep(5)
            sys.stdout.write(".")
            dotCount += 1
            if (dotCount >= 12):
                dotCount = 0
                sys.stdout.write("\n")

            if (datetime.now() >= endTime):
                if (dotCount != 0): sys.stdout.write("\n")
                checkStatus = 1
                self.logger.log("Timeout in %d seconds. Cluster is still not available!" % timeout)
                break

            clusterStatus = OMCommand.getClusterStatus(user, isExpandScene)
            if(clusterStatus is None):
                self.logger.debug("Failed to obtain the cluster status.")
            elif(clusterStatus.clusterStatus in ["Normal", "Degraded"]):
                checkStatus = 0
                if (dotCount != 0): sys.stdout.write("\n")
                self.logger.log("The cluster status is %s." % clusterStatus.clusterStatus)
                break
            else:
                self.logger.debug("Cluster status is %s(%s)." % \
                     (clusterStatus.clusterStatus, clusterStatus.clusterStatusDetail))

        if (checkStatus != 0):
            raise Exception(ErrorCode.GAUSS_516["GAUSS_51602"])

    def waitProcessStop(self, processKeywords, hostname):
        """
        function: Wait the process stop
        input : process name 
        output: NA
        """
        count = 0
        while(True):
            psCmd = "ps ux|grep -v grep |awk '{print \$11}'|grep '%s' " % processKeywords.strip()
            (status, output) = self.sshTool.getSshStatusOutput(psCmd, [hostname])
            #Determine whether the process can be found.
            if (status[hostname] != DefaultValue.SUCCESS):
                self.logger.debug("The %s process stopped." % processKeywords)
                break

            count += 1
            if (count%20 == 0):
                self.logger.debug("The %s process exists." % processKeywords)
            time.sleep(3)

    def checkCluseterStatus(self, user, expected_status = None, expected_redistributing = None):
        """
        function: Check cluster status
        input : user, expected_status, expected_redistributing
        output: NA
        """
        self.logger.log("Checking the cluster status.")
        try:
            #query and save status into a file
            statusFile = "/home/%s/gauss_check_status_%d.dat" % (user, os.getpid())
            TempfileManagement.removeTempFile(statusFile)
            cmd = ClusterCommand.getQueryStatusCmd(user, 0, statusFile)
            (status, output) = commands.getstatusoutput(cmd)
            if (status != 0):
                fp = open(statusFile, 'r')
                output = fp.read()
                fp.close()
                TempfileManagement.removeTempFile(statusFile)
                if (output.find("cm_ctl: can't connect to cm_server.") >= 0):
                    raise Exception(ErrorCode.GAUSS_516["GAUSS_51640"])
                raise Exception(ErrorCode.GAUSS_516["GAUSS_51601"] % "cluster"  + " Error: \n%s" % output)
                
            #check cluster status
            clusterStatus = DbClusterStatus()
            clusterStatus.initFromFile(statusFile)
            TempfileManagement.removeTempFile(statusFile)

            self.logger.debug("The cluster status:%s\nredistributing:\
%s\nbalanced:%s" % (clusterStatus.clusterStatus
                , clusterStatus.redistributing
                , clusterStatus.balanced))
            #check if current status match with expected status
            if not clusterStatus.isAllHealthy(expected_status, self.logger):
                raise Exception(ErrorCode.GAUSS_516["GAUSS_51602"] + " Please check it.")
            if (expected_redistributing is not None and clusterStatus.redistributing != expected_redistributing):
                raise Exception(ErrorCode.GAUSS_516["GAUSS_51641"] + " The expected redistributing status should be %s, Please check it." % expected_redistributing)
        except Exception, e:
            raise Exception(str(e))
            
        self.logger.log("Successfully checked the cluster status.")
    
    def managerOperateStepDir(self, action='create', nodes = []):
        """
        function: manager operate step directory 
        input : NA
        output: currentStep
        """
        try:
            #Creating the backup directory
            if( action == "create"):
                cmd = "(if [ ! -d '%s' ];then mkdir -p '%s' -m %s;fi)" % (self.operateStepDir, 
                                                                          self.operateStepDir, 
                                                                          DefaultValue.KEY_DIRECTORY_MODE)
            else:
                cmd = "(if [ -d '%s' ];then rm -rf '%s';fi)" % (self.operateStepDir, self.operateStepDir)
            DefaultValue.execCommandWithMode(cmd, 
                                             "%s temporary directory" % action, 
                                             self.sshTool, 
                                             self.localMode or self.isSingle, 
                                             "", 
                                             nodes)
        except Exception, e:
            raise Exception(str(e))

    def readOperateStep(self):
        """
        function: read operate step signal 
        input : NA
        output: currentStep
        """
        currentStep = self.initStep

        if (not os.path.exists(self.operateStepFile)):
            self.logger.debug("The %s does not exits." % self.operateStepFile)
            return currentStep

        if (not os.path.isfile(self.operateStepFile)):
            self.logger.debug("The %s must be a file." % self.operateStepFile)
            return currentStep

        try:
            #read the step from INSTALL_STEP
            #open file
            fp = open(self.operateStepFile, "r")
            line = fp.readline().strip()
            if(line != None and line != ""):
                currentStep = line
            #close the file
            fp.close()
        except Exception, e:
            if(fp):
                fp.close()
            #failed to read the INSTALL_STEP
            raise Exception(str(e))

        return currentStep
    
    def writeOperateStep(self, stepName, nodes = []):
        """
        function: write operate step signal 
        input : step
        output: NA
        """
        try:
            #write the step into INSTALL_STEP
            #open the INSTALL_STEP
            with open(self.operateStepFile, "w") as g_DB:
                #write the INSTALL_STEP
                g_DB.write(stepName)
                g_DB.write(os.linesep)
                g_DB.flush()
            #change the INSTALL_STEP permissions
            g_file.changeMode(DefaultValue.KEY_FILE_MODE, self.operateStepFile)

            #distribute file to all nodes
            cmd = "mkdir -p -m %s '%s'" % (DefaultValue.KEY_DIRECTORY_MODE, self.operateStepDir)
            DefaultValue.execCommandWithMode(cmd, 
                                             "create backup directory on all nodes", 
                                             self.sshTool, 
                                             self.localMode or self.isSingle, 
                                             "", 
                                             nodes)

            if(not self.localMode and not self.isSingle):
                self.sshTool.scpFiles(self.operateStepFile, self.operateStepDir, nodes)
        except Exception, e:
            #failed to write the step into INSTALL_STEP
            raise Exception(str(e))
    
    def distributeFiles(self):
        """
        function: distribute package to every host
        input : NA
        output: NA
        """
        self.logger.debug("Distributing files.")
        try:
            #get the all nodes
            hosts = self.clusterInfo.getClusterNodeNames()
            if DefaultValue.GetHostIpOrName() not in hosts:
                raise Exception(ErrorCode.GAUSS_516["GAUSS_51619"] % DefaultValue.GetHostIpOrName())
            hosts.remove(DefaultValue.GetHostIpOrName())
            #Send xml file to every host
            DefaultValue.distributeXmlConfFile(self.sshTool, self.xmlFile, hosts, self.mpprcFile)
            agent_conf_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), "../../../../om_agent.conf")
            if os.path.exists(agent_conf_file):
                self.logger.debug("Distribute tmptoken for public cloud scene.")
                hosts = self.sshTool.hostNames[:]
                local_node = DefaultValue.GetHostIpOrName()
                if local_node in hosts:
                    hosts.remove(local_node)
                destDir = os.path.normpath(os.path.dirname(DefaultValue.PUBLIC_CLOUD_TOKEN_FILE))
                cmd = "source /etc/profile 2&>/dev/null;source ~/.bashrc;pssh %s \"if [ -e %s ];then rm -rf %s;fi\"" % (" ".join(["-H %s"%node for node in hosts]), DefaultValue.PUBLIC_CLOUD_TOKEN_FILE, DefaultValue.PUBLIC_CLOUD_TOKEN_FILE)
                cmd += " && pscp %s %s %s" % (" ".join(["-H %s"%node for node in hosts]), DefaultValue.PUBLIC_CLOUD_TOKEN_FILE, destDir)
                self.logger.debug("Commands for distribute tmptoken to remote node: [%s]" % cmd)
                (status, output) = commands.getstatusoutput(cmd)
                if status != 0:
                    raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % str(cmd) + " Error: \n%s" % str(output))
            #Successfully distributed files
            self.logger.debug("Successfully distributed files.")
        except Exception, e:
            #failed to distribute package to every host
            raise Exception(str(e))
            
    def checkPreInstall(self, user, flag, nodes = []):
        """
        function: check if have done preinstall on given nodes
        input : user, nodes
        output: NA
        """
        try:
            cmd = "%s -U %s -t %s" % (OMCommand.getLocalScript("Local_Check_PreInstall"), user, flag)
            DefaultValue.execCommandWithMode(cmd, 
                                             "check preinstall", 
                                             self.sshTool, 
                                             self.localMode or self.isSingle, 
                                             "", 
                                             nodes)
        except Exception, e:
            raise Exception(str(e))
        
    def checkNodeInstall(self, nodes = [], checkParams = [], strictUserCheck = True):
        """
        function: Check node install
        input : nodes, checkParams, strictUserCheck
        output: NA
        """
        validParam = ["shared_buffers", "max_connections"]    
        cooGucParam = ""
        for param in checkParams:
            entry = param.split("=")
            if (len(entry) != 2):
                raise Exception(ErrorCode.GAUSS_500["GAUSS_50009"])
            if (entry[0].strip() in validParam):
                cooGucParam += " -C \\\"%s\\\"" % param
        self.logger.log("Checking installation environment on all nodes.")
        cmd = "%s -U %s:%s -R %s %s -l %s -X '%s'" % (OMCommand.getLocalScript("Local_Check_Install"), 
                                                      self.user, 
                                                      self.group, 
                                                      self.clusterInfo.appPath, 
                                                      cooGucParam, 
                                                      self.localLog, 
                                                      self.xmlFile)
        if (not strictUserCheck):
            cmd += " -O"
        self.logger.debug("Checking the install command: %s." % cmd)
        DefaultValue.execCommandWithMode(cmd, 
                                         "check installation environment", 
                                         self.sshTool, 
                                         self.localMode or self.isSingle, 
                                         "", 
                                         nodes)

    def installApp(self, nodes = [], static_config_file = "", dws_mode = False, action = "install_cluster"):
        """
        function: Install the applications
        input : nodes, static_config_file, dws_mode
        output: NA
        """
        if self.productVersion is None:
            # Check the product version.
            featureFile = "%s/bin/%s" % (self.clusterInfo.appPath, DefaultValue.DEFAULT_DISABLED_FEATURE_FILE_NAME)
            if os.path.isfile(featureFile):
                self.productVersion, disabledFeatureList = DefaultValue.unencryptedFeatureInfo(featureFile)

        self.logger.log("Installing applications on all nodes.")
        cmd = "%s -U %s:%s -X '%s' -R %s -c %s -l %s -T -t %s" % (OMCommand.getLocalScript("Local_Install"),
                                                                        self.user,
                                                                        self.group,
                                                                        self.xmlFile,
                                                                        self.clusterInfo.appPath, 
                                                                        self.clusterInfo.name, 
                                                                        self.localLog, 
                                                                        action)
        if self.productVersion:
            cmd += " --product=%s" % self.productVersion
        if(static_config_file):
            cmd += " -f %s" % static_config_file
        if dws_mode:
            cmd += " --dws-mode"
        self.rollbackCommands.insert(0, ["%s -U %s -R %s -l %s -T" % (OMCommand.getLocalScript("Local_Uninstall"),
                                                                      self.user,
                                                                      self.clusterInfo.appPath,
                                                                      self.localLog),
                                                                      nodes])
        self.logger.debug("Install applications command: %s." % cmd)
        DefaultValue.execCommandWithMode(cmd, 
                                         "install application on all nodes", 
                                         self.sshTool, 
                                         self.localMode or self.isSingle, 
                                         "", 
                                         nodes)
    
    def uninstallApp(self, nodes = []): 
        """
        function: Uninstall the applications
        input : nodes
        output: NA
        """
        self.logger.log("Uninstalling applications on all nodes.")
        cmd = "%s -U %s -R %s -l %s -T" % (OMCommand.getLocalScript("Local_Uninstall"), 
                                           self.user, 
                                           self.clusterInfo.appPath, 
                                           self.localLog)
        self.logger.debug("Uninstall applications command: %s." % cmd)
        DefaultValue.execCommandWithMode(cmd, 
                                         "uninstall application on all nodes", 
                                         self.sshTool, 
                                         self.localMode or self.isSingle, 
                                         "", 
                                         nodes)
        self.logger.log("Successfully uninstalled applications on all nodes.")
    
    def cleanNodeConfig(self, nodes = [], datadirs = []):
        """
        function: Clean instance
        input : nodes, datadirs
        output: NA
        """
        self.logger.log("Deleting instances from all nodes.")
        if (self.clusterInfo.isMasterStandbyCluster()):
            #clean building standby dn process
            cmd = "%s --clean_install_process -U %s -l %s" % (OMCommand.getLocalScript("Local_Query")
                                                              , self.user
                                                              , self.localLog)
            DefaultValue.execCommandWithMode(cmd, 
                                             "clean process", 
                                             self.sshTool, 
                                             self.localMode or self.isSingle, 
                                             "", 
                                             nodes)
        cmdParam = ""
        for datadir in datadirs:
            cmdParam += " -D %s " % datadir
        cmd = "%s -U %s %s -l %s" % (OMCommand.getLocalScript("Local_Clean_Instance"), 
                                     self.user, 
                                     cmdParam, 
                                     self.localLog)
        DefaultValue.execCommandWithMode(cmd, 
                                         "clean instance", 
                                         self.sshTool, 
                                         self.localMode or self.isSingle, 
                                         "", 
                                         nodes)
        self.logger.log("Successfully deleted instances from all nodes.")
            
    def checkNodeConfig(self, nodes = [], cooGucParam = [], dataGucParam= []):
        """
        function: Check node config on all nodes
        input : nodes, cooGucParam, dataGucParam
        output: NA
        """
        self.logger.log("Checking node configuration on all nodes.")
        cmdParam = ""
        for param in cooGucParam:
            cmdParam += " -C \\\"%s\\\"" % param
        for param in dataGucParam:
            cmdParam += " -D \\\"%s\\\"" % param
        cmd = "%s -U %s -l %s %s" % (OMCommand.getLocalScript("Local_Check_Config"), 
                                     self.user, 
                                     self.localLog, 
                                     cmdParam)
        DefaultValue.execCommandWithMode(cmd, 
                                         "check node configuration", 
                                         self.sshTool, 
                                         self.localMode or self.isSingle, 
                                         "", 
                                         nodes)

    def cleanEtcdData(self, nodes):
        """
        function: Clean instance info in etcd
        input : nodes
        output: NA
        """
        self.logger.log("Deleting new node info in etcd.")
        # get etcd ca.crt path
        gaussHome = DefaultValue.getInstallDir(self.user)
        crtPath = "%s/share/sslcert/etcd/etcdca.crt" % gaussHome
        clientkeyPath = "%s/share/sslcert/etcd/client.key" % gaussHome
        clientcrtPath = "%s/share/sslcert/etcd/client.crt" % gaussHome
        cmd_pre = "export ETCDCTL_API=3;etcdctl --cacert %s --cert %s --key %s --endpoints " % (
        crtPath, clientcrtPath, clientkeyPath)
        for node in self.oldClusterInfo.dbNodes:
            if len(node.etcds) > 0:
                etcdinst = node.etcds[0]
                cmd_pre += "https://%s:%s " % (etcdinst.listenIps[0], etcdinst.port)
                break

        get_cmd = "%s get / --prefix" % cmd_pre
        (status, output) = commands.getstatusoutput(get_cmd)
        if status != 0:
            self.logger.log("Warning: Failed to get etcd data:\n%s" % output)
        else:
            self.logger.debug("Old etcd data:\n%s" % output)

        for nodeName in nodes:
            nodeInfo = self.clusterInfo.getDbNodeByName(nodeName)
            self.logger.debug("Clean etcd node data.")
            cmd = "%s del /%s/command/%s/command_status " % (cmd_pre, self.user, nodeInfo.id)
            cmd += " && %s del /%s/command/%s/pending_command " % (cmd_pre, self.user, nodeInfo.id)
            self.logger.debug("Clean node: %s" % cmd)
            DefaultValue.execCommandWithMode(cmd,
                                             "clean etcd data",
                                             self.sshTool,
                                             True,
                                             "",
                                             nodes)
            self.logger.debug("Clean etcd instance data.")
            instance_list = nodeInfo.coordinators + nodeInfo.datanodes + nodeInfo.gtms + nodeInfo.cmservers
            for inst in instance_list:
                if inst.instanceRole == DefaultValue.INSTANCE_ROLE_COODINATOR:
                    cmd = "%s del /%s/dynamic_config/coordinators/%s " % (cmd_pre, self.user, inst.instanceId)
                elif inst.instanceRole == DefaultValue.INSTANCE_ROLE_DATANODE:
                    cmd = "%s del /%s/dynamic_config/datanodes/%s " % (cmd_pre, self.user, inst.instanceId)
                elif inst.instanceRole == DefaultValue.INSTANCE_ROLE_GTM:
                    cmd = "%s del /%s/dynamic_config/GTM/%s " % (cmd_pre, self.user, inst.instanceId)
                elif inst.instanceRole == DefaultValue.INSTANCE_ROLE_CMSERVER:
                    cmd = "%s del /%s/CMServer/heart_beat/%s " % (cmd_pre, self.user, nodeInfo.id)
                self.logger.debug("Clean instance: %s" % cmd)
                DefaultValue.execCommandWithMode(cmd,
                                                 "clean etcd data",
                                                 self.sshTool,
                                                 True,
                                                 "",
                                                 nodes)
        (status, output) = commands.getstatusoutput(get_cmd)
        if status != 0:
            self.logger.log("Warning: Failed to get before etcd data:\n%s" % output)
        else:
            self.logger.debug("New etcd data:\n%s" % output)
        self.logger.log("Successfully deleted etcd data.")

    def initNodeInstance(self, nodes = [], dbInitParam = [], gtmInitParam = [], dws_mode = False, vc_mode = False):
        """
        function: Init instances on all nodes
        input : nodes, dbInitParam, gtmInitParam, dws_mode
        output: NA
        """
        self.logger.log("Initializing instances on all nodes.")
        cmdParam = ""
        for param in dbInitParam:
            cmdParam += " -P \\\"%s\\\"" % param
        for param in gtmInitParam:
            cmdParam += " -G \\\"%s\\\"" % param
        cmd = "%s -U %s %s -l %s" % (OMCommand.getLocalScript("Local_Init_Instance"),
                                     self.user, 
                                     cmdParam, 
                                     self.localLog)
        if dws_mode:
            cmd += " --dws_mode"
        if vc_mode:
            cmd += " --vc_mode"
        self.rollbackCommands.insert(0, [self.cleanNodeConfig, (nodes,)])
        DefaultValue.execCommandWithMode(cmd, 
                                         "initialize instances", 
                                         self.sshTool, 
                                         self.localMode or self.isSingle, 
                                         "", 
                                         nodes)

    def configNodeInstance(self, nodes = [], cooGucParam = [], dataGucParam = [], cmsGucParam = [], alarm_component = "", vc_mode = False):
        """
        function: Update instances config on all nodes
        input : nodes, cooGucParam, dataGucParam, cmsGucParam, alarm_component
        output: NA
        """
        self.logger.log("Updating instances configuration on all nodes.")
        cmdParam = ""
        # *==SYMBOL==* is a symbol which is used to split the parameter string.
        for param in cooGucParam:
            cmdParam += "*==SYMBOL==*-C*==SYMBOL==*%s" % param
        for param in dataGucParam:
            cmdParam += "*==SYMBOL==*-D*==SYMBOL==*%s" % param
        for param in cmsGucParam:
            cmdParam += "*==SYMBOL==*-S*==SYMBOL==*%s" % param
        if(alarm_component != ""):
            cmdParam += "*==SYMBOL==*--alarm=%s" % alarm_component
        cmd = "%s " % (OMCommand.getLocalScript("Local_Config_Instance"))
        paraLine = "*==SYMBOL==*-U*==SYMBOL==*%s%s*==SYMBOL==*-l*==SYMBOL==*%s" % (self.user, cmdParam, self.localLog)
        if vc_mode:
            paraLine += "*==SYMBOL==*--vc_mode"
        cmd += DefaultValue.encodeParaline(paraLine, DefaultValue.BASE_ENCODE)
        DefaultValue.execCommandWithMode(cmd, 
                                         "update instances configuration", 
                                         self.sshTool, 
                                         self.localMode or self.isSingle, 
                                         "", 
                                         nodes)

    def configHba(self, nodes = [], addIPs = []):
        """
        function: Configure pg_hba on all nodes
        input : nodes, addIPs
        output: NA
        """
        self.logger.log("Configuring pg_hba on all nodes.")
        cmd = "%s -U %s -l %s" % (OMCommand.getLocalScript("Local_Config_Hba"), 
                                  self.user, 
                                  self.localLog)
        if(addIPs):
            cmd += " --add-ip=%s" % ','.join(addIPs)

        self.logger.debug("Command for configuring pg_hba: %s." % cmd)
        DefaultValue.execCommandWithMode(cmd, 
                                         "update instances configuration", 
                                         self.sshTool, 
                                         self.localMode or self.isSingle, 
                                         "", 
                                         nodes)
        self.logger.log("Successfully configured pg_hba on all nodes.")

    def installKerberos(self, nodes = []):
        """
        function: Configure pg_hba on all nodes
        input : nodes
        output: NA
        """
        self.logger.debug("Installing kerberos on all nodes.")
        #ELK and MPPDB enable install kerberos
        userProfile = DefaultValue.getMpprcFile()
        if (DefaultValue.checkKerberos(userProfile)):
            if(not self.localMode and not self.isSingle):
                # distribute keytab and config
                gaussHome = DefaultValue.getInstallDir(self.user)
                gausshome_kerberso = os.path.join(gaussHome, "kerberos")
                if not os.path.exists(gausshome_kerberso):
                    self.g_logger.log("There is not exists kerberos configuration file in %s." % gausshome_kerberso)
                    raise Exception(ErrorCode.GAUSS_502["GAUSS_50201"] % gausshome_kerberso)

                dest_kerberos_dir = gaussHome + '/'
                self.sshTool.scpFiles(gausshome_kerberso, dest_kerberos_dir, self.sshTool.hostNames)

                # create auth_config
                mppdb_site_dir = os.path.join(os.path.dirname(DefaultValue.getMpprcFile()), "auth_config")
                cmd = "if [ ! -d '%s' ]; then mkdir %s; fi" % (mppdb_site_dir, mppdb_site_dir)
                self.sshTool.executeCommand(cmd, "create auth_config directory", DefaultValue.SUCCESS, self.sshTool.hostNames, DefaultValue.getMpprcFile())
                dest_path = os.path.realpath(os.path.join(mppdb_site_dir, "mppdb-site.xml"))
                # distribute mppdb-site.xml
                self.sshTool.scpFiles(dest_path, dest_path, self.sshTool.hostNames)

            self.logger.debug("Installing kerberos on all nodes.")
            cmd = "%s -m %s -U %s --krb-client" %  (OMCommand.getLocalScript("Local_Kerberos"), 
                                       "install", 
                                       self.user)
            DefaultValue.execCommandWithMode(cmd, 
                                             "install kerberos", 
                                             self.sshTool, 
                                             self.localMode or self.isSingle, 
                                             "", 
                                             nodes)
            self.logger.debug("Successfully installed kerberos on all nodes.")

    def excuteRollbackCommands(self):
        """
        support two kinds of rollback command:
        1. rollback commands + execute nodes
        2. rollback function + function args
        """
        self.logger.log("Rolling back.")
        for rollback in self.rollbackCommands:
            if (callable(rollback[0])): 
                func, args = rollback[0], rollback[1]
                try:
                    if (hasattr(func, "func_name")):
                        func_name = func.func_name
                    elif (hasattr(func, "im_func") and hasattr(func.im_func, "func_name")):
                        func_name = func.im_func.func_name
                    else:
                        func_name = str(func)
                except Exception as e:
                    func_name = str(func)
                try:
                    self.logger.log("Rollback function: %s." % func_name)
                    result = func(*args)
                    self.logger.debug("Rollback function: %s. Result: %s." % (func_name, str(result) if result is not None else "finished."))
                except Exception as e:
                    self.logger.error(ErrorCode.GAUSS_516["GAUSS_51626"] + " Failed to execute function[%s]. Error: \n%s" % (func_name, str(e)))
                    sys.exit(3)
            else:   
                cmd, nodes = rollback[0], rollback[1]         
                self.logger.log("Rollback command: %s" % cmd)
                if (nodes is None):
                    (status, output) = self.sshTool.getSshStatusOutput(cmd)
                else:
                    (status, output) = self.sshTool.getSshStatusOutput(cmd, nodes)
                for ret in status.values():
                    if (ret != DefaultValue.SUCCESS):
                        self.logger.error(ErrorCode.GAUSS_516["GAUSS_51626"] + " Failed to execute command[%s]. Result: \n%s" % (cmd, status))
                        sys.exit(3)
                self.logger.debug(output)
        self.logger.log("Rollback succeeded.") 
        
    def lockClusterInternal(self, dbNodes, lockTime, logger, sshTool, action, g_opts = None):
        '''
        function: Lock cluster internal
        input :
            dbNodes:  cluster node info
            lockTime: time out number 
            logger: 
            sshTool: 
            action: the action name which call it. eg "managecn", "replace" 
            g_opts: command option
        output: 
            curNode: Successfully locked  on curNode node
            None: Failed to lock cluster 
        ''' 
        self.logger = logger
        # When executing a lock cluster, multithreading is implemented using daemon threads. 
        extra = "--set-daemon"
        cmd = "%s --lock-cluster %s --lock-time=%d -U %s -l %s --lock-mode %s" % (OMCommand.getLocalScript("Local_Query"), extra, lockTime, g_opts.user, g_opts.localLog, self.lockMode)
        self.logger.debug("Command for locking cluster: %s." % cmd)
        if (self.lockMode == "waitingLock"):
            lockTime = 3600
        lock_success = False
        curNode = ""
        self.logger.debug("All node names: %s." % ",".join([dbNode.name for dbNode in dbNodes]))
            
        for dbNode in dbNodes:
            # when delete CN at local host or the CN is abnormal, then skip lock cluster
            if (action == "managecn"):
                if (g_opts.mode == "delete" and dbNode.name == g_opts.hostname):
                    continue
                if(len(g_opts.abnormalCN) and dbNode.name == g_opts.abnormalCN[0].hostname):
                    continue

            if (len(dbNode.coordinators) > 0):
                curNode = dbNode.name
                agent_conf_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), "../../../../om_agent.conf")
                #when do gs_replace,we should execute gs_clean before lock Cluster
                if (action == "replace"):
                    gs_clean_cmd = "gs_clean -a -p %d -v -r " % dbNode.coordinators[0].port
                    if os.path.exists(agent_conf_file):
                        gs_clean_cmd = "gs_clean -U %s -W %s -a -p %d -v -r " % (DefaultValue.PGUSER, ClusterCommand.get_passwd().replace("-W ", ""), dbNode.coordinators[0].port)
                    self.logger.debug("Command for gs_clean:%s" % gs_clean_cmd.replace(ClusterCommand.get_passwd().replace("-W ", ""), "*****"))
                    sshTool.executeCommand(gs_clean_cmd, "execute gs_clean", DefaultValue.SUCCESS, [curNode])
                #Lock cluster retry 3 times
                lock_success = self.lockClusterImpl(cmd, lockTime, sshTool, curNode, logger, action, g_opts)
                break
        if (not lock_success):
            self.logger.debug("Failed to lock cluster on all nodes.")
            return None
        else:
            self.logger.debug("Successfully locked cluster on node[%s]." % curNode)
            return curNode
    
    def lockClusterImpl(self, cmd, lockTime, sshTool, curNode, logger, action, g_opts):
        '''
        function: Lock cluster internal
        input :
            curNode:  cluster node
            lockTime: time out number 
            logger: 
            sshTool: 
            action: the action name which call it. eg "managecn", "replace" 
            g_opts: command option
        output: 
            lock_success:False/True
        ''' 
        lock_success = False
        #retry 3 times for waiting lock.
        for i in range(3):
            try:
                self.setTimer(lockTime)
                sshTool.executeCommand(cmd, "lock cluster", DefaultValue.SUCCESS, [curNode])
                self.logger.debug("Successfully locked cluster.")
                lock_success = True
                break
            except Exception as lockError: 
                try:
                    self.unlockClusterInternal([curNode], lockTime, logger, sshTool, action, g_opts)
                except Timeout as to:
                    self.logger.debug("Caught timeout error [%s] when unlocking the cluster. Ignore the error." % str(to)) 
                except Exception as unlockError:
                    self.logger.debug("Failed to unlock cluster on node [%s]. Error: \n%s" % (curNode, str(unlockError)))  
                finally:
                    self.logger.debug("Failed to lock cluster on node [%s] at the %d times. Error: \n%s" % (curNode, (i+1), str(lockError)))
        return lock_success
    
    def unlockClusterInternal(self, dbNodeNames, lockTime, logger, sshTool, action, g_opts = None):
        '''
        function: unLock cluster internal
        input :
            dbNodeNames:  cluster node name
            lockTime: time out number 
            logger: 
            sshTool: 
            action: the action name which call it. eg "managecn", "replace" 
            g_opts: command option
        ''' 
        self.logger = logger
        try:
            self.resetTimer()
            cmd = "%s --release-cluster --lock-time=%d -U %s --lock-mode %s -l %s"  % (OMCommand.getLocalScript("Local_Query")
                                                                                   , lockTime
                                                                                   , g_opts.user
                                                                                   , self.lockMode
                                                                                   , g_opts.localLog)
            # now when do gs_replace, we unlock the cluster by root user
            self.logger.debug("Command for unlocking cluster: %s." % cmd)
            sshTool.executeCommand(cmd, "unlock cluster", DefaultValue.SUCCESS, dbNodeNames)
        except Exception as e:
            self.logger.log("Warning: Failed to unlock cluster.")
            self.logger.debug(str(e))
            raise Exception(str(e))

    @staticmethod
    def getPrepareKeysCmd(key_file, user, confFile, destPath, logfile, userProfile="", localMode=False):
        """
        function: get  etcd communication keys command
        input: key_file, user, confFile, destPath, localMode:do not scp keys
        output: NA
        """
        if(not os.path.exists(key_file)):
            raise Exception(ErrorCode.GAUSS_502["GAUSS_50201"] % key_file)
        if(not userProfile):
            userProfile = DefaultValue.getMpprcFile()
        #create the directory on all nodes
        cmd = "source %s; %s -U %s -X %s --src-file=%s --dest-path=%s -l %s" %  (userProfile,
                    OMCommand.getLocalScript("Local_PrepareKeys"),  user, confFile, key_file, destPath, logfile)
        #if local mode, only prepare keys, do not scp keys to cluster nodes
        if(localMode):
            cmd += " -L"
        return cmd

    def getClusterRings(self, clusterInfo):
        """
        function: get clusterRings from cluster info
        input: DbclusterInfo() instance
        output: list
        """
        hostPerNodeList = self.getDNHostnamesPerNode(clusterInfo)
        #Loop the hostname list on each node where the master and slave of the DN instance.
        for i in  range(len(hostPerNodeList)):
            #Loop the list after the i-th list
            for perNodelist in hostPerNodeList[i+1:len(hostPerNodeList)]:
                #Define a tag 
                flag = 0
                #Loop the elements of each perNodelist
                for hostNameElement in perNodelist:
                    #If elements on the i-th node, each element of the list are joined in hostPerNodeList[i
                    if hostNameElement in hostPerNodeList[i]:
                        flag = 1
                        for element in perNodelist:
                            if element not in hostPerNodeList[i]:
                                hostPerNodeList[i].append(element)
                if (flag == 1):
                    hostPerNodeList.remove(perNodelist)

        return hostPerNodeList

    def getDNHostnamesPerNode(self, clusterInfo):
        """
        function: get DN hostnames per node
        input: DbclusterInfo() instance
        output: list
        """
        hostPerNodeList = []
        for dbNode in clusterInfo.dbNodes:
            nodeDnlist = []
            #loop per node
            for dnInst in dbNode.datanodes:
                if (dnInst.instanceType == DefaultValue.MASTER_INSTANCE):
                    if dnInst.hostname not in nodeDnlist:
                        nodeDnlist.append(dnInst.hostname)
                    #get other standby and dummy hostname
                    instances = clusterInfo.getPeerInstance(dnInst)
                    for inst in instances:
                        if inst.hostname not in nodeDnlist:
                            nodeDnlist.append(inst.hostname)
            if nodeDnlist != []:  
                hostPerNodeList.append(nodeDnlist)
        return hostPerNodeList

    #for olap function
    def checkIsElasticGroupExist(self, dbNodes):
        """
        function: Check if elastic_group exists.
        input : NA
        output: NA
        """
        self.logger.debug("Checking if elastic group exists.")

        self.isElasticGroup = False
        coorNode = []
        #traverse old nodes
        for dbNode in dbNodes:
            if (len(dbNode.coordinators) >= 1):
                coorNode.append(dbNode.coordinators[0])
                break

        #check elastic group
        CHECK_GROUP_SQL = "SELECT count(*) FROM pg_catalog.pgxc_group WHERE group_name='elastic_group' and group_kind='e'; "
        (checkstatus, checkoutput) = ClusterCommand.remoteSQLCommand(CHECK_GROUP_SQL, self.user, 
                                  coorNode[0].hostname, coorNode[0].port)
        if (checkstatus != 0 or not checkoutput.isdigit()):
            raise Exception(ErrorCode.GAUSS_502["GAUSS_50219"] % "node group" + " Error:\n%s" % str(checkoutput))
        elif (checkoutput.strip() == '1'):
            self.isElasticGroup = True
        elif (checkoutput.strip() == '0'):
            self.isElasticGroup = False
        else:
            raise Exception(ErrorCode.GAUSS_502["GAUSS_50219"] % "the number of node group")

        self.logger.debug("Successfully checked if elastic group exists.") 

    def checkHostnameIsLoop(self, nodenameList):
        """
        function: check if hostname is looped
        input : NA
        output: NA
        """
        isRing = True
        #1.get ring information in the cluster
        clusterRings = self.getClusterRings(self.clusterInfo)
        nodeRing = ""
        nodenameRings = []
        #2.Check if the node is in the ring
        for num in range(len(clusterRings)):
            ringNodeList = []
            for nodename in nodenameList:
                if (nodename in clusterRings[num]):
                    ringNodeList.append(nodename)
            if (len(ringNodeList) != 0 and len(ringNodeList) == len(clusterRings[num])):
                nodenameRings.append(ringNodeList)
            if (len(ringNodeList) != 0 and len(ringNodeList) != len(clusterRings[num])):
                isRing = False
                break
            else:
                continue
        if not isRing:
            raise Exception(ErrorCode.GAUSS_500["GAUSS_50004"] % "h" + " The hostname (%s) specified by the -h parameter must be looped." % nodeRing)
        return (clusterRings, nodenameRings)
        
    def getAlterPasswdTimes(self):
        """
        function: Obtain super user alter password times
        input : NA
        output: NA
        """
        coorNode = []
        #traverse old nodes
        for dbNode in self.clusterInfo.dbNodes:
            if (len(dbNode.coordinators) >= 1):
                coorNode.append(dbNode.coordinators[0])
                break
        
        sql = "select count(1) from pg_auth_history where roloid = (select usesysid from pg_user where usename = '%s');" % (self.user)
        (status, output) = ClusterCommand.remoteSQLCommand(sql, self.user, 
                        coorNode[0].hostname, coorNode[0].port, False, DefaultValue.DEFAULT_DB_NAME)
        if(status != 0 or not output.isdigit()):
            raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))
        numTimes = int(output)
        if (numTimes == 0):
            self.logger.debug("Obtain super user alter password times result: %d." % numTimes)
            raise Exception(ErrorCode.GAUSS_503["GAUSS_50309"] + "from node[%s]. Error:\n%s" % (coorNode[0].hostname, output)) 

        return numTimes

    def getDNinstanceByNodeName(self, hostname, isMaster = True):
        """
        function: Get the DN instance of the node based on the node name.
        input : hostname
                isMaster: get master dn instance
        output: NA
        """
        masterdnInsts = []
        standbydnInsts = []
        #notice
        for dbNode in self.clusterInfo.dbNodes:
            if (dbNode.name == hostname):
                for dbInst in dbNode.datanodes:
                    #get master dn instance
                    if(dbInst.instanceType == DefaultValue.MASTER_INSTANCE):
                        masterdnInsts.append(dbInst)
                    #get standby or dummy dn instance
                    else:
                        standbydnInsts.append(dbInst)

        if(isMaster):
            return masterdnInsts
        else: 
            return standbydnInsts

    def stopDnInstance(self, dnInst):
        """
        function: stop all dn instances on the node.
        input : NA
        output: NA
        """
        dbNode = self.clusterInfo.getDbNodeByName(dnInst.hostname)
        if not dbNode:
            raise Exception(ErrorCode.GAUSS_516["GAUSS_51619"] % dnInst.hostname)
        nodeId = dbNode.id
        cmd = ClusterCommand.getStopCmd(self.user, nodeId, "i", 0, dnInst.datadir)

        self.logger.debug("Command for stopping node: %s" % cmd)
        (status, output) = commands.getstatusoutput(cmd)
        if (status != 0):
            raise Exception(ErrorCode.GAUSS_516["GAUSS_51610"] % "dn instances" +" Error: \n%s" % output)

    def stopMasterDnInstanceOnNode(self, hostname):
        """
        function: stop all master dn instances on the node.
        input : hostname
        output: NA
        """
        #Get the master DN instance info of the node
        masterdnInstances = self.getDNinstanceByNodeName(hostname, True)
        if (len(masterdnInstances) > 0):
            try:
                pool = ThreadPool(DefaultValue.getCpuSet())
                pool.map(self.stopDnInstance, masterdnInstances)
                pool.close()
                pool.join()
            except Exception, e:
                raise Exception(str(e))

    def stopStandbyDnInstanceOnNode(self, hostname):
        """
        function: stop all standby dn instances on the node.
        input : hostname
        output: NA
        """
        #Get the standby DN instance info of the node
        standbydnInstances = self.getDNinstanceByNodeName(hostname, False)
        if (len(standbydnInstances) > 0):
            try:
                pool = ThreadPool(DefaultValue.getCpuSet())
                pool.map(self.stopDnInstance, standbydnInstances)
                pool.close()
                pool.join()
            except Exception, e:
                raise Exception(str(e))

    def stopDnInstanceInParallel(self, hostList, isMaster = True):
        """
        function: Instances of DNs on nodes that are stopped in parallel between nodes
        input: hostList
        output: NA 
        """
        self.logger.debug("Stopping dn instances in parallel.")
        if (len(hostList) > 0):
            try:
                pool = ThreadPool(DefaultValue.getCpuSet())
                if (isMaster):
                    pool.map(self.stopMasterDnInstanceOnNode, hostList)
                else:
                    pool.map(self.stopStandbyDnInstanceOnNode, hostList)
                pool.close()
                pool.join()
            except Exception, e:
                raise Exception(str(e))
        self.logger.debug("Successfully stopped dn instances in parallel.")

    def startDnInstance(self, dnInstances):
        """
        function: stop dn instance.
        input : NA
        output: NA
        """
        self.logger.debug("Starting dn instances.")
        for inst in dnInstances:
            dbNode = self.clusterInfo.getDbNodeByName(inst.hostname)
            if not dbNode:
                raise Exception(ErrorCode.GAUSS_516["GAUSS_51619"] % inst.hostname)
            nodeId = dbNode.id
            cmd = ClusterCommand.getStartCmd(self.user, nodeId, 0, inst.datadir)

            self.logger.debug("Command for Starting node: %s" % cmd)
            (status, output) = commands.getstatusoutput(cmd)
            if (status != 0):
                raise Exception(ErrorCode.GAUSS_516["GAUSS_51610"] % "dn instances" +" Error: \n%s" % output)

        self.logger.debug("Successfully started dn instances.")

    #for olap function
    def dumpAndRestoreNewNodes(self, needDumpCn, needDumpDn, cooInst=None, lcname = "", lcNodeStr = "", isElasticMode = False):
        """
        function: Dump and restore coordinators and datanodes on new nodes
        input : NA
        output: NA 
        """
        coorNodes = []
        dnNodes = []
        cnInsts = []
        dnSoFile = ""
        hosts = []

        #Find the first CN information in the physical cluster.
        #Consider the following scenario:
        #    In virtual cluster mode, self.oldNodes is the node information of the virtual cluster, but it may not contain CN instances.
        #    For example: install a 6-node cluster and create VCs using 1 2 3 nodes that do not contain CN instances.
        cnInsts.append(cooInst)
        coorNodes.append(cooInst.hostname)
        #traverse old nodes
        #In physical cluster and virtual cluster mode, self.oldNodes is different.
        #    physical cluster mode : self.oldNodes contains all nodes before expansion.
        #    virtual cluster mode  : self.oldNodes contains the nodes of the specified virtual cluster.
        #traverse old nodes
        for dbNode in self.oldNodes:
            hosts.append(dbNode.name)
            #Gets the node name of a DN master and standby instance for the master/standby switch.
            if (len(dbNode.datanodes) >= 1 and len(dnNodes) == 0):
                for dnInst in dbNode.datanodes:
                    if (dnInst.instanceType == DefaultValue.MASTER_INSTANCE):
                        dnNodes.append(dnInst.hostname)
                        peerInsts = self.clusterInfo.getPeerInstance(dnInst)
                        for peerInst in peerInsts:
                            if (peerInst.instanceType == DefaultValue.STANDBY_INSTANCE):
                                dnNodes.append(peerInst.hostname)
                        break
        self.logger.debug("CNs: %s\n DNs: %s" % (coorNodes, dnNodes))

        #Obtain super user alter password times
        alterTimes = self.getAlterPasswdTimes()
        #Get the temporary directory from PGHOST
        tmpDir = DefaultValue.getTmpDirFromEnv()
        #check if tmp dir exists
        if(not os.path.exists(tmpDir)):
            raise Exception(ErrorCode.GAUSS_502["GAUSS_50201"] % tmpDir + " Please check it.")
        #get the dump schema_coordinator file
        coorSqlFile = "%s/%s" % (tmpDir, DefaultValue.SCHEMA_COORDINATOR)
        #get the dump schema_datanode file
        dnSqlFile = "%s/%s" % (tmpDir, DefaultValue.SCHEMA_DATANODE)

        dnTableFile = "%s/%s" % (tmpDir, DefaultValue.DUMP_TABLES_DATANODE)

        dnTableSqlFile = "%s/%s" % (tmpDir, DefaultValue.DUMP_Output_DATANODE)
        #get the dump schema_coordinator file prefix
        coorSqlFilePre = coorSqlFile[:-4]
        #get the dump schema_datanode file prefix
        dnSqlFilePre = dnSqlFile[:-4]

        dnTableFilePre = dnTableFile[:-4]

        dnTableSqlFilePre = dnTableSqlFile[:-4]

        #get .so file of the C function path
        soFilePath = "%s/lib/postgresql/pg_plugin/" % self.clusterInfo.appPath
        try:
            self.logger.log("Rebuilding new nodes.")
            if (needDumpCn):
                self.logger.log("Dumping CN.")
                cmd = "%s --dump-cn -U %s -l %s" % (OMCommand.getLocalScript("Local_Query"), self.user, self.localLog)
                if(alterTimes > 1):
                    cmd += " -a"
                self.logger.debug("Command for dumpping CN: %s" % cmd)
                currNode = coorNodes[0]
                try:                
                    self.sshTool.executeCommand(cmd, "dump CN", DefaultValue.SUCCESS, [currNode])
                except Timeout:
                    raise Exception(ErrorCode.GAUSS_526["GAUSS_52615"])
                except Exception as e:
                    self.logger.debug("Failed to dump CN on node[%s]. Error:\n%s" % (currNode, str(e))) 
                    raise Exception(ErrorCode.GAUSS_516["GAUSS_51616"] % 'CN')
                
                #remotely copy .so file of the C function for CN instance.
                cnSoFile = "%scn_%s#*" % (soFilePath, cnInsts[0].instanceId)
                scpCmd_soFile = g_Platform.getRemoteCopyCmd(cnSoFile, soFilePath, currNode, False, "directory")
                #remotely copy CN file
                scpCmd_cnFile = g_Platform.getRemoteCopyCmd("'%s'*" % coorSqlFilePre, tmpDir, currNode, False, "file")
                scpCmd = scpCmd_soFile + ';' + scpCmd_cnFile
                self.logger.debug("Command for remotely copying CN: %s" % scpCmd)
                self.sshTool.executeCommand(scpCmd, "remotely copy CN file", DefaultValue.SUCCESS, self.newNodeNames)

                #changing the owner of CN
                chownCmd = g_Platform.getChownCmd(self.user, self.group, coorSqlFile)
                chmodCmd = g_Platform.getChmodCmd(str(DefaultValue.FILE_MODE), coorSqlFile, False)
                cmd = chownCmd + ';' + chmodCmd
                self.logger.debug("Command for changing the owner of CN: %s" % cmd)
                self.sshTool.executeCommand(cmd, "change the owner of CN SQL file", DefaultValue.SUCCESS, self.newNodeNames)
                self.logger.log("Successfully dumped CN.")

            if (needDumpDn):
                cmdcn = "%s --dump-dn-tables -U %s -l %s" % (OMCommand.getLocalScript("Local_Query"), self.user, self.localLog)
                if (lcname != "" and lcNodeStr != ""):
                    cmdcn += " --lcname='%s' --lcnodestring='%s'" % (lcname, lcNodeStr)
                self.logger.debug("Command for get dump DN tables file: %s" % cmdcn)
                try:
                    self.sshTool.executeCommand(cmdcn, "get dump DN tables file", DefaultValue.SUCCESS, [coorNodes[0]])
                except Exception as e:
                    self.logger.debug("Failed to get dump DN tables file on node[%s]." % coorNodes[0] ) 
                    raise Exception(str(e))

                self.logger.log("Dumping DN.")
                dumpDnSuccess = False
                cmd = "%s --dump-dn -U %s -l %s" % (OMCommand.getLocalScript("Local_Query"), self.user, self.localLog)
                if (lcname != "" and lcNodeStr != ""):
                    cmd += " --lcname='%s' --lcnodestring='%s'" % (lcname, lcNodeStr)
                self.logger.debug("Command for dumping DN: %s" % cmd)
                errorInfo = ""
                if(alterTimes > 1):
                    cmd += " -a"
                for currNode in dnNodes:
                    try:
                        #dump DN
                        self.sshTool.executeCommand(cmd, "dump DN", DefaultValue.SUCCESS, [currNode])
                        dumpDnSuccess = True
                        break
                    except Timeout:
                        raise Exception(ErrorCode.GAUSS_526["GAUSS_52615"])
                    except Exception as e:
                        errorInfo += "Failed to dump DN on node[%s]. Error:\n%s" % (currNode, str(e))

                if (dumpDnSuccess):
                    #get dump successful DN number to a temporary file[]
                    sourceFile = DefaultValue.getMpprcFile()
                    cmd = g_file.SHELL_CMD_DICT["sshCmd"] % (currNode, sourceFile, "cat %sDnId" % dnSqlFilePre)
                    (status, output) = DefaultValue.retryGetstatusoutput(cmd)
                    if (status != 0 or output == ""):
                        raise Exception("Failed to get dump DN instance id. Error:%s." % output)
                    dumpDnInstId = output.split()[-1].strip()
                    #remotely copy .so file of the C function for CN instance.
                    dnSoFile = "%sdn*_%s*" % (soFilePath, dumpDnInstId)
                    scpCmd_soFile = g_Platform.getRemoteCopyCmd(dnSoFile, soFilePath, currNode, False, "directory")

                    #remotely copying DN
                    scpCmd_dnFile = g_Platform.getRemoteCopyCmd("'%s'*" % dnSqlFilePre, tmpDir, currNode, False, "file")
                    scpCmd = scpCmd_soFile + ';' + scpCmd_dnFile
                    self.logger.debug("Command for remotely copying DN: %s" % scpCmd)
                    self.sshTool.executeCommand(scpCmd, "remotely copy DN file", DefaultValue.SUCCESS, self.newNodeNames)

                    #remotely copying DN
                    scpCmd = g_Platform.getRemoteCopyCmd("'%s'*" % dnTableSqlFilePre, tmpDir, currNode, False, "file")
                    self.logger.debug("Command for remotely copying DN: %s" % scpCmd)
                    self.sshTool.executeCommand(scpCmd, "remotely copy DN file", DefaultValue.SUCCESS, self.newNodeNames)

                    #changing the owner of DN
                    chownCmd = g_Platform.getChownCmd(self.user, self.group, dnSqlFile)
                    chmodCmd = g_Platform.getChmodCmd(str(DefaultValue.FILE_MODE), dnSqlFile, False)
                    cmd = chownCmd + ';' + chmodCmd
                    self.logger.debug("Command for changing the owner of DN: %s" % cmd) 
                    self.sshTool.executeCommand(cmd, "change the owner of DN SQL file", DefaultValue.SUCCESS, self.newNodeNames)

                    #changing the owner of DN
                    dnTableSqlFilePre = "'%s'*" % dnTableSqlFilePre
                    chownCmd = g_Platform.getChownCmd(self.user, self.group, dnTableSqlFilePre)
                    chmodCmd = g_Platform.getChmodCmd(str(DefaultValue.FILE_MODE), dnTableSqlFilePre, False)
                    cmd = chownCmd + ';' + chmodCmd
                    self.logger.debug("Command for changing the owner of DN: %s" % cmd) 
                    self.sshTool.executeCommand(cmd, "change the owner of DN SQL file", DefaultValue.SUCCESS, self.newNodeNames)

                    self.logger.log("Successfully dumped DN.")
                else:
                    raise Exception(errorInfo)

            #only used by logic cluster mode
            #If the expansion node contains cmserver or gtm, the node cannot be stopped. Otherwise, the restore DN will fail.
            if (lcname != "" and lcNodeStr != ""):
                self.logger.log("Stopping DN instance on node.")
                #Instances of DNs on nodes that are stopped in parallel between nodes
                #stop standby/dummy dn instance
                self.stopDnInstanceInParallel(self.newNodeNames, False)
                ##stop master dn instance
                self.stopDnInstanceInParallel(self.newNodeNames, True)
                self.logger.log("Successfully stopped DN instance on node.")

            self.logger.log("Restoring new nodes.")
            buildInstanceString = ""
            if (needDumpCn and needDumpDn):
                buildInstanceString = "--build-node=all"
            elif (needDumpCn):
                buildInstanceString = "--build-node=coordinators"
            elif (needDumpDn):
                buildInstanceString = "--build-node=datanodes"
            #restore new nodes
            cmd = "%s %s -U %s --parallel-jobs %s -l %s" % (OMCommand.getLocalScript("Local_Query"), buildInstanceString, self.user, self.jobs, self.localLog)
            if (self.dws_mode):
                cmd += " --dws-mode"
            if (lcname != "" and lcNodeStr != ""):
                cmd += " --replace --lcname=%s" % lcname
            if (isElasticMode):
                cmd += " --elastic-cluster "
            self.logger.debug("Command for restoring nodes: %s" % cmd)
            self.sshTool.executeCommand(cmd, "restore nodes", DefaultValue.SUCCESS, self.newNodeNames)
            self.logger.log("Successfully restored new nodes.")
            #remove .so file of the C function on the new nodes.
            rmCmd = g_Platform.getRemoveCmd("directory")
            cmd = "%s %s;" % (rmCmd, dnSoFile)
            if (needDumpCn):
                cmd += "%s %s" % (rmCmd, cnSoFile)
            self.logger.debug("Command for remove .so file: %s" % cmd)
            self.sshTool.executeCommand(cmd, "remove .so file", DefaultValue.SUCCESS, self.newNodeNames)
            self.logger.log("Successfully rebuild new nodes.")
        except Exception as e:
            raise Exception(str(e))
        finally:
            rmCmd = g_Platform.getRemoveCmd("file")
            cmd = ""
            #remove all nodes cn sql file
            cmd += "%s '%s'* ;" % (rmCmd, coorSqlFilePre)
            #remove all nodes dn sql file
            cmd += "%s '%s'* ;" % (rmCmd, dnSqlFilePre)
            #remove all nodes dump table file
            cmd += "%s '%s'* ;" % (rmCmd, dnTableFilePre)
            #remove all nodes dn table sql file
            cmd += "%s %s*" % (rmCmd, dnTableSqlFilePre)
            try:
                #Clean up the gaussdb process that may remain
                killCmd = ""
                for newNodeName in self.newNodeNames:
                    killCmd += " pssh -s -H %s '%s -9 gaussdb;%s -9 gsql' 2>/dev/null;" % (newNodeName,
                                                                                           findCmdInPath("killall"),
                                                                                           findCmdInPath("killall"))
                os.system(killCmd[:-1])
                if (lcname != "" and lcNodeStr != ""):
                    self.sshTool.executeCommand(cmd, "remove all nodes sql file", DefaultValue.SUCCESS, hosts + self.newNodeNames)
                else:
                    self.sshTool.executeCommand(cmd, "remove all nodes sql file", DefaultValue.SUCCESS, self.nodeNames)
                self.logger.debug("Successfully remove all nodes sql file.")   
            except Exception as e:
                self.logger.debug("Failed to remove all nodes sql file. Output:%s." % str(e))

    def getSQLResultList(self, sql, user, hostname, port, database = "postgres"):
        """
        """
        (status, output) = ClusterCommand.remoteSQLCommand(sql, user, hostname, port, False, database)
        if status != 0 or ClusterCommand.findErrorInSql(output):
            raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))
        #split the output string with '\n'
        resultList = output.split("\n")
        return resultList

    def cleanAndReloadSingleGPUCache(self, tmpExecPara):
        """
        """
        database = tmpExecPara[0]
        schema = tmpExecPara[1]
        hostname = tmpExecPara[2]
        port = tmpExecPara[3]
        executeSQL = "set application_name='OM'; SET search_path = \'%s\'; SELECT * FROM pg_catalog.clean_vector_gpu();" \
                     " SELECT * FROM pg_catalog.add_vector_gpu();" % schema
        self.logger.debug("SQL for clean and reload GPU on database %s: %s." % (database, executeSQL))
        (status, output) = ClusterCommand.remoteSQLCommand(executeSQL, self.user, hostname, port, False, database)
        if status != 0 or ClusterCommand.findErrorInSql(output):
            raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % executeSQL + " Error:\n%s" % str(output))

    def cleanAndReloadGPUCache(self, coorNode):
        """
        function: excute sql to clean and reload GPU cache
        input : NA
        output: NA
        """
        # Check if GPU exists
        # GPU is no supported in Gauss300 
        if self.productVersion is None:
            # Check the product version.
            featureFile = "%s/bin/%s" % (self.clusterInfo.appPath, DefaultValue.DEFAULT_DISABLED_FEATURE_FILE_NAME)
            if os.path.isfile(featureFile):
                self.productVersion, disabledFeatureList = DefaultValue.unencryptedFeatureInfo(featureFile)
        if (self.productVersion is not None) and (self.productVersion.lower() == DefaultValue.PRODUCT_VERSION_GAUSS300.lower()):
            return
        sql = "SELECT COUNT(proname) FROM pg_catalog.pg_proc WHERE proname = 'pgxc_get_simsearch_lib_load_status';"
        (status, output) = ClusterCommand.remoteSQLCommand(sql, self.user, coorNode.hostname, coorNode.port, False)
        if (status != 0 or not output.isdigit()):
            raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))
        if (1 != int(output)):
            self.logger.debug("The SQL used to check the GPU cache is %s: The output value is %s." % (sql, output))
            return

        # check if GPU is be used
        sql = "set application_name='OM'; SET search_path = 'public'; SELECT COUNT(lib_load_status) FROM" \
              " pg_catalog.pgxc_get_simsearch_lib_load_status() WHERE lib_load_status = false;"
        (status, output) = ClusterCommand.remoteSQLCommand(sql, self.user, coorNode.hostname, coorNode.port, False)
        output = output.replace("SET\n" ,'')
        if (status != 0 or not output.isdigit()):
            raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))
        if (int(output) > 0):
            self.logger.debug("The SQL used to check the GPU cache is %s: The output value is %s." % (sql, output))
            return

        # if the table 'gpu_vector_info' is in user schema, then clean and reload GPU cache
        tmpExecPara = []
        tmpExecParaList = []
        databaseList = []
        schemaList = []

        # get all database
        databaseSQL = "SELECT datname FROM pg_catalog.pg_database WHERE datallowconn ORDER BY 1;"
        databaseList = self.getSQLResultList(databaseSQL, self.user, coorNode.hostname, coorNode.port)
        # we can makesure that the length of databaseList must greate than 1
        for database in databaseList:
            schemaSQL = "SELECT n.nspname FROM pg_catalog.pg_class c JOIN pg_catalog.pg_namespace n ON \
n.nspname NOT IN ('pg_toast', 'pg_catalog', 'information_schema', 'cstore') AND \
c.relnamespace = n.oid AND c.oid >= 16384 \
WHERE c.relname = 'gpu_vector_info';"
            schemaList = self.getSQLResultList(schemaSQL, self.user, coorNode.hostname, coorNode.port, database)
            for schema in schemaList:
                if ("" == schema):
                    continue
                tmpExecPara = []
                tmpExecPara.append(database)
                tmpExecPara.append(schema)
                tmpExecPara.append(coorNode.hostname)
                tmpExecPara.append(coorNode.port)
                tmpExecParaList.append(tmpExecPara)

        if (len(tmpExecParaList) == 0):
            return
        try:
            # Parallel processing cannot be used. 
            # The background will update a system table at the same time and generate an error.
            for execPara in tmpExecParaList:
                self.cleanAndReloadSingleGPUCache(execPara)
        except Exception, e:
            raise Exception(str(e))
        
    def getLcGroupMessage(self, LClist):
        """
        function: Obtain the corresponding OID through the logic cluster name
        input : LClist---the logic cluster name list
        output: group_member---mapping of logic cluster name and hostname
        """
        self.logger.debug("Obtaining OID through the logic cluster name.")
        try:
            #the map of group---hostname
            groupmessage = {}
            #get CN instance info from cluster
            cooInst = self.getCooInst()
            for lcname in LClist:
                #save the node ip contained in the logical cluster
                node_ips = []
                #save the node name contained in the logical cluster
                node_hosts = []
                #1.get node host ip by lc name
                host_sql = "SELECT distinct(pgxc_node.node_host) FROM pg_catalog.pgxc_node, pg_catalog.pgxc_group \
                           WHERE pgxc_group.group_name='%s' AND pgxc_node.oid = ANY(pgxc_group.group_members) \
                           order by pgxc_node.node_host; " % lcname
                (status, output) = ClusterCommand.remoteSQLCommand(host_sql,
                                                                   self.user, 
                                                                   cooInst[0].hostname,
                                                                   cooInst[0].port,
                                                                   ignoreError = False)
                if (status != 0):
                    raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % host_sql + " Error:\n%s" % str(output))
                elif (output != ""):
                    # use the split method of string to transform the output string to a list
                    node_ips = output.split('\n')

                #2.get hostname by hostip
                for dbnode in self.clusterInfo.dbNodes:
                    if ((dbnode.backIps[0] in node_ips) and (dbnode.name not in node_hosts)):
                        node_hosts.append(dbnode.name)

                #3.map of lcname-hostname
                groupmessage[lcname] = node_hosts

            self.logger.debug("Successfully obtained OID through the logic cluster name.")
        except Exception, e:
            raise Exception(str(e)) 

        return groupmessage
    
    def killCmserver(self, killSignal=DefaultValue.SIGNAL_RELOAD_PARA, cmserverNodes=[]):
        """
        function: Kill CMServer
        input : NA
        output: NA
        """
        self.logger.debug("Killing CMServer.")
        if(cmserverNodes == []):
            #traverse old nodes
            for dbNode in self.clusterInfo.dbNodes:
                if (len(dbNode.cmservers) > 0):
                    cmserverNodes.append(dbNode.name)
        self.logger.debug("CMServer nodes: %s." % cmserverNodes)

        failedNodes = []
        killCmd = "rm -rf '%s/bin/cluster_dynamic_config' && " % self.clusterInfo.appPath
        killCmd += "%s" % DefaultValue.killInstProcessCmd("cm_server", True, killSignal)
        (status, output) = self.sshTool.getSshStatusOutput(killCmd, cmserverNodes)

        for node in cmserverNodes:
            if(status[node] != DefaultValue.SUCCESS):
                failedNodes.append(node)

        #judge failed nodes
        if(len(failedNodes)):
            time.sleep(1)
            self.logger.debug("Failed nodes: %s." %  failedNodes)
            (status, output) = self.sshTool.getSshStatusOutput(killCmd, failedNodes)
            for node in failedNodes:
                if(status[node] != DefaultValue.SUCCESS):
                    raise Exception(output)
        self.logger.debug("Successfully killed CMServer.")

    def killAgent(self, killSignal=DefaultValue.SIGNAL_RELOAD_PARA, cmagentNodeNames=[]):
        """
        function: Kill Agent
        input : NA
        output: NA
        """
        self.logger.debug("Killing CMAgent.")
        #Get the cluster's node names
        if (len(cmagentNodeNames) == 0):
            cmagentNodes = self.clusterInfo.getClusterNodeNames()
        else:
            cmagentNodes = cmagentNodeNames
        self.logger.debug("CMAgent nodes: %s." % cmagentNodes)

        failedNodes = []
        killCmd = DefaultValue.killInstProcessCmd("cm_agent", True, killSignal)
        (status, output) = self.sshTool.getSshStatusOutput(killCmd, cmagentNodes)

        for node in cmagentNodes:
            if(status[node] != DefaultValue.SUCCESS):
                failedNodes.append(node)

        #judge failed nodes
        if(len(failedNodes)):
            time.sleep(1)
            self.logger.debug("Failed nodes: %s." %  failedNodes)
            (status, output) = self.sshTool.getSshStatusOutput(killCmd, failedNodes)
            for node in failedNodes:
                if(status[node] != DefaultValue.SUCCESS):
                    raise Exception(output)
        self.logger.debug("Successfully killed CMAgent.")
    
    
    def getListenIpList(self, dbNodeNames):
        """
        function:Obtain the listening IPs of the primary DN on the nodes
        input : NA
        output: NA
        """
        ipList = []
        
        #obtain the node IP list
        for nodename in dbNodeNames:
            dbNode = self.clusterInfo.getDbNodeByName(nodename)    
            for inst in dbNode.datanodes:
                if (inst.instanceType == DefaultValue.MASTER_INSTANCE and inst.listenIps[0] not in ipList):
                    ipList.append(inst.listenIps[0])
        
        #check the ipList
        if (len(ipList) == 0):
            raise Exception(ErrorCode.GAUSS_502["GAUSS_50219"] % "the node ip list")
        return ipList
    
    def getDnNameByListenIp(self, iplist):
        """
        function: Obtain the IP list of the node through the hostname list
        input : NA
        output: NA
        """
        self.logger.debug("Obtaining DN instance info through the IP list.")

        pgxc_node_name = []
        #get CN instance info from cluster
        cooInst = self.getCooInst()
        for ip in iplist:
            #obtain DN information
            obtainCmd = "SELECT node_name FROM pg_catalog.pgxc_node WHERE node_name like 'dn_%'"
            obtainCmd += " and node_host = '%s' ORDER BY node_name;" % ip
            #execute the sql command
            (status, output) = ClusterCommand.remoteSQLCommand(obtainCmd,
                                                               self.user, 
                                                               cooInst[0].hostname,
                                                               cooInst[0].port,
                                                               ignoreError = False)
            if (status != 0):
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % obtainCmd + " Error:\n%s" % str(output))
            elif (len(output.split("\n")) == 0):
                self.logger.log("Warning: node [%s] does not contain DN instance information." % ip)
            else:
                pgxc_node_name.extend(output.split("\n"))
      
        self.logger.debug("Successfully obtained DN instance info through the IP list.")
         
        return pgxc_node_name
    
    
    def createLogicalNodeGroup(self, lcGroupName, lcNodeNames):
        """
        function: Create node group for logic cluster.
        input : lcGroupName, nodeNames
        output: NA
        """
        self.logger.debug("Creating node group for logic cluster [%s]." % lcGroupName)
        temp_group_name = "group_version_%s_%s" % (time.strftime("%Y%m%d_%H%M%S"), os.getpid())
        tmpCgroupFilePath = "%s/etc/%s.gscgroup_perfadm.cfg" % (self.clusterInfo.appPath, temp_group_name)
        try:
            #get CN instance info from cluster
            cooInst = self.getCooInst()
            #if output=1, explain that the cluster has created a new nodegroup
            CHECK_GROUP_SQL = "select count(group_name) from pgxc_group where group_name = '%s' and in_redistribution = 'y';" % lcGroupName
            (status, output) = ClusterCommand.remoteSQLCommand(CHECK_GROUP_SQL,
                                                               self.user, 
                                                               cooInst[0].hostname,
                                                               cooInst[0].port,
                                                               ignoreError = False)
            if (status != 0):
                raise Exception(ErrorCode.GAUSS_516["GAUSS_51601"] % "node group information" + " Error:\n%s" % str(output))

            if (str(1) == output.strip()):
                self.logger.debug("The cluster has created a new nodegroup for logic cluster [%s]." % lcGroupName)
                return
            #Get IP address by node name '-h' + self.lcName
            iplist = self.getListenIpList(lcNodeNames)
            #get DN info by ip
            nodenameList = self.getDnNameByListenIp(iplist)
            
            #check if contain DN instance
            if (len(nodenameList) == 0):
                raise Exception(ErrorCode.GAUSS_502["GAUSS_50219"] % "DN instance" + " Maybe the node you are pointing to does not contain a DN instance.")
    
            #convert list to str
            dnstr = ",".join(nodenameList)
            #create tmp cgroup file for temp_group_name
            cmd = "%s/bin/gs_cgroup -c -N %s" % (self.clusterInfo.appPath, temp_group_name)
            self.sshTool.executeCommand(cmd, "create tmp cgroup file", DefaultValue.SUCCESS, lcNodeNames)
            #3.the sql command to create node group
            #it will automatically change the old LC's in_redistribution to 'y', and the new LC's in_redistribution to 't'
            #Bind users and LCs after redistribution
            START_TRANSACTION_SQL = "SET xc_maintenance_mode = on; START TRANSACTION; "
            END_TRANSACTION_SQL = "COMMIT; RESET xc_maintenance_mode; "
            CREATE_GROUP_SQL = "create node group \"%s\" with (%s) vcgroup DISTRIBUTE FROM \"%s\";" % (temp_group_name
                                                                                                    , dnstr
                                                                                                    , lcGroupName)
            LOAD_LCNAME_SQL = "select * from pgxc_cgroup_map_ng_conf('%s'); " % lcGroupName
            sql = "%s%s%s%s" % (START_TRANSACTION_SQL
                                , CREATE_GROUP_SQL
                                , LOAD_LCNAME_SQL
                                , END_TRANSACTION_SQL)
            self.logger.debug("Sql command for creating nodegroup and user: %s\n" % sql)
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.user, 
                                                               cooInst[0].hostname,
                                                               cooInst[0].port,
                                                               ignoreError = False)
            if (status != 0):
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))
        except Exception as e:
            self.cleanFilesOnAllNodes(tmpCgroupFilePath)
            raise Exception(str(e))
            
        self.logger.debug("Successfully created node group for logic cluster [%s]." % lcGroupName)
    
    def cleanFilesOnAllNodes(self, filePath, nodeNames=None):
        """
        function: clean files on all nodes.
        input : NA
        output: NA
        """
        try:
            cmd = g_file.SHELL_CMD_DICT["deleteFile"] % (filePath, filePath)
            if (nodeNames is None):
                self.sshTool.executeCommand(cmd, "clean files on all nodes")
            else:
                self.sshTool.executeCommand(cmd, "clean files on all nodes", DefaultValue.SUCCESS, nodeNames)
        except Exception as e:
            self.logger.debug("Failed to clean files on all nodes. Output:\n%s" % str(e))

    def loadCgroupFile(self, lcName):
        """
        function: Load the logical cluster cgroup configuration file on the DN and CN of all physical machine nodes
        input : lcName
        output: NA
        """
        self.logger.log("Loading the cgroup config file.")
        try:
            #get CN instance info from cluster
            cooInst = self.getCooInst()
            #map the DN and gs_cgroup 
            sql = "select * from pgxc_cgroup_map_ng_conf('%s'); " % lcName
            self.logger.debug("Sql command for loading the cgroup config file: %s" % sql)
            #load cgroup config file
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.user, 
                                                               cooInst[0].hostname,
                                                               cooInst[0].port,
                                                               ignoreError = False)
            if (status != 0):
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))
        except Exception, e:
            raise Exception(str(e))
            
        self.logger.log("Successfully loaded the cgroup config file.")

    def buildNewNodeInst(self, oldCnList, newNodeInstList, buildMode='full', dilatation_mode = "read-only"):
        """
        function: Getting DN building mapping.
        input : NA
        output: NA
        """
        #starting build
        self.logger.debug("Building the instances on new nodes by %s mode." % buildMode)
        try:
            newNodeInstList_bak = copy.deepcopy(newNodeInstList)
            while(len(newNodeInstList_bak) > 0):
                instMapping = self.getBuildInstMapping(oldCnList, newNodeInstList_bak)
                cmd = self.getBuildCommand(buildMode, instMapping)
                if (dilatation_mode == "insert"):
                    cmd += " --online-build"
                self.logger.debug("Command for building the instances on new nodes: %s" % cmd)
                self.sshTool.executeCommand(cmd, "build the new instances", DefaultValue.SUCCESS, self.newNodeNames)
                for instId in instMapping.keys():
                    newNodeInstList_bak.remove(instId)
        except Exception as e:
            killCmd = ""
            #Clean up the gaussdb process that may remain
            for newNodeName in self.newNodeNames:
                killCmd += " pssh -s -H %s '%s -9 gaussdb;%s -9 gsql' 2>/dev/null;" % (newNodeName,
                                                                                       findCmdInPath("killall"),
                                                                                       findCmdInPath("killall"))
            os.system(killCmd[:-1])
            raise Exception(str(e))
    
    def getAllNewNodeInst(self, ignoreCN=False):
        """
        function: Getting new node instance building mapping.
        input : NA
        output: NA
        """
        newNodeDnTmpList = []
        maxDnNumPerNode = 0
        newNodeInstList = []
        #Get a two-dimensional array of the primary DN of each node:newNodeDnList
        for i in range(len(self.newNodes)):
            dbNode = self.newNodes[i]
            if (not ignoreCN and len(dbNode.coordinators) >= 1):
                newNodeInstList.append(dbNode.coordinators[0].instanceId)
            masterDnList = []
            for datanode in dbNode.datanodes:
                if (datanode.instanceType == DefaultValue.MASTER_INSTANCE):
                    masterDnList.append(datanode.instanceId)
            if (len(masterDnList) > maxDnNumPerNode):
                maxDnNumPerNode = len(masterDnList)
            newNodeDnTmpList.append(masterDnList)
        
        #Rearrange two-dimensional arrays by column
        for i in range(maxDnNumPerNode):
            for j in newNodeDnTmpList:
                if (len(j) > i):
                    newNodeInstList.append(j[i])

        return newNodeInstList
    
    def getBuildInstMapping(self, cnList, newNodeInstList):
        """
        function: Getting new node instance building mapping.
        input : cnList, newNodeInstList
        output: newNodeInstMapping
        """
        newNodeInstMapping = {}
        cnNums = len(cnList)
        for j in range(cnNums):
            for i in range(len(newNodeInstList)):
                if (i/cnNums == self.jobs):
                    break
                if ( i % cnNums == j):
                    newNodeInstMapping[newNodeInstList[i]] = cnList[j].instanceId
        return newNodeInstMapping

    def manageGucParameterForBuild(self, nodeNames, mode = "set", NormalcooInsts = None):
        """
        function: Set "wal_keep_segments=1000000" and 'enable_prevent_job_task_startup=on' by guc tool.
                  Before starting CN build DN, you need to set the parameters of 
                  Cn and set the parameter value to 1000000 to ensure that 
                  the xlog log of the remote CN is not deleted.
        input : mode : set,restore 
                     set: Save the values of the parameters to a temporary file
                     restore: Read the wal_keep_segments value of the parameter and restore it.
        output: NA
        """
        self.logger.debug("Setting guc parameter for building new nodes.")
        buildParaMap = {'wal_keep_segments':1000000,
                        'enable_prevent_job_task_startup':'on'}
        tmpFile = "%s/wal_keep_segments.dat" % DefaultValue.getTmpDirFromEnv(self.user)
        gucStr = ""
        if (mode == "set"):
            if (not os.path.isfile(tmpFile)):

                sql = "select name,setting from pg_settings where name in ('wal_keep_segments', 'enable_prevent_job_task_startup');"
                if (NormalcooInsts is None):
                    (status, output) = ClusterCommand.remoteSQLCommand(sql, self.user, self.cooInst.hostname, self.cooInst.port, False, DefaultValue.DEFAULT_DB_NAME)
                else:
                    (status, output) = ClusterCommand.remoteSQLCommand(sql, self.user, NormalcooInsts[0].hostname, NormalcooInsts[0].port, False, DefaultValue.DEFAULT_DB_NAME)
                if (status !=0 ):
                    raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))
                self.logger.debug("Record the wal_keep_segments settings before building:\n %s" % output)
                g_file.createFile(tmpFile, True, DefaultValue.KEY_FILE_MODE)
                g_file.writeFile(tmpFile, [output], mode="w")
                self.sshTool.scpFiles(tmpFile, tmpFile, nodeNames)
            else:
                self.logger.debug("Record the wal_keep_segments settings file already exists.")
            for key in buildParaMap.keys():
                gucStr += " -c \"%s=%s\" " % (key, buildParaMap[key])
        else:
            if (os.path.isfile(tmpFile)):
                lines = g_file.readFile(tmpFile)
                for line in lines:
                    paraInfo = line.split('|')
                    gucStr += "-c \"%s=%s\" " % (paraInfo[0].strip(), paraInfo[1].strip())
            else:
                self.logger.debug("Record the wal_keep_segments settings file does not exists.")
                return

        cmd = ""
        needDelayXlogRecycleNodeNames = []
        if (NormalcooInsts is None):
            cmd = "gs_guc reload -Z coordinator -N all -I all %s" % gucStr
            needDelayXlogRecycleNodeNames = nodeNames
        else:
            for cooInst in NormalcooInsts:
                needDelayXlogRecycleNodeNames.append(cooInst.hostname)
                if (cmd == ""):
                    cmd = "gs_guc reload -Z coordinator -N %s -I all %s" % (cooInst.hostname, gucStr)
                else:
                    cmd += " && gs_guc reload -Z coordinator -N %s -I all %s" % (cooInst.hostname, gucStr)
        
        self.logger.debug("Command to set GUC parameters: %s" % cmd)
        (status, output) = DefaultValue.retryGetstatusoutput(cmd)
        if (status != 0):
            raise Exception(ErrorCode.GAUSS_500["GAUSS_50007"] % "GUC" + " Command:%s. Error:\n%s" % (cmd , output))
        else:
            if (mode == "restore"):
                #After restoring the wal_keep_segments settings, delete the temporary file
                cmd = g_file.SHELL_CMD_DICT["deleteFile"] % (tmpFile, tmpFile)
                self.sshTool.executeCommand(cmd, "clean the record file of guc settings", DefaultValue.SUCCESS, nodeNames)

        self.logger.debug("Successfully set guc parameter for building new nodes.")
        self.setDelayXlogRecycleForCN(needDelayXlogRecycleNodeNames, mode)

    def setDelayXlogRecycleForCN(self, nodeNames, mode="restore"):
        """
        function: In the CNbuildCN or CnbuildDn scenario, logs cannot be reclaimed before build.
                  After the build is complete, logs are reclaimed.
        input : nodeNames
                mode
        output: NA
        """
        self.logger.debug("Setting xlog recycel for cn instance.")
        self.logger.debug("The following nodes need to be set: %s" % nodeNames)
        cmd = "%s --set-cnlog-recycle %s -U %s -l '%s'" % (
                                                            OMCommand.getLocalScript("Local_Query")
                                                            , mode
                                                            , self.user
                                                            , self.localLog)
        self.logger.debug("Command for setting xlog recycel: %s" % cmd)
        self.sshTool.executeCommand(cmd, "set xlog recycle", DefaultValue.SUCCESS, nodeNames)
        self.logger.debug("Successfully setting xlog recycel for cn instance.")
    
    def removeBuildFile(self, nodeNames):
        """
        function: Remove build file (gs_build.pid) of master DN instance for new nodes.
                  If you do not delete the file, it will cause the new node to fail to start.
        input : NA
        output: NA
        """
        self.logger.debug("Removing build file for new nodes.")
        cmd = "%s --target=remove_build_file -X '%s' -l '%s'" % (
                                                            OMCommand.getLocalScript("Local_Dilatation_Config")
                                                            , self.xmlFile
                                                            , self.localLog)
        self.logger.debug("Command for removing build file: %s" % cmd)
        self.sshTool.executeCommand(cmd, "remove build file", DefaultValue.SUCCESS, nodeNames)
        self.logger.debug("Successfully removed build file.")
    
    def checkLcGroupNameExist(self, lcName):
        """
        function: Check if the logical cluster name exists
        input :NA
        output:NA
        """
        self.logger.debug("Checking the logical cluster name.")
        #get CN instance info from cluster
        cooInst = self.getCooInst()
        sql = "SELECT count(*) FROM pg_catalog.pgxc_group \
               WHERE group_name='%s' and group_kind = 'v';" % lcName
        (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                           self.user, 
                                                           cooInst[0].hostname,
                                                           cooInst[0].port,
                                                           ignoreError = False)
        if (status != 0 or not output.isdigit()):
            raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))
        elif (int(output) == 0):
            raise Exception(ErrorCode.GAUSS_500["GAUSS_50008"] + " The specified logical cluster name does not exist in the cluster.")
        self.logger.debug("Successfully checked the logical cluster name.")
    
    def getCooInst(self):
        """
        function: get CN instance
        input : NA
        output: CN instance
        """
        coorInst = []
        #get CN on nodes
        for dbNode in self.clusterInfo.dbNodes:
            if (len(dbNode.coordinators) >= 1):
                coorInst.append(dbNode.coordinators[0])
        #check if contain CN on nodes
        if (len(coorInst) == 0):
            raise Exception(ErrorCode.GAUSS_526["GAUSS_52602"])
        else:
            return coorInst
    
    def getGroupName(self, fieldName, fieldVaule):
        """
        function: Get nodegroup name by field name and field vaule.
        input : field name and field vaule
        output: node group name
        """
        #1.get CN instance info from cluster
        cooInst = self.getCooInst()

        #2.obtain the node group
        OBTAIN_SQL = "select group_name from pgxc_group where %s = %s; " % (fieldName, fieldVaule)
        #execute the sql command
        (status, output) = ClusterCommand.remoteSQLCommand(OBTAIN_SQL,
                                                           self.user,
                                                           cooInst[0].hostname,
                                                           cooInst[0].port,
                                                           ignoreError = False)
        if (status != 0):
            raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % OBTAIN_SQL + " Error:\n%s" % str(output))

        return output.strip()

    def getAllDatabase(self):
        """
        function: Get all the databases in the cluster
        output: databaseList
        """
        self.logger.debug("Getting all the databases.")
        try:
            #get CN instance info from cluster
            cooInst = self.getCooInst()

            # get databases from cluster
            selectSql = "SELECT datname FROM pg_catalog.pg_database WHERE datallowconn ORDER BY 1"
            self.logger.debug("Get databases from cluster command: %s." % selectSql)
            #execute the sql command
            (status, output) = ClusterCommand.remoteSQLCommand(selectSql,
                                                               self.user,
                                                               cooInst[0].hostname,
                                                               cooInst[0].port,
                                                               ignoreError = False)
            if (status != 0):
                raise Exception(ErrorCode.GAUSS_513["GAUSS_51300"] % selectSql + " Error: \n%s" % str(output))
            self.logger.debug("Get databases from cluster result: %s." % output)
            #split the output string with '\n'
            databaseList = output.split("\n")

            self.logger.debug("Successfully get all the databases.")

            return databaseList
        except Exception, e:
            raise Exception(str(e))
    
    def updateTableDepend(self, sourceGroup, destGroup):
        """
        function: Update table dependencies
        input : sourceGroup---source node group
                destGroup---destination node group
        output: NA
        """
        self.logger.log("Updating table dependencies.")
        try:
            #1.get CN instance info from cluster
            cooInst = self.getCooInst()

            #2..get all database
            databases = self.getAllDatabase()
            #3.remove 'template1'
            databases.remove("template1")

            #4.exectue sql in every database. bind user and node group
            START_TRANSACTION_SQL = "SET xc_maintenance_mode = on; START TRANSACTION; "
            ALTER_SQL = "alter node group \"%s\" set table group \"%s\";" % (sourceGroup , destGroup)

            END_TRANSACTION_SQL = "COMMIT; RESET xc_maintenance_mode; "
            #total sql command
            rename_sql = "%s%s%s" % (START_TRANSACTION_SQL, ALTER_SQL, END_TRANSACTION_SQL)
            #execute sql statement in each database
            for database in databases:
                self.logger.debug("exectue rename operation in database [%s]: %s." % (database, rename_sql))
                #execute the sql command in a different database
                (status, output) = ClusterCommand.remoteSQLCommand(rename_sql,
                                                                   self.user,
                                                                   cooInst[0].hostname,
                                                                   cooInst[0].port,
                                                                   False,
                                                                   database)
                if (status != 0):
                    raise Exception(ErrorCode.GAUSS_513["GAUSS_51300"] % rename_sql + " Error:\n%s" % output)
        except Exception, e:
            raise Exception(str(e))

        self.logger.log("Successfully updated table dependencies.")

    def updateGroupInfo(self, sourceGroup, destGroup):
        """
        function: Update group_buckets and group_members
        input : sourceGroup---source node group
                destGroup---destination node group
        output: NA
        """
        self.checkClusterLock()
        self.logger.log("Updating group buckets and node list.")
        try:
            #1.turn on enable_online_ddl_waitlock
            self.setOnlineGucPamater(True)
            #2.lock cluster
            self.logger.log("Locking cluster.")
            self.lockMode = "waitingLock"
            lockNode = self.lockClusterInternal(self.clusterInfo.dbNodes,DefaultValue.CLUSTER_LOCK_TIME,self.logger,self.sshTool,'redis',self)
            if (lockNode is None):
                self.setOnlineGucPamater(False)
                raise Exception(ErrorCode.GAUSS_516["GAUSS_51608"] + ".")
            self.logger.log("Successfully locked cluster.")
            #3.get CN instance info from cluster
            cooInst = self.getCooInst()
            #4.build sql for copy buckets
            START_TRANSACTION_SQL = "SET xc_maintenance_mode = on; START TRANSACTION;"
            ALTER_SQL = "ALTER NODE GROUP \"%s\" COPY BUCKETS FROM \"%s\";" % (destGroup,sourceGroup)
            END_TRANSACTION_SQL = "COMMIT; RESET xc_maintenance_mode;"
            update_sql = "%s%s%s" % (START_TRANSACTION_SQL, ALTER_SQL, END_TRANSACTION_SQL)
            #5.execute query
            (status, output) = ClusterCommand.remoteSQLCommand(update_sql,
                                                                       self.user,
                                                                       cooInst[0].hostname,
                                                                       cooInst[0].port,
                                                                       False)
            #6.unlock cluster
            self.logger.log("Unlocking cluster.")
            coorNodes = []
            coorNodes.append(lockNode)
            self.unlockClusterInternal(coorNodes,DefaultValue.CLUSTER_LOCK_TIME,self.logger,self.sshTool,'redis',self)
            self.logger.log("Successfully unlocked cluster.")
            #7.turn off enable_online_ddl_waitlock
            self.setOnlineGucPamater(False)
            #query failed
            if (status != 0):
                raise Exception(ErrorCode.GAUSS_513["GAUSS_51300"] % update_sql + " Error:\n%s" % output)
        except Exception, e:
            raise Exception(str(e))
        self.logger.log("Successfully update group buckets and node list.")

    def checkClusterLock(self):
        """
        function: Check if another cluster lock in use
        input : NA
        output: NA
        """
        self.logger.log("Check if another cluster lock in use.")
        checkLockSql = "select * from pg_stat_activity where query like "
        checkLockSql += "'select case (select pgxc_lock_for_backup()) when true then (select pg_sleep(%)::text) end;' "
        cooInst = self.getCooInst()
        (status, output) = ClusterCommand.remoteSQLCommand(checkLockSql,
                                                                   self.user,
                                                                   cooInst[0].hostname,
                                                                   cooInst[0].port,
                                                                   False)
        if (status != 0):
            raise Exception(ErrorCode.GAUSS_513["GAUSS_51300"] % checkLockSql + " Error:\n%s" % output)
        if(output != ""):
            self.logger.log("There is another cluster lock in use.\nCluster info in pg_stat_activity is below.")
            raise Exception(output)
        self.logger.log("There is no cluster lock in use.")

    def setOnlineGucPamater(self, mode=False, Nodenames = None):
        """
        function: Set GUC parameters for online dilatation
        input : mode
        output: NA
        """
        self.logger.debug("Setting GUC parameter for online.")
        if not Nodenames:
            cmd = "gs_guc reload -Z coordinator -N all -I all -c \"enable_online_ddl_waitlock=%s\"" % mode
            self.logger.debug("Command for setting GUC parameter for online: %s" % cmd)
            (status, output) = DefaultValue.retryGetstatusoutput(cmd)
            if (status != 0):
                raise Exception(ErrorCode.GAUSS_500["GAUSS_50008"] + "Error:\n%s" % output)
        else:
            for nodename in Nodenames:
                cmd = "gs_guc reload -Z coordinator -N %s -I all -c \"enable_online_ddl_waitlock=%s\"" % (nodename, mode)
                self.logger.debug("Command for setting GUC parameter for online: %s" % cmd)
                (status, output) = DefaultValue.retryGetstatusoutput(cmd)
                if (status != 0):
                    raise Exception(ErrorCode.GAUSS_500["GAUSS_50008"] + "Error:\n%s" % output)
        time.sleep(5)
        self.logger.debug("Successfully set GUC parameter for online.")

    def cleanPreTransaction(self):
        """
        function:Run gs_clean on all cns to clean temp table
        input:NA
        output:NA
        """
        self.logger.debug("Cleaning up outstanding 2PCs before delete nodegroup.")
        cnNodenames = []
        for dbNode in self.clusterInfo.dbNodes:
            if (len(dbNode.coordinators) > 0):
                cnNodenames.append(dbNode.name)
        cmd = "%s --target=cleanup2PCs -l '%s'" % (OMCommand.getLocalScript("Local_Dilatation_Config"), self.localLog)
        self.sshTool.executeCommand(cmd, "execute gs_clean", DefaultValue.SUCCESS, cnNodenames)
        self.logger.debug("Successfully cleaned up outstanding 2PCs before delete nodegroup.")
    
    def dropNodeGroup(self, sourceGroup, destGroup):
        """
        function: drop new group for logical cluster.
        input : sourceGroup---source node group
                destGroup---destination node group
        output: NA
        """
        self.cleanPreTransaction()
        self.logger.debug("Deleting node group.")

        #1.get CN instance info from cluster
        cooInst = self.getCooInst()

        #2.the sql command to drop node group
        START_TRANSACTION_SQL = "SET xc_maintenance_mode = on; START TRANSACTION; "
        DROP_SQL = "drop node group \"%s\" distribute from \"%s\"; " % (sourceGroup, destGroup)
        END_TRANSACTION_SQL = "COMMIT; RESET xc_maintenance_mode; "

        #total drop node group sql
        drop_sql = "%s%s%s" % (START_TRANSACTION_SQL, DROP_SQL, END_TRANSACTION_SQL)
        self.logger.debug("Sql command for drop nodegroup: %s\n" % drop_sql)
        #execute the sql command
        (status, output) = ClusterCommand.remoteSQLCommand(drop_sql,
                                                           self.user,
                                                           cooInst[0].hostname,
                                                           cooInst[0].port,
                                                           ignoreError = False)
        if (status != 0):
            raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % DROP_SQL + " Error:\n%s" % str(output))
        tmpCgroupFilePath = "%s/etc/%s.gscgroup_%s.cfg" % (self.clusterInfo.appPath, sourceGroup, self.user)
        self.cleanFilesOnAllNodes(tmpCgroupFilePath)
        self.logger.debug("Successfully deleted node group.")

    def dropAndUpdateGroupMessage(self):
        """
        function: drop and update group step
                  1. update table depend.
                  2. exchange group_members of new group with old group.
                  3. update the in_redistribution field information of the old node group.
        input : NA
        output: NA
        """
        self.logger.debug("Deleting and updating node group.")

        #1.get new group name and old group name
        #get new group name
        new_group = self.getGroupName("in_redistribution", "'t'")
        #get old group name
        old_group = self.getGroupName("in_redistribution", "'y'")
        if (new_group == "" or old_group == ""):
            self.logger.debug("Can not find new and old node groups, group information ('%s', '%s')." % (new_group
                                                                                                        , old_group))
            return
        #2.Update group_buckets and group_members
        self.updateGroupInfo(new_group, old_group)
        #3.Update table dependencies
        self.updateTableDepend(new_group, old_group)

        #4.drop new group
        self.dropNodeGroup(new_group, old_group)

        self.logger.debug("Successfully deleted and updated node group.")

    def deleteRedisStatusAndDetailTable(self):
        """
        function: Delete the last remaining status and details table of redis
        input : NA
        output: NA
        """
        self.logger.debug("Delete the last remaining status and details table of redis.")
        cooInst = self.getCooInst()[0]
        sql = "DROP TABLE IF EXISTS redis_progress_detail CASCADE;"
        (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                           self.user,
                                                           cooInst.hostname,
                                                           cooInst.port,
                                                           ignoreError = False)
        if(status != 0):
            raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        sql = "DROP TABLE IF EXISTS redis_status CASCADE;"
        (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                           self.user,
                                                           cooInst.hostname,
                                                           cooInst.port,
                                                           ignoreError = False)
        if(status != 0):
            raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))
        self.logger.debug("Successfully deleted the last remaining status and details table of redis.")

    def manageCMControlFile(self, mode = "delete"):
        """
        function: manage CM control file:
                Create a tag file in $GAUSSHOME/bin for the instance being expanded. 
               The name is instance_maintance, The file records the instance ID of the primary DN on the new node.
               and the expansion is completed or deleted after the rollback. The CM encounters an instance that is being expanded. 
               For the static primary dn, it is directly pulled up. Otherwise, Performing primary and standby arbitration.
        input : mode:'add'/'delete'
        output: NA
        """
        self.logger.debug("Managing CMServer control file for '%s' mode." % mode)
        cmservers = []

        for dbNode in self.clusterInfo.dbNodes:
            if (len(dbNode.cmservers) > 0):
                cmservers.append(dbNode.name)
        controlFilePath = "%s/bin/instance_maintance" % self.clusterInfo.appPath
        if (mode == "add"):
            cmd = "%s --target generateCmMaintanceFile -l %s --nodes %s" % (OMCommand.getLocalScript("Local_Dilatation_Config"), 
                                                                            self.localLog,
                                                                            ",".join(self.newNodeNames))
            self.sshTool.executeCommand(cmd, "generate instance maintance file", DefaultValue.SUCCESS, cmservers)
        else:
            cmd = g_file.SHELL_CMD_DICT["deleteFile"] % (controlFilePath
                                                        , controlFilePath)
            failedNodes = []
            (status, output) = self.sshTool.getSshStatusOutput(cmd, cmservers)
         
            for node in cmservers:
                if(status[node] != DefaultValue.SUCCESS):
                    failedNodes.append(node)
    
            #judge failed nodes
            if(len(failedNodes)):
                time.sleep(1)
                self.logger.debug("Failed nodes: %s." %  failedNodes)
                (status, output) = self.sshTool.getSshStatusOutput(cmd, failedNodes)
                for node in failedNodes:
                    if(status[node] != DefaultValue.SUCCESS):
                        raise Exception(ErrorCode.GAUSS_502["GAUSS_50207"] % "instance_maintance" + " Error:\n%s" % output)
        
        self.logger.debug("Successfully managed CMServer control file.")

    def getEtcdUrl_v2(self):
        """
        function: get etcd url and etcd host names
        input : NA
        output: NA
        """
        etcdhost = []
        gaussHome       = DefaultValue.getInstallDir(self.user)
        crtPath         = "%s/share/sslcert/etcd/etcdca.crt" % gaussHome
        clientkeyPath   = "%s/share/sslcert/etcd/client.key" % gaussHome
        clientcrtPath   = "%s/share/sslcert/etcd/client.crt" % gaussHome
        cmdpre          = "export ETCDCTL_API=2;etcdctl --ca-file %s --cert-file %s --key-file %s --total-timeout 60s --peers " % \
                          (crtPath, clientcrtPath, clientkeyPath)
        tmpcmdpre       = cmdpre

        for node in self.clusterInfo.dbNodes:
            if (len(node.etcds) > 0):
                etcdhost.append(node.name)
                etcdinst = node.etcds[0]
                cmdpre += "https://%s:%s," % (etcdinst.listenIps[0], etcdinst.port)

        if (cmdpre == tmpcmdpre):
            self.logger.debug("The cluster is not installed etcd.")
            return

        cmdpre = cmdpre[:-1]

        self.logger.debug("The etcdctl cmdpre:%s" % cmdpre)
        return cmdpre,etcdhost

    def getEtcdUrl_v3(self):
        """
        function: get etcd url and etcd host names
        input : NA
        output: NA
        """
        etcdhost = []
        gaussHome       = DefaultValue.getInstallDir(self.user)
        crtPath         = "%s/share/sslcert/etcd/etcdca.crt" % gaussHome
        clientkeyPath   = "%s/share/sslcert/etcd/client.key" % gaussHome
        clientcrtPath   = "%s/share/sslcert/etcd/client.crt" % gaussHome
        cmdpre          = "export ETCDCTL_API=3;etcdctl --cacert %s --cert %s --key %s --command-timeout 60s --endpoints " % \
                          (crtPath, clientcrtPath, clientkeyPath)
        tmpcmdpre       = cmdpre

        for node in self.clusterInfo.dbNodes:
            if (len(node.etcds) > 0):
                etcdhost.append(node.name)
                etcdinst = node.etcds[0]
                cmdpre += "https://%s:%s," % (etcdinst.listenIps[0], etcdinst.port)

        if (cmdpre == tmpcmdpre):
            self.logger.debug("The cluster is not installed etcd.")
            return

        if (cmdpre != tmpcmdpre):
            cmdpre = cmdpre[:-1]

        self.logger.debug("The etcdctl cmdpre:%s" % cmdpre)
        return cmdpre,etcdhost

    def hasETCD(self):
        """
        function: has etcd in cluster or not.
        input : NA
        output: NA
        """
        for dbNodes in self.clusterInfo.dbNodes:
            if(len(dbNodes.etcds) > 0):
                return True
        return False

    def startClusterEtcd(self):
        """
        function: start the etcd and check etcd is healthy
        input : NA
        output: NA
        """
        self.logger.debug("Starting cluster Etcd.")

        (cmdpre, etchhost) = self.getEtcdUrl_v3()
        gaussHome = DefaultValue.getInstallDir(self.user)
        etcdflagfile = "%s/bin/etcd_manual_start" % gaussHome
        cmd = g_file.SHELL_CMD_DICT["deleteFile"] % (etcdflagfile, etcdflagfile)
        self.logger.debug("Command for deleting ETCD manual start file: %s" % cmd)
        DefaultValue.execCommandWithMode( cmd, 
                                          "start etcd", 
                                          self.sshTool,
                                          self.isSingle, 
                                          self.mpprcFile, 
                                          etchhost)
        time.sleep(5)

        checkEtcdHealthcmd = "%s endpoint health" % cmdpre
        check_times = 0
        while True:
            (status, output) = commands.getstatusoutput(checkEtcdHealthcmd)
            self.logger.debug("The ETCD healthy cmd:%s\nstatus:%s\noutput:%s" % \
                 (checkEtcdHealthcmd, status, output))
            if (status != 0 or output.find("unhealthy") >= 0):
                if (check_times >= 12):
                    raise Exception(ErrorCode.GAUSS_516["GAUSS_51607"] % "etcd")
                time.sleep(5)
                check_times += 1
            else:
                break

        self.logger.debug("Successfully start cluster Etcd.")

    def checkEtcdStatus(self):
        """
        function: set gxid and check it's value equal 20000
        input : NA
        output: NA
        """
        self.logger.debug("check etcd status")

        (cmdpre, etchhost) = self.getEtcdUrl_v3()
        set_times = 0
        checkcmd = "%s endpoint status" % (cmdpre)
        status = 0
        output = ""
        while True:
            (status, output) = commands.getstatusoutput(checkcmd)
            self.logger.debug("check etcd status cmd:%s\nstatus:%s\noutput:%s" % \
                 (checkcmd, status, output))
            if (status != 0 or output.find("Failed") >= 0):
                if (set_times >= 12):
                    raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % \
                          checkcmd + " Error:\n%s" % str(output))
                time.sleep(5)
                set_times += 1
            else:
                break

        self.logger.debug("Successfully check etcd status.")
        return (status,output)

    def setGxid(self):
        """
        function: set gxid and check it's value equal 20000
        input : NA
        output: NA
        """
        self.logger.debug("Set gxid on etcd.")

        (cmdpre, etchhost) = self.getEtcdUrl_v3()
        set_times = 0
        setgxidcmd = "%s put /%s/gxid 20000" % (cmdpre, self.user)
        while True:    
            (status, output) = commands.getstatusoutput(setgxidcmd)
            self.logger.debug("Set ETCD gxid cmd:%s\nstatus:%s\noutput:%s" % \
                 (setgxidcmd, status, output))
            if (status != 0):
                if (set_times >= 12):
                    raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % \
                          setgxidcmd + " Error:\n%s" % str(output))
                time.sleep(5)
                set_times += 1
            else:
                checkgxidcmd = "%s get /%s/gxid" % (cmdpre, self.user)
                self.logger.debug("Check ETCD gxid cmd:%s\nstatus:%s\noutput:%s" % \
                     (checkgxidcmd, status, output))
                if (status != 0 or output.find("OK") < 0):
                    if (set_times >= 12):
                        raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % \
                              checkgxidcmd + " Error:\n%s" % str(output))
                    time.sleep(5)
                    set_times += 1
                else:
                    break

        self.logger.debug("Successfully set the gxid.")

    def getGxid_v3(self):
        """
        function: get gxid
        input : NA
        output: NA
        """
        self.logger.debug("Get gxid on etcd.")

        (cmdpre, etchhost) = self.getEtcdUrl_v3()
        set_times = 0
        getgxidcmd = "%s get /%s/gxid" % (cmdpre, self.user)
        while True:    
            (status, output) = commands.getstatusoutput(getgxidcmd)
            self.logger.debug("Get ETCD gxid cmd:%s\nstatus:%s\noutput:%s" % \
                 (getgxidcmd, status, output))
            if (status != 0):
                if (set_times >= 12):
                    raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % \
                          getgxidcmd + " Error:\n%s" % str(output))
                time.sleep(5)
                set_times += 1
            else:
                break

        self.logger.debug("Successfully get the gxid.")
        
        return (status,output.split()[1])
    
    def getTerm_v3(self):
        """
        function: get Term
        input : NA
        output: NA
        """
        self.logger.debug("Get term on etcd.")

        (cmdpre, etchhost) = self.getEtcdUrl_v3()
        set_times = 0
        gettermcmd = "%s get /%s/CMServer/status_key/term" % (cmdpre, self.user)
        while True:    
            (status, output) = commands.getstatusoutput(gettermcmd)
            self.logger.debug("Get ETCD term cmd:%s\nstatus:%s\noutput:%s" % \
                 (gettermcmd, status, output))
            if (status != 0):
                if (set_times >= 12):
                    raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % \
                          gettermcmd + " Error:\n%s" % str(output))
                time.sleep(5)
                set_times += 1
            else:
                break

        self.logger.debug("Successfully get the gxid.")

        return (status,output.split()[1])

    def setGxid_v2(self, value):
        """
        function: set gxid and check it's value
        input : NA
        output: NA
        """
        self.logger.debug("Set gxid on etcd.")

        (cmdpre, etchhost) = self.getEtcdUrl_v2()
        set_times = 0
        setgxidcmd = "%s set /%s/gxid %s" % (cmdpre, self.user, value)
        while True:    
            (status, output) = commands.getstatusoutput(setgxidcmd)
            self.logger.debug("Set ETCD gxid cmd:%s\nstatus:%s\noutput:%s" % \
                 (setgxidcmd, status, output))
            if (status != 0):
                if (set_times >= 12):
                    raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % \
                          setgxidcmd + " Error:\n%s" % str(output))
                time.sleep(5)
                set_times += 1
            else:
                checkgxidcmd = "%s get /%s/gxid" % (cmdpre, self.user)
                self.logger.debug("Check ETCD gxid cmd:%s\nstatus:%s\noutput:%s" % \
                     (checkgxidcmd, status, output))
                if (status != 0 or output.find("OK") < 0):
                    if (set_times >= 12):
                        raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % \
                              checkgxidcmd + " Error:\n%s" % str(output))
                    time.sleep(5)
                    set_times += 1
                else:
                    break

        self.logger.debug("Successfully set the gxid.")

    def setTerm_v2(self, value):
        """
        function: set term and check it's value
        input : NA
        output: NA
        """
        self.logger.debug("Set term on etcd.")

        (cmdpre, etchhost) = self.getEtcdUrl_v2()
        set_times = 0
        setgxidcmd = "%s set /%s/CMServer/status_key/term %s" % (cmdpre, self.user, value)
        while True:
            (status, output) = commands.getstatusoutput(setgxidcmd)
            self.logger.debug("Set ETCD term cmd:%s\nstatus:%s\noutput:%s" % \
                 (setgxidcmd, status, output))
            if (status != 0):
                if (set_times >= 12):
                    raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % \
                          setgxidcmd + " Error:\n%s" % str(output))
                time.sleep(5)
                set_times += 1
            else:
                checktermcmd = "%s get /%s/CMServer/status_key/term" % (cmdpre, self.user)
                self.logger.debug("Check ETCD term cmd:%s\nstatus:%s\noutput:%s" % \
                     (checktermcmd, status, output))
                if (status != 0 or output.find("OK") < 0):
                    if (set_times >= 12):
                        raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % \
                              checktermcmd + " Error:\n%s" % str(output))
                    time.sleep(5)
                    set_times += 1
                else:
                    break

        self.logger.debug("Successfully set the term.")
    
    def stopClusterEtcd(self):
        """
        function: stop etcd
        input : NA
        output: NA
        """
        self.logger.debug("Stoping cluster ETCD.")

        (cmdpre, etchhost) = self.getEtcdUrl_v3()
        gaussHome = DefaultValue.getInstallDir(self.user)
        etcdflagfile = "%s/bin/etcd_manual_start" % gaussHome
        cmd = g_file.SHELL_CMD_DICT["createFile"] % \
              (etcdflagfile, DefaultValue.KEY_FILE_MODE, etcdflagfile)
        DefaultValue.execCommandWithMode( cmd, 
                                          "stop etcd",
                                          self.sshTool, 
                                          self.isSingle, 
                                          self.mpprcFile, 
                                          etchhost)
        time.sleep(10)

        self.logger.debug("Successfully stoped cluster ETCD.")

    def killKernalSnapshotThread(self, coorInst):
        """
        function: kill snapshot thread in Kernal,
                avoid dead lock with redistribution)
        input : NA
        output: NA
        """
        self.logger.debug("Stopping snapshot thread in DB Kernal.")
        killSnapshotSQL = "select * from kill_snapshot();"

        (status, output) = ClusterCommand.remoteSQLCommand(killSnapshotSQL, self.user, coorInst.hostname, coorInst.port, False, DefaultValue.DEFAULT_DB_NAME)
        if (status !=0 ):
            raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % killSnapshotSQL + " Error:\n%s" % str(output))
        self.logger.debug("Successfully stopped snapshot thread in DB Kernal.")

    def createGsqlCA(self, hostList=None):
        """
        function: create gsql ca file
        input : NA
        output: NA
        """
        huaweiyun = DefaultValue.isHuaweiyunScene()
        if huaweiyun:
            self.logger.debug("HuaweiyunScene:so skip generate gsql CA files.")
            return
        self.logger.debug("Generating gsql CA files.")
        if hostList is None:
            hostList = []
        appPath = DefaultValue.getInstallDir(self.user)
        caPath = os.path.join(appPath, "share/sslcert/gsql")
        self.logger.debug("The ca file dir is: %s." % caPath)
        if (len(hostList) == 0):
            for dbNode in self.clusterInfo.dbNodes:
                hostList.append(dbNode.name)
        try:
            # Create CA dir and prepare files for using.
            self.logger.debug("Create gsql CA file directory.")
            DefaultValue.localCreateCaDir(caPath)
            self.logger.debug("Successfully create gsql CA file directory.")
            # create cert
            randpass = DefaultValue.getRandStr()
            DefaultValue.createRootCA(randpass, caPath, self.logger)
            DefaultValue.createServerCA(DefaultValue.SERVER_CA, randpass, caPath, self.logger)
            DefaultValue.createClientCA(DefaultValue.Client_CA, randpass, caPath, self.logger)
            randpass = ""
            # Clean useless files.
            self.logger.debug("clean the dir of ca file：demoCA, server.req, client.req")
            DefaultValue.cleanServerCaDir(caPath)
            self.logger.debug("Successfully clean the dir of ca file：demoCA, server.req, client.req")

            #change permission of ca files to 600
            self.logger.debug("change permission of ca files to 600")
            for certFile in DefaultValue.GSQL_CERT_LIST:
                chmod_file = os.path.join(caPath, "%s" % certFile)
                g_file.changeMode(DefaultValue.KEY_FILE_MODE, chmod_file)
            self.logger.debug("Successfully change permission of ca files to 600")
        except Exception as e:
            randpass = ""
            certFile = caPath + "/demoCA/cacert.pem"
            if os.path.exists(certFile):
                g_file.removeFile(certFile)
            DefaultValue.cleanServerCaDir(caPath)
            raise Exception(str(e))
        self.logger.debug("Successfully generated server gsql CA files.")

    def copyDWSCA(self, component):
        self.logger.debug("Copying %s CA files." % component)
        appPath = DefaultValue.getInstallDir(self.user)
        caPath = os.path.join(appPath, "share/sslcert", component)

        try:
            DefaultValue.localCreateCaDir(caPath)
            self.logger.debug("Successfully create %s CA file directory." % component)
            caSrcDir = os.path.join(DefaultValue.getEnv("GPHOME"), "lib/sslcert", component)
            certList = self.DWS_COPY_CERT_LIST[component]
            for certFile in certList:
                if certFile != "openssl.cnf":
                    caSrcFile = os.path.join(caSrcDir, "%s" % certFile)
                    caDstFile = os.path.join(caPath, "%s" % certFile)
                    g_file.cpFile(caSrcFile, caDstFile)
                    g_file.changeMode(DefaultValue.KEY_FILE_MODE, caDstFile)
            self.logger.debug("copied files to dest path.")
            # remove redundant files and directory
            DefaultValue.cleanServerCaDir(caPath)
            self.logger.debug("Successfully clean redundant files and directory.")
        except Exception as e:
            certFile = os.path.join(caPath, "demoCA/cacert.pem")
            if os.path.exists(certFile):
                g_file.removeFile(certFile)
            DefaultValue.cleanServerCaDir(caPath)
            raise Exception(str(e))
        self.logger.debug("Successfully copy %s CA files." % component)

    def createGdsCA(self, hostList=None):
        """
        function: create gds ca file
        input : NA
        output: NA
        """

        if hostList is None:
            hostList = []
        if (len(hostList) == 0):
            for dbNode in self.clusterInfo.dbNodes:
                hostList.append(dbNode.name)
        self.logger.debug("Generating gds CA files.")
        appPath = DefaultValue.getInstallDir(self.user)
        caPath = os.path.join(appPath, "share/sslcert/gds")
        self.logger.debug("The ca file dir is: %s." % caPath)

        try:
            # Create CA dir and prepare files for using.
            self.logger.debug("Create gds CA file directory.")
            DefaultValue.localCreateCaDir(caPath)
            self.logger.debug("Successfully create gds CA file directory.")
            # create cert
            randpass = DefaultValue.getRandStr()
            DefaultValue.createRootCA(randpass, caPath, self.logger)
            DefaultValue.createServerCA(DefaultValue.SERVER_CA, randpass, caPath, self.logger)
            DefaultValue.createClientCA(DefaultValue.Client_CA, randpass, caPath, self.logger)
            randpass = ""

            # Clean useless files.
            self.logger.debug("clean the dir of ca file：demoCA, server.req, client.req")
            DefaultValue.cleanServerCaDir(caPath)
            self.logger.debug("Successfully clean the dir of ca file：demoCA, server.req, client.req")

            #change permission of ca file to 600
            self.logger.debug("change permission of ca files to 600")
            for certFile in DefaultValue.GDS_CERT_LIST:
                chmod_file = os.path.join(caPath, "%s" % certFile)
                g_file.changeMode(DefaultValue.KEY_FILE_MODE, chmod_file)
            self.logger.debug("Successfully change permission of ca files to 600")
        except Exception as e:
            randpass = ""
            certFile = os.path.join(caPath, "demoCA/cacert.pem")
            if os.path.exists(certFile):
                g_file.removeFile(certFile)
            DefaultValue.cleanServerCaDir(caPath)
            raise Exception(str(e))
        self.logger.debug("Successfully generated server gds CA files.")

    def createGrpcCa(self, hostList=[], appPath=""):
        """
        function: create grpc ca file
        input : NA
        output: NA
        """
        self.logger.debug("Generating grpc CA files.")
        hosts = []
        if appPath == "":
            appPath = DefaultValue.getInstallDir(self.user)
        caPath = os.path.join(appPath, "share/sslcert/grpc")
        self.logger.debug("The ca file dir is: %s." % caPath)
        if (len(hostList) == 0):
            for dbNode in self.clusterInfo.dbNodes:
                hostList.append(dbNode.name)
        # Create CA dir and prepare files for using.
        self.logger.debug("Create CA file directory.")
        DefaultValue.createCADir(self.sshTool, caPath, hostList)
        self.logger.debug("Add hostname to config file.")
        configPath = os.path.join(appPath, "share/sslcert/grpc/openssl.cnf")
        self.logger.debug("The ca file dir is: %s." % caPath)
        # Add hostname to openssl.cnf file.
        DefaultValue.changeOpenSslConf(configPath, hostList)
        self.logger.debug("Generate CA files.")
        DefaultValue.createCA(DefaultValue.GRPC_CA, caPath)
        # Clean useless files, and change permission of ca file to 600.
        DefaultValue.cleanCaDir(caPath)
        self.logger.debug("Scp CA files to all nodes.")
        hosts = list(hostList)
        if DefaultValue.GetHostIpOrName() in hosts:
            hosts.remove(DefaultValue.GetHostIpOrName())
        for certFile in DefaultValue.GRPC_CERT_LIST:
            scpFile = os.path.join(caPath, "%s" % certFile)
            self.sshTool.scpFiles(scpFile, caPath, hosts)
        self.logger.debug("Successfully generated grpc CA files.")

    def checkUsedSpaceForDbUsers(self, nodegroup=""):
        """
        function:
            whether the space usage of the database users is greater than 80%.
            If the number of data records exceeds the threshold,
            the space for data redistribution will be insufficient.
        input : nodegroup string
        output: NA
        """
        self.checkIsElasticGroupExist(self.clusterInfo.dbNodes)
        if (self.isElasticGroup and nodegroup == ""):
            nodegroup = self.getGroupName("in_redistribution", "'y'")

        cooInst = self.getCooInst()
        sql = ('SET search_path = "public";select u.rolname, '
               '((s.used_space * 100.0) / s.total_space)::numeric(5, 2) '
               'AS used_space_perc '
               'from pg_catalog.pg_total_user_resource_info_oid s, '
               'pg_catalog.pg_roles u '
               'where s.userid=u.oid and s.total_space > 0 ')
        if (nodegroup != ""):
            sql += "and u.nodegroup = '%s'" % nodegroup
        sql += ";"
        self.logger.debug(
            "Sql statement for getting space usage for database users: %s." %
            sql)
        (status, output) = ClusterCommand.remoteSQLCommand(
             sql, self.user, cooInst[0].hostname,
             cooInst[0].port, ignoreError=False)
        if (status != 0):
            raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql +
                            " Error:\n%s" % str(output))
        if (output.strip() == ""):
            return
        lines = output.split('\n')
        detailInfo = ""
        for line in lines:
            resultList = line.split('|')
            if (len(resultList) != 2):
                continue
            (username, usedSpacePerc) = resultList
            if (float(usedSpacePerc) > SPACE_USAGE_DBUSER):
                detailInfo += "\n The space usage of user [%s] is %s." % (
                              username, usedSpacePerc)
        if (detailInfo != ""):
            raise Exception(ErrorCode.GAUSS_530["GAUSS_53011"] % (
                    "used space of the database users") +
                    " Used space percentage cannot exceed 80%%. Detail:%s" % (
                    detailInfo))

    def doBuildStandbyDN(self, nodeNames=[], isStartInst=True):
        """
        function: Use the gs_ctl tool to start master instances on all nodes, 
                 build the standby instance by full mode, and stop all master and standby instances.
        input : nodeNames:list
        output: NA
        """
        for dbNode in self.clusterInfo.dbNodes:
            for dnInst in dbNode.datanodes:
                if (self.clusterInfo.isSingleInstCluster() and
                    len(self.clusterInfo.getPeerInstance(dnInst)) != 1):
                    return
        self.logger.debug("Building the standby DN.")
        try:
            if (isStartInst):
                #start master dn
                cmd = "%s --manageDnInstance --actionType start --dnType master -U %s -l %s" % (
                                                                                        OMCommand.getLocalScript("Local_Query")
                                                                                        , self.user
                                                                                        , self.localLog)
                self.logger.debug("Command for starting master instances: %s" % cmd)
                self.sshTool.executeCommand(cmd, "start master DN", DefaultValue.SUCCESS, nodeNames)
            #Building standby dn
            cmd = "%s --manageDnInstance --actionType build --dnType standby -U %s -l %s" % (OMCommand.getLocalScript("Local_Query")
                                                                                    , self.user
                                                                                    , self.localLog)
            self.logger.debug("Command for building standby instances: %s" % cmd)
            self.sshTool.executeCommand(cmd, "build standby DN", DefaultValue.SUCCESS, nodeNames)
            
            if (isStartInst):
                #stop all dn 
                cmd = "%s --manageDnInstance --actionType stop --dnType all -U %s -l %s" % (
                                                                                        OMCommand.getLocalScript("Local_Query")
                                                                                        , self.user
                                                                                        , self.localLog)
                self.logger.debug("Command for stopping master and standby instances: %s" % cmd)
                self.sshTool.executeCommand(cmd, "stop master and standby DN", DefaultValue.SUCCESS, nodeNames)
        except Exception, e:
            try:
                if (isStartInst):
                    cmd = "%s --clean_install_process -U %s -l %s" % (OMCommand.getLocalScript("Local_Query"), self.user, self.localLog)
                    self.sshTool.executeCommand(cmd, "clean process", DefaultValue.SUCCESS, nodeNames)
            except Exception, e1:
                self.logger.debug("Failed to clean building process. ERROR:%s" % str(e1))
            raise Exception(str(e))
        self.logger.debug("Successfully built the standby DN.")

class Timeout(Exception): pass

