#!/usr/bin/env python
#-*- coding:utf-8 -*-
#############################################################################
#Copyright (c): 2012-2019, Huawei Tech. Co., Ltd.
#FileName     : ResourceControl.py
#Version      : Gauss200
#Date         : 2017-09-28
#Description  : ResourceControl.py is a utility to manage and control resource during data redistribution.
#############################################################################

try:
    import os
    import sys
    import commands
    import pwd
    import math
    import time
    import random

    sys.path.append(sys.path[0] + "/../../")
    from gspylib.common.Common import DefaultValue, ClusterCommand
    from gspylib.common.ErrorCode import ErrorCode

except ImportError, e:
    sys.exit("[GAUSS-52200] : Unable to import module: %s." % str(e))


class ResourceControl():

    """
    This class is used for resource control
    """

    # self.defaultuser will be set to the default database superuser which is always the os user
    REDIS_DEFAULT_USER_NAME = pwd.getpwuid(os.getuid()).pw_name



#############################################################################
# init a resource control instance
#############################################################################

    def __init__(self, redisInstance):
        """
        function: initialize the parameters for a resource control instance
        input : NA
        output: NA
        """

        self.redisInstanceForResCtrl = redisInstance

        self.redisuser = self.redisInstanceForResCtrl.redisuser
        self.redispassword = self.redisInstanceForResCtrl.redispasswd
        self.defaultuser = ResourceControl.REDIS_DEFAULT_USER_NAME

        self.reslevel = self.redisInstanceForResCtrl.resourcelevel
        self.action = self.redisInstanceForResCtrl.action
        self.redisCN = self.redisInstanceForResCtrl.redisNode

        # flags for objects in cgroup
        self.redisclassexist = False
        self.redisgrpexist = False

        # flags for objects in database
        self.redisuserexist = False
        self.redisrespoolexist = False
        self.redisrelation = False

        # in default setting, enableSeparationOfDuty is off
        self.separationofduty = False

        #Mark whether it is virtual cluster mode
        self.isElasticGroup = False
        #Indicates the corresponding nodegroup
        self.vcGroupName = ""


#############################################################################
# run a resource control instance
#############################################################################

    def run(self, vcMode = False, groupname = ""):
        """
        function: perform the whole resource control process
        input : NA
        output: NA
        """

        try:
            #init variable
            self.isElasticGroup = vcMode
            self.vcGroupName = groupname
            self.redisInstanceForResCtrl.logger.debug("message is :%d   %s" % (self.isElasticGroup, self.vcGroupName))
            # check validity of parameters
            if(self.action ==  "nodegroup_migrate"):
                self.defaultuser = 'rdsAdmin'
                self.resourceControlEnd()
                if self.reslevel == 'reset':
                    return
                self.resourceCtlPrePhase()
            self.resourceControlStart()

            if (self.action == "redistribute"):
                self.redisInstanceForResCtrl.logger.debug("The action is: %s" % self.action)
                # - If the command executed is "gs_expand -t residtribution",
                #   resource level is fixed to "f" (full).
                if (self.reslevel != 'f'):
                    raise Exception("The resource level for starting redistribution should be f.")
                else:
                    self.resourceControlForRedistribute()
            if (self.action == "resourcectl"):
                self.redisInstanceForResCtrl.logger.debug("The action is: %s" % self.action)
                self.resourceControlForResourcectl()
            if(self.action == "nodegroup_migrate"):
                self.redisInstanceForResCtrl.logger.debug("The action is: %s" % self.action)
                self.resourceControlForNodegroupMigrate()
            # get the resource pool setting level and log in logFile
            self.getResourcePoolLevel()
        except Exception, e:
            if (str(e).find("GAUSS-5") > 0):
                raise Exception(str(e))
            else:
                raise Exception(ErrorCode.GAUSS_516["GAUSS_51644"] + " Error:\n%s" % str(e))



#############################################################################
# abstract interfaces to achieve resource control functionality
#############################################################################

    def resourceControlStart(self):
        """
        function: check validity of parameters, contains:
                  1.check whether the action is valid
                  2.check whether the resouce level is valid
                  3.Check GUCs
                  4.Check whether GUC enableSeparationOfDuty is on
                  5.Check the existence of redisclass and redisgrp
                  6.Check the existence of redisuser
                  7.Check the existence of redisrespool
                  8.Check the association relation of redisuser and redisrespool
        input : NA
        output: NA
        """

        self.checkActionIsLegal()
        self.checkResourceLevelIsLegal()

        self.checkGUCsAreOn()
        self.checkGUCForSeparationOfDutyIsOn()

        self.checkIfRedisCgroupExists()

        self.checkIfRedisUserExists()
        self.checkIfRedisRespoolExists()
        self.checkRelationOfRedisuserAndRedisrespool()

    def resourceControlEnd(self):
        """
        function: clear up redisuser related info after redistribution succeeds:
                  1.Alter the owner of tables for statistic purpose
                  2.Revoke all schemas and tables of these schemas to redistribution user
                    if and only if the GUC enableSeparationOfDuty is on
                  3.Clear resource control related objects
        input : NA
        output: NA
        """

        # - Assume that buildTable is true, it means there will exist a pgxc_redistb
        #   table in each database.
        # - If you want to drop redisuser, you MUST drop these pgxc_redistb tables.
        # - This action will not be accordance with the meaning of buildTable parameter.

        # - Alter the owner of left tables which are for statistics purpose
        #   including redis_status, redis_progress_detail and redis_stage_detail.
        self.alterOwnerOfTablesForStatistics()

        # - This branch is to deal with the situation where enableSeparationOfDuty is on.
        # - More specifically, it revokes all schema relevant privileges after redistribution succeeds.
        if self.separationofduty:
            self.revokeAllSchemaPrivilegesToRedisuser()

        # Clear resource control related objects
        self.clearResourceControlObjects()

    def resourceCtlPrePhase(self):
        """
        function: entry for ResourceControl.run() methond when the parameter value after "-t" is "redistribute"
                  1.deal with redisclass, redisgrp
                  2.deal with redisuser, redisrespool
        input : NA
        output: NA
        """
        ## Part 1: deal with redisclass, redisgrp
        if not self.redisclassexist and not self.redisgrpexist:
            # prepare subclass group and workload group
            self.prepareResourceControlObjects('cgroup')
        elif self.redisclassexist and not self.redisgrpexist:
            # in this case, only prepare 'redisgrp'
            self.createWorkloadCgroup()
        elif self.redisclassexist and self.redisgrpexist:
            # no extra operations with respect to 'redisclass' and 'redisgrp' need to be conducted
            pass
        # - In the expected case, it is impossible that 'redisgrp' exists but 'redisclass' does not exist
        #   because 'redisgrp' depends on 'redisclass' in our design.
        # - So in this scenario, we are able to be sure that the current workload group 'redisgrp' in the cluster
        #   was not created by redistribution, instead, it already existed before. The reason is that it is not
        #   correctly associated to 'redisclass'.
        # - We deal with this issue by: exit and prompt a hint to the client, who will decide how to
        #   deal with this situation.
        else:
            raise Exception("The workload group for redistribution (redisgrp) has already existed but not associated to"
                            "the expected subclass group (redisclass). Please resolve the conflict at first by clear "
                            "existing 'redisgrp' and 'redisclass' because data redistribution needs 'redisgrp' to "
                            "associate to 'redisclass' so that it can finish redistribution task.")

        ## Part 2: deal with redisuser, redisrespool
        if not self.redisuserexist and not self.redisrespoolexist:
            # prepare resource pool and associate to the correct workload group
            self.prepareResourceControlObjects('resourcepool')

            # prepare redisuser and grant enough privileges to it
            self.prepareResourceControlObjects('redisuser')

            # this branch is to deal with the situation where enableSeparationOfDuty is on
            if self.separationofduty:
                self.grantAllSchemaPrivilegesToRedisuser()
        elif not self.redisuserexist and self.redisrespoolexist:
            # prepare redisuser
            self.prepareResourceControlObjects('redisuser')

            # this branch is to deal with the situation where enableSeparationOfDuty is on
            if self.separationofduty:
                self.grantAllSchemaPrivilegesToRedisuser()
        elif self.redisuserexist and self.redisrespoolexist:
            if not self.redisrelation:
                raise Exception("The two resource control objects for redistribution (redisuser and redisrespool) "
                                "are not correctly associated. Please resolve the conflict at first since data "
                                "redistribution needs the two objects to finish redis task.")
            # reset the (randomized) passwd of redisuser
            self.resetRedisUser()
            # this branch is to deal with the situation where enableSeparationOfDuty is on
            if self.separationofduty:
                self.grantAllSchemaPrivilegesToRedisuser()
        # - It is impossible that redisuser exists but redisrespool does not exist
        #   because redisuser depends on redisrespool in terms of resource control logic.
        # - So in this scenario, we are able to be sure that the current redisuser in the database
        #   was not created by redistribution, instead, it already existed before.
        # - We deal with this issue by: exit and prompt a hint to the client, who will decide how to
        #   deal with this situation.
        else:
            raise Exception("The database user for redistribution (redisuser) has already existed. Please resolve "
                            "the conflict at first since data redistribution needs redisuser to finish redis task.")

    def resourceControlForRedistribute(self):
        """
        function: entry for ResourceControl.run() methond when the parameter value after "-t" is "redistribute"
                  After prepare phase, change resource level to f
        input : NA
        output: NA
        """
        self.resourceCtlPrePhase()
        ##  After prepare phase, change resource level to f except the io_priority which is set to l.
        self.changeResourceLevelToF()

    def resourceControlForResourcectl(self):
        """
        function: entry for ResourceControl.run() methond when the parameter value after "-t" is "resourcectl"
        input : NA
        output: NA
        """

        # This branch is mainly used to detect the incorrect situation for resourcectl.
        if not self.redisuserexist or not self.redisrespoolexist or not self.redisrelation:
            message = "The resource control objects are not correctly established, detailed information is as follows:"
            if not self.redisuserexist:
                message += "\nThe special user (redisuser) for redistribution does not exist!"
            if not self.redisrespoolexist:
                message += "\nThe special resource pool (redisrespool) for redistribution does not exist!"
            if not self.redisrelation:
                message += "\nThe two objects (redisuser and redisrespool) are not correctly associated!"
            raise Exception(message)

        # 'f' means 'full resource level'
        if (self.reslevel == 'f'):
            self.changeResourceLevelToF()
        # 'h' means 'high resource level'
        elif (self.reslevel == 'h'):
            self.changeResourceLevelToH()
        # 'm' means 'mediate resource level'
        elif (self.reslevel == 'm'):
            self.changeResourceLevelToM()
        # 'l' means 'low resource level'
        elif (self.reslevel == 'l'):
            self.changeResourceLevelToL()
        else:
            raise Exception("Error in resource level, program will exit!")

    def resourceControlForNodegroupMigrate(self):
        """
        function: entry for ResourceControl.run() methond when the parameter value after "-t" is "nodegroup_migrate"
        input : NA
        output: NA
        """
        self.resourceControlForResourcectl()

    def checkActionIsLegal(self):
        """
        function: check whether the action is valid
        input : NA
        output: NA
        """

        try:
            if self.action not in ["redistribute", "resourcectl", "nodegroup_migrate"]:
                raise Exception("Parameter value of '-t' is incorrect! It must be redistribute or resourcectl")
        except Exception, e:
            raise Exception(str(e))

    def checkResourceLevelIsLegal(self):
        """
        function: check whether the resouce level is valid
        input : NA
        output: NA
        """

        if (self.reslevel not in ['f', 'l' ,'h', 'm']):
            raise Exception("parameter value of resource level is incorrect! It must be "
                            "f(full), l(low), h(high), m(medium)")

    def checkGUCForSeparationOfDutyIsOn(self):
        """
        function: Check whether GUC enableSeparationOfDuty is on
        input : NA
        output: NA
        """

        # check GUC enableSeparationOfDuty
        try:
            sql = "SHOW enableSeparationOfDuty;"
            self.redisInstanceForResCtrl.logger.debug("Sql command to check GUC enableSeparationOfDuty is: %s"
                                                      % sql)

            # execute command to alter redisuser
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (output == "on"):
                self.separationofduty = True

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to check GUC enableSeparationOfDuty failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))
        except Exception, e:
            raise Exception(str(e))

    def checkGUCsAreOn(self):
        """
        function: Check GUCs,
                  1.check GUC enable_memory_limit
                  2.check GUC bypass_workload_manager
                  3.check GUC enable_control_group
                  4.check GUC enable_backend_control
                  5.check GUC enable_vacuum_control
        input : NA
        output: NA
        """

        # check GUC enable_memory_limit
        try:
            sql = "SHOW enable_memory_limit;"
            self.redisInstanceForResCtrl.logger.debug("Sql command to check GUC enable_memory_limit: %s"
                                                      % sql)

            # execute command to alter redisuser
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to check GUC enable_memory_limit failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

            else:
                if (output != "on"):
                    if (self.action == "resourcectl"):
                        raise Exception("GUC parameter enable_memory_limit is not on! "
                               "Please turn it on first for resource control functionality.")

                    print ("Warning! GUC parameter enable_memory_limit is not on! "
                           "Can not use redistribute resource control! "
                           "Please turn it on first for resource control functionality.")
        except Exception, e:
            raise Exception(str(e))

        # check GUC bypass_workload_manager
        try:
            sql = "SHOW bypass_workload_manager;"
            self.redisInstanceForResCtrl.logger.debug("Sql command to check GUC bypass_workload_manager: %s"
                                                      % sql)

            # execute command to alter redisuser
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to check GUC bypass_workload_manager failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

            else:
                if (output != "on"):
                    raise Exception("GUC parameter bypass_workload_manager is not on! "
                           "Please turn it on first for resource control functionality.")
        except Exception, e:
            raise Exception(str(e))

        # check GUC enable_control_group
        try:
            sql = "SHOW enable_control_group;"
            self.redisInstanceForResCtrl.logger.debug("Sql command to check GUC enable_control_group: %s"
                                                      % sql)

            # execute command to alter redisuser
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to check GUC enable_control_group failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

            else:
                if (output != "on"):
                    raise Exception("GUC parameter enable_control_group is not on! "
                           "Please turn it on first for resource control functionality.")
        except Exception, e:
            raise Exception(str(e))

        # check GUC enable_backend_control
        try:
            sql = "SHOW enable_backend_control;"
            self.redisInstanceForResCtrl.logger.debug("Sql command to check GUC enable_backend_control: %s"
                                                      % sql)

            # execute command to alter redisuser
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to check GUC enable_backend_control failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

            else:
                if (output != "on"):
                    raise Exception("GUC parameter enable_backend_control is not on! "
                           "Please turn it on first for resource control functionality.")
        except Exception, e:
            raise Exception(str(e))

        # check GUC enable_vacuum_control
        try:
            sql = "SHOW enable_vacuum_control;"
            self.redisInstanceForResCtrl.logger.debug("Sql command to check GUC enable_vacuum_control: %s"
                                                      % sql)

            # execute command to alter redisuser
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to check GUC enable_vacuum_control failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

            else:
                if (output != "on"):
                    raise Exception("GUC parameter enable_vacuum_control is not on! "
                           "Please turn it on first for resource control functionality.")
        except Exception, e:
            raise Exception(str(e))

    def checkIfRedisCgroupExists(self):
        """
        function: Check the existence of redisclass and redisgrp,
                  1.get MpprcFile
                  2.get gausshome
                  3.form ssh command, currently we only check the cgroup info of the main host
                  4.execute the ssh command
        input : NA
        output: NA
        """

        try:
            #get MpprcFile
            sourceFile = DefaultValue.getMpprcFile()

            #get gausshome
            gaussHome = DefaultValue.getEnv("GAUSSHOME")
            if not gaussHome:
                return

            # form ssh command
            # currently we only check the cgroup info of the main host
            # Physical cluster mode
            if not self.isElasticGroup:
                sshcmd = "source %s;%s/bin/gs_cgroup -p" % (sourceFile, gaussHome)
            else:
                sshcmd = "source %s;%s/bin/gs_cgroup -p -N %s" % (sourceFile, gaussHome, self.vcGroupName)
            self.redisInstanceForResCtrl.logger.debug("Command to obtain the cgroup information: %s"
                                                      % sshcmd)

            # execute the ssh command
            (status, output) = commands.getstatusoutput(sshcmd)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Command to obtain the cgroup information failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sshcmd + " Error:\n%s" % str(output))
            else:
                # - We can split the output string into a list by whitespace
                #   and then deal with it as a python list cgroupinfo = output.split(" ").
                # - But here string method 'in' is better since redis workload group always has a level and a colon
                #   after its name.
                # - It is also possible to import re module and use re.split() with multiple delimiters including both
                #   whitespace and colon.
                if ('redisclass' in output):
                    self.redisclassexist = True
                # - Note that in the output information of 'gs_cgroup -p', redisgrp is followed by its level.
                # - In the common case, it is 'redisgrp:2'.
                if ('redisgrp' in output):
                    self.redisgrpexist = True

        except Exception, e:
            raise Exception(str(e))

    def checkIfRedisUserExists(self):
        """
        function: Check the existence of redisuser
        input : NA
        output: NA
        """

        try:
            sql = "SELECT COUNT(*) FROM PG_USER WHERE USENAME = 'redisuser';"
            self.redisInstanceForResCtrl.logger.debug("Sql command to check if redisuser exists: %s"
                                                      % sql)

            # execute command to check the existence of redisuser
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (output == '1' ):
                self.redisuserexist = True

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Check the existence of redisuser failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))
        except Exception, e:
            raise Exception(str(e))

    def checkIfRedisRespoolExists(self):
        """
        function: Check the existence of redisrespool
        input : NA
        output: NA
        """

        try:
            sql = "SELECT COUNT(*) FROM PG_RESOURCE_POOL WHERE RESPOOL_NAME = 'redisrespool';"
            self.redisInstanceForResCtrl.logger.debug("Sql command to check if redisrespool exists: %s"
                                                      % sql)

            # execute command to check the existence of redisrespool
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)
            if (output == '1'):
                self.redisrespoolexist = True

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Check the existence of redisrespool failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))
        except Exception, e:
            raise Exception(str(e))

    def checkRelationOfRedisuserAndRedisrespool(self):
        """
        function: Check the association relation of redisuser and redisrespool
        input : NA
        output: NA
        """

        try:
            sql = "SELECT COUNT(*) FROM PG_USER WHERE USENAME = 'redisuser' AND RESPOOL = 'redisrespool';"
            self.redisInstanceForResCtrl.logger.debug("Sql command to check the relation between redisuser and"
                                                      " redisrespool: %s" % sql)

            # execute command to check the existence of redisrespool
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)
            if (output == '1'):
                self.redisrelation = True

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Check the relation between redisuser and redisrespool failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))
        except Exception, e:
            raise Exception(str(e))

    def changeResourceLevelToF(self):
        """
        function: change Resource Level To F:
                1.Alter memory percentage "mem_percent" to 100%
                2.Alter workload group to CPU resource level "f"
                3.alter the resource pool to limit the io to level f
                4.alter the resource pool to limit the io_priority f when action is redistribute 
                5.alter the resource pool to limit the io_priority f when action is resourcectl 
        input : NA
        output: NA
        """
        self.alterResourcePoolChangeMemPercentToLevelF()
        self.alterWorkloadCgroupToLevelF()
        # preserved interface, waiting for IO control patch -- It is enabled in version V100R007C10B011, after IO resource
        # control functionality had been merged and stable.
        self.alterResourcePoolChangeIOLimitToLevelF()
        if (self.action == "redistribute"):
            self.alterResourcePoolChangeIOPriorityToLevelL()
        if (self.action == "resourcectl"):
            self.alterResourcePoolChangeIOPriorityToLevelF()

    def changeResourceLevelToH(self):
        """
        function: change Resource Level To H:
                  1.Alter memory percentage "mem_percent" to 50%
                  2.Alter workload group to CPU resource level "h"
                  3.Alter the resource pool to limit the io_priority h
        input : NA
        output: NA
        """
        self.alterResourcePoolChangeMemPercentToLevelH()
        self.alterWorkloadCgroupToLevelH()
        # preserved interface, waiting for IO control patch
        self.alterResourcePoolChangeIOPriorityToLevelH()

    def changeResourceLevelToM(self):
        """
        function: change Resource Level To M:
                  1.Alter memory percentage "mem_percent" to 30%
                  2.Alter workload group to CPU resource level "m"
                  3.Alter the resource pool to limit the io_priority m
        input : NA
        output: NA
        """
        self.alterResourcePoolChangeMemPercentToLevelM()
        self.alterWorkloadCgroupToLevelM()
        # preserved interface, waiting for IO control patch
        self.alterResourcePoolChangeIOPriorityToLevelM()

    def changeResourceLevelToL(self):
        """
        function: change Resource Level To L:
                  1.Alter memory percentage "mem_percent" to 10%
                  2.Alter workload group to CPU resource level "l"
                  3.alter the resource pool to limit the io to level l
        input : NA
        output: NA
        """
        self.alterResourcePoolChangeMemPercentToLevelL()
        self.alterWorkloadCgroupToLevelL()
        # preserved interface, waiting for IO control patch -- It is enabled in version V100R007C10B011, after IO resource
        # control functionality had been merged and stable.
        self.alterResourcePoolChangeIOLimitToLevelL()
        self.alterResourcePoolChangeIOPriorityToLevelL()

    def prepareResourceControlObjects(self, objType):
        """
        function: Prepare resource control related objects according to the type of objects.
                  More specifically, the input parameter objType has three values:
                  - cgroup: this value means that next the program will prepare cgroup objects including a subclass group
                            called 'redisclass' and a workload group called 'redisgrp'. Besides, 'redisgrp' is associated
                            to 'redisclass'.
                  - resourcepool: this value means that next the program will prepare a resource pool called 'redisrespool'
                                  for redistribution.
                  - redisuser: this value means that next the program will prepare a database user (with enough privilege)
                                called 'redisrespool' for redistribution.
        input : objType
        output: NA
        """

        # check validity of objType
        if (objType not in ['cgroup','resourcepool','redisuser']):
            raise Exception("The value of resource control objType (%s) is not legal. "
                                                    "It should be 'cgroup' or 'resourcepool' or 'redisuser'."
                                                    % objType)

        # prepare resource control objects in terms of objType
        if ('cgroup' == objType):
            self.prepareCgroup()
        elif ('resourcepool' == objType):
            self.prepareResourcePool()
        elif ('redisuser' == objType):
            self.prepareRedisUser()
        else:
            raise Exception("The type (%s) of resource control objects need to be prepared is incorrect!" % objType)



    def clearResourceControlObjects(self):
        """
        function: Clear resource control related objects:
                  1.Drop the user redisuser for redistribution
                  2.Drop the resource pool redisrespool
                  3.drop the workload group
                  4.drop the subclass resource group for redistribution
        input : NA
        output: NA
        """

        self.dropRedisUser()
        # - If a certain role is associated with a resource pool,
        #   the pool can not be dropped.
        # - So we should drop redisuser at first.
        self.dropResourcePool()
        self.dropWorkloadCgroup()
        self.dropSubclassCgroup()



#############################################################################
# concrete interfaces to achieve resource control functionality
#############################################################################

    def isGroupClassExist(self, name):
        try:
            #get MpprcFile
            sourceFile = DefaultValue.getMpprcFile()

            #get gausshome
            gaussHome = DefaultValue.getEnv("GAUSSHOME")
            if not gaussHome:
                self.redisInstanceForResCtrl.logger.log("Error:get GAUSSHOME")
                raise Exception(ErrorCode.GAUSS_518["GAUSS_51807"] + " Error:get GAUSSHOME")

            # form ssh command
            # currently we only check the cgroup info of the main host
            #Virtual cluster mode
            if not self.isElasticGroup:
                # form ssh command
                sshcmd = "source %s;%s/bin/gs_cgroup -p" % (sourceFile, gaussHome)
            else:
                sshcmd = "source %s;%s/bin/gs_cgroup -p -N %s" % (sourceFile, gaussHome, self.vcGroupName)
            self.redisInstanceForResCtrl.logger.debug("Command to get the cgroup information: %s"
                                                      % sshcmd)

            # execute the ssh command
            (status, output) = commands.getstatusoutput(sshcmd)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Command to get the cgroup %s information failed." % (name))
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sshcmd + " Error:get %s" % str(name))
            else:
                if (name in output):
                    return True
                else:
                    return False

        except Exception, e:
            self.redisInstanceForResCtrl.logger.logExit(str(e))

    def getGroupClassConf(self, name):
        try:
            #get MpprcFile
            sourceFile = DefaultValue.getMpprcFile()

            #get gausshome
            gaussHome = DefaultValue.getEnv("GAUSSHOME")
            if not gaussHome:
                self.redisInstanceForResCtrl.logger.log("Error:get GAUSSHOME")
                raise Exception(ErrorCode.GAUSS_518["GAUSS_51807"] + " Error:get GAUSSHOME")

            # form ssh command
            # currently we only check the cgroup info of the main host
            #Virtual cluster mode
            if not self.isElasticGroup:
                # form ssh command
                sshcmd = "source %s;%s/bin/gs_cgroup -p | grep -E '\<Name: %s\>'" % (sourceFile, gaussHome, name)
            else:
                sshcmd = "source %s;%s/bin/gs_cgroup -p -N %s | grep -E '\<Name: %s\>'" % (sourceFile, gaussHome, self.vcGroupName, name)
            self.redisInstanceForResCtrl.logger.debug("Command to get the cgroup information: %s"
                                                      % sshcmd)

            # execute the ssh command
            (status, output) = commands.getstatusoutput(sshcmd)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Command to get the cgroup %s information failed." % (name))
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sshcmd + " Error:get %s" % str(name))
            else:
                if (name in output):
                    for line in output.splitlines():
                        if (line.find(name) < 0):
                            continue
                        else:
                            items = line.replace('(', '#')
                            items = items.replace(')', '#')
                            items = items.split('#')
                            for item in items:
                                item = item.strip()
                                if (str(item).isdigit()):
                                    return int(item.strip())
                            self.redisInstanceForResCtrl.logger.log("Command to get the cgroup %s Percent failed." % (name))
                            raise Exception(ErrorCode.GAUSS_517["GAUSS_51706"] + " Error:get %s" % str(name))
                else:
                    self.redisInstanceForResCtrl.logger.log("Command to get the cgroup %s is not exist. failed." % (name))
                    raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sshcmd + " Error:get %s" % str(name))

        except Exception, e:
            self.redisInstanceForResCtrl.logger.logExit(str(e))

    def prepareCgroup(self):
        """
        function: Create a user and grant enough privilege to her for redistribution:
                  1.create a subclass resource group for redistribution
                  2.create the workload group
        input : NA
        output: NA
        """

        # prepare subclass group and workload group
        self.createSubclassCgroup()
        self.createWorkloadCgroup()

    def prepareResourcePool(self):
        """
        function: Create the resource pool redisrespool for resource control in redistribution
        input : NA
        output: NA
        """

        try:
            # sql for a transaction block
            START_TRANSACTION_SQL = "START TRANSACTION;"
            END_TRANSACTION_SQL = "COMMIT;"

            # sql to create the resource pool
            # set ACTIVE_STATEMENTS=-1 to avoid potential high concurrency issue
            # Physical cluster mode
            if not self.isElasticGroup:
                # Create a default resource pool
                CREATE_POOL_SQL = "CREATE RESOURCE POOL redisrespool with (ACTIVE_STATEMENTS=-1);"
                # sql to associate redisrespool to workload group
                ASSOCIATE_TO_WKLGROUP_SQL = "ALTER RESOURCE POOL redisrespool WITH (control_group='redisclass:redisgrp');"
                # combine them to form the final sql
                sql = "%s%s%s%s" % (START_TRANSACTION_SQL, CREATE_POOL_SQL, ASSOCIATE_TO_WKLGROUP_SQL, END_TRANSACTION_SQL)
            else:
                # sql to associate redisrespool to workload group
                CREATE_POOL_SQL = "CREATE RESOURCE POOL redisrespool with (control_group='redisclass:redisgrp', nodegroup= '%s');" % self.vcGroupName
                # combine them to form the final sql
                sql = "%s%s%s" % (START_TRANSACTION_SQL, CREATE_POOL_SQL, END_TRANSACTION_SQL)

            self.redisInstanceForResCtrl.logger.debug("Sql command to create the resource pool for redistribution is: %s"
                                                      % sql)

            # execute command to create resource pool
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Create the resource pool for redistribution failed." )
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

            # - A potential problem is that either this combined sql succeeds or fails, the program will continue.
            # - This is not the correct case in which the resource pool has not been established.
            # - So after executing the command, we further check whether resource pool exists, if not, the program will
            #   exit with an error.
            self.checkIfRedisRespoolExists()
            if not self.redisrespoolexist:
                raise Exception("Create the resource pool for redistribution failed. Please check:\n"
                                                        "1.Cgroup class group 'redisclass' and workload group 'redisgrp' are "
                                                        "correct in each node.\n"
                                                        "2.If the resource pool name 'redisrespool' already exists in the database.")

        except Exception, e:
            raise Exception(str(e))

    def prepareRedisUser(self):
        """
        function: Create a user and grant enough privilege to her for redistribution
        input : NA
        output: NA
        """

        try:
            # sql for a transaction block
            START_TRANSACTION_SQL = "START TRANSACTION;"
            END_TRANSACTION_SQL = "COMMIT;"

            # sql to create the redisuser in physical cluster mode
            CREATE_USER_SQL = "CREATE USER redisuser PASSWORD '%s';" % self.redispassword
            # sql to create the redisuser in virtual cluster mode
            # NOTICE: The execution of the 'create role' command is inconsistent in the two system tables pg_authid and pg_user.
            #         The execution of the 'create user' command is consistent in the pg_authid and pg_user system tables.
            CREATE_USER_VC_MODE_SQL = "CREATE USER redisuser node group \"%s\" PASSWORD '%s';" % (self.vcGroupName, self.redispassword)
            # this is for protecting sensitive information
            CREATE_USER_SQL_WITHOUT_PASSWD = "CREATE USER redisuser PASSWORD 'xxx';"
            CREATE_USER_VC_MODE_NO_PASSWD = "CREATE USER redisuser node group \"%s\" PASSWORD 'xxx';" % self.vcGroupName

            # sql to drop the schema with the same name of redisuser
            # the reason for this operation is:
            # - This is because if we do not drop this schema, some tables for statistic purpose in database postgres
            #   will be created on schema 'redisuser', which will incur problem in dropping resource control objects
            #   after redistribution succeeds.
            # - After this step, all the tables in DB postgres will be created on schema 'public' or 'data_redis'.
            DROP_SCHEMA_SQL = "DROP SCHEMA IF EXISTS redisuser;"

            # sql to associate redisuser to redisrespool
            ASSOCIATE_TO_POOL_SQL = "ALTER USER redisuser WITH RESOURCE POOL 'redisrespool';"

            # sql to grant all privilege to redisuser
            GRANT_ALL_PRIVILEGE_SQL = "GRANT ALL PRIVILEGE TO redisuser;"

            # Physical cluster mode
            if not self.isElasticGroup:
                # combine them to form the final sql
                sql = "%s%s%s%s%s%s" % (START_TRANSACTION_SQL, CREATE_USER_SQL, DROP_SCHEMA_SQL, ASSOCIATE_TO_POOL_SQL,
                                        GRANT_ALL_PRIVILEGE_SQL, END_TRANSACTION_SQL)

                # sqlforlog is only used to prevent from printing plaintext password message in the log
                sqlforlog = "%s%s%s%s%s%s" % (START_TRANSACTION_SQL, CREATE_USER_SQL_WITHOUT_PASSWD, DROP_SCHEMA_SQL, ASSOCIATE_TO_POOL_SQL,
                                              GRANT_ALL_PRIVILEGE_SQL, END_TRANSACTION_SQL)
            else:
                # combine them to form the final sql
                sql = "%s%s%s%s%s%s" % (START_TRANSACTION_SQL, CREATE_USER_VC_MODE_SQL, DROP_SCHEMA_SQL, ASSOCIATE_TO_POOL_SQL,
                                        GRANT_ALL_PRIVILEGE_SQL, END_TRANSACTION_SQL)

                # sqlforlog is only used to prevent from printing plaintext password message in the log
                sqlforlog = "%s%s%s%s%s%s" % (START_TRANSACTION_SQL, CREATE_USER_VC_MODE_NO_PASSWD, DROP_SCHEMA_SQL, ASSOCIATE_TO_POOL_SQL,
                                              GRANT_ALL_PRIVILEGE_SQL, END_TRANSACTION_SQL)

            self.redisInstanceForResCtrl.logger.debug("Sql command to create the special user to conduct redistribution: %s"
                                                      % sqlforlog)

            # execute command to create redisuser
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Create user %s for redistribution failed." % self.redisuser)
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sqlforlog + " Error:\n%s" % str(output))

            # - A potential problem is that either this combined sql succeeds or fails, the program will continue.
            # - This is not the correct case in which the redistribution user has not been established.
            # - So after executing the command, we further check whether redistribution user exists, if not, the program will
            #   exit with an error.
            self.checkIfRedisUserExists()
            if not self.redisuserexist:
                raise Exception("Create the user 'redisuser' for redistribution failed. Please check:\n"
                                                        "1.If the resource pool 'redisrespool' to which the user "
                                                        "will associate does not exist.\n"
                                                        "2.If the user name 'redisuser' already exists in the database.")

        except Exception, e:
            raise Exception(str(e))

    def getRandomPasswd(self):
        """
        function: get random passwd
        input: NA
        output: passwd
        """
        strupper = ''.join(random.sample('ABCDEFGHIJKLMNOPQRSTUVWXYZ', 3))
        strlower = ''.join(random.sample('abcdefghijklmnopqrstuvwxyz', 4))
        strnum = ''.join(random.sample('0123456789', 3))
        strspecial = ''.join(random.sample('~@_#*', 1))
        return "G" + strupper + strlower + strnum + strspecial

    def resetRedisUser(self):
        """
        function: - Reset the password of the redisuser for redistribution.
                  - This method is needed for the reentry of redistribution, because after we discard the -W parameter, we
                    should reset the randomized password to be the newly generated one.
        input : NA
        output: NA
        """

        try:
            if (self.action == "nodegroup_migrate"):
                self.defaultuser = 'rdsAdmin'
                self.redispassword = self.getRandomPasswd()
            sql = "ALTER USER redisuser IDENTIFIED BY '%s';" % self.redispassword
            self.redisInstanceForResCtrl.logger.debug("Resetting password of the special redis user.")

            # execute command to drop redisuser
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("ALTER the random password of user %s for "
                                                        "redistribution failed." % self.redisuser)
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def dropRedisUser(self):
        """
        function: Drop the user redisuser for redistribution
        input : NA
        output: NA
        """

        try:
            sql = "DROP USER IF EXISTS redisuser CASCADE;"
            self.redisInstanceForResCtrl.logger.debug("Sql command to drop the special user is: %s"
                                                      % sql)

            # execute command to drop redisuser
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)
            
            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Drop user %s for redistribution failed." % self.redisuser)
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def grantAllSchemaPrivilegesToRedisuser(self):
        """
        function: Grant all schemas and tables of these schemas to redistribution user if and only if
                  the GUC enableSeparationOfDuty is on!!!
                  This is because in this mode, even sysadmin is not enough for redisuser to conduct redistribution.
        input : NA
        output: NA
        """

        # define some essential variables
        schemalist = []
        dbname = ""
        schemaname = ""

        # obtain the non-template databases
        dbnamelist = self.obtainDatabaseInfo()
        self.redisInstanceForResCtrl.logger.debug("The database list is: %s" % dbnamelist)

        for dbname in dbnamelist:
            self.redisInstanceForResCtrl.logger.debug("The current working database is: %s" % dbname)
            # obtain the non-system schemas
            schemalist = self.obtainSchemaInfoPerDatabase(dbname)
            self.redisInstanceForResCtrl.logger.debug("The schema list in this database is: %s" % schemalist)
            for schemaname in schemalist:
                self.redisInstanceForResCtrl.logger.debug("The current working schema is: %s" % schemaname)
                self.grantTablesAndSchemaToRedisuserPerSchema(dbname, schemaname)

    def grantTablesAndSchemaToRedisuserPerSchema(self, dbname, schemaname):
        """
        function: Grant all privileges on the input schema and the tables belonging to this schema to redistribution user
        input : NA
        output: NA
        """

        # check the validity of input parameter dbname
        if (dbname == ""):
            raise Exception("The input database name should not be an empty string!")

        # check the validity of input parameter schemaname
        if (schemaname == ""):
            self.redisInstanceForResCtrl.logger.debug("There is no non-system schema in database %s." % dbname)
            return

        try:
            # sql to grant related tables and schema privileges to the user
            # - Notice that GRANT operations can be repeatedly executed, thus there is no need for them to use transaction protection
            #   like "START TRANSACTION;GRANT ...;END;"
            sql = "GRANT ALL PRIVILEGES ON SCHEMA %s TO redisuser;" \
                  "GRANT ALL PRIVILEGES ON ALL TABLES IN SCHEMA %s TO redisuser;" % (schemaname, schemaname)
            self.redisInstanceForResCtrl.logger.debug("Sql command to grant ALL TABLES and the corresponding schema "
                                                      "to redisuser is: %s"
                                                      % sql)

            # execute command to grant all the required privileges to redisuser
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               True,
                                                               dbname,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to grant ALL TABLES and the corresponding schema "
                                                      "to redisuser failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def revokeAllSchemaPrivilegesToRedisuser(self):
        """
        function: Revoke all schemas and tables of these schemas to redistribution user if and only if
                  the GUC enableSeparationOfDuty is on!!!
                  This is because in this mode, when you try to clear resource control objects after redistribution
                  succeeds, there will be privilege problem due to previous grant privilege operations in method
                  grantAllSchemaPrivilegesToRedisuser().
        input : NA
        output: NA
        """

        # define some essential variables
        schemalist = []
        dbname = ""
        schemaname = ""

        # obtain the non-template databases
        dbnamelist = self.obtainDatabaseInfo()
        self.redisInstanceForResCtrl.logger.debug("The database list is: %s" % dbnamelist)

        for dbname in dbnamelist:
            self.redisInstanceForResCtrl.logger.debug("The current working database is: %s" % dbname)
            # obtain the non-system schemas
            schemalist = self.obtainSchemaInfoPerDatabase(dbname)
            self.redisInstanceForResCtrl.logger.debug("The schema list in this database is: %s" % schemalist)
            for schemaname in schemalist:
                self.redisInstanceForResCtrl.logger.debug("The current working schema is: %s" % schemaname)
                self.revokeTablesAndSchemaToRedisuserPerSchema(dbname, schemaname)

    def revokeTablesAndSchemaToRedisuserPerSchema(self, dbname, schemaname):
        """
        function: Revoke all privileges on the input schema and the tables belonging to this schema to redistribution user
        input : NA
        output: NA
        """

        # check the validity of input parameter dbname
        if (dbname == ""):
            raise Exception("The input database name should not be an empty string!")

        # check the validity of input parameter schemaname
        if (schemaname == ""):
            self.redisInstanceForResCtrl.logger.debug("There is no non-system schema in database %s." % dbname)
            return

        try:
            # sql to revoke related tables and schema privileges to the user
            sql = "REVOKE ALL PRIVILEGES ON SCHEMA %s FROM redisuser CASCADE;" \
                  "REVOKE ALL PRIVILEGES ON ALL TABLES IN SCHEMA %s FROM redisuser CASCADE;" % (schemaname, schemaname)
            self.redisInstanceForResCtrl.logger.debug("Sql command to revoke ALL TABLES and the corresponding schema "
                                                      "to redisuser is: %s"
                                                      % sql)

            # execute command to revoke all the required privileges to redisuser
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               True,
                                                               dbname,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to revoke ALL TABLES and the corresponding schema "
                                                      "to redisuser failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def obtainDatabaseInfo(self):
        """
        function: To obtain the database information of the cluster, except template database template0 and template1
        input : NA
        output: NA
        """

        try:
            # sql to obtain the DBs in the cluster
            sql = "SELECT datname FROM pg_database WHERE datname NOT IN ('template0', 'template1');"
            self.redisInstanceForResCtrl.logger.debug("Sql command to obtain the database information is: %s"
                                                      % sql)

            # execute command to issue the select SQL statement
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to obtain the database information failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

            dbnamelist = []

            # use the split method of string to transform the output string to a list
            dbnamelist = output.split("\n")

        except Exception, e:
            raise Exception(str(e))

        return dbnamelist

    def obtainSchemaInfoPerDatabase(self, dbname):
        """
        function: To obtain the non-system schema information of the input database
        input : NA
        output: NA
        """

        # check the validity of input parameter
        if (dbname == ""):
            raise Exception("The input database name should not be an empty string!")

        try:
            # sql to obtain the schemas in the input database
            sql = "SELECT nspname FROM pg_namespace WHERE nspname NOT IN ('cstore', 'dbe_job', 'dbe_lob', " \
                  "'dbe_output', 'dbe_random', 'information_schema', 'pg_catalog', 'pg_toast', 'public'," \
                  " 'sys', 'dbe_raw', 'dbe_file');"
            self.redisInstanceForResCtrl.logger.debug("Sql command to obtain the schema information in "
                                                      "database %s is: %s"
                                                      % (dbname, sql))

            # execute command to issue the select SQL statement
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               True,
                                                               dbname,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to obtain the schema information failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

            schemalist = []

            # use the split method of string to transform the output string to a list
            schemalist = output.split("\n")

        except Exception, e:
            raise Exception(str(e))

        return schemalist

    def alterOwnerOfTablesForStatistics(self):
        """
        function: Alter the owner of tables for statistic purpose
        input : NA
        output: NA
        """

        try:
            sql = "ALTER TABLE IF EXISTS redis_status OWNER TO \"%s\";" % self.defaultuser
            self.redisInstanceForResCtrl.logger.debug("Sql command to alter the owner of table redis_status: %s"
                                                      % sql)

            # execute command to alter the owner of statistics table redis_status
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to alter the owner of table redis_status failed." )
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

        try:
            sql = "ALTER TABLE IF EXISTS redis_progress_detail OWNER TO \"%s\";" % self.defaultuser
            self.redisInstanceForResCtrl.logger.debug("Sql command to alter the owner of table redis_progress_detail: %s"
                                                      % sql)

            # execute command to alter the owner of statistics table redis_progress_detail
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to alter the owner of table redis_progress_detail failed." )
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

        try:
            sql = "ALTER TABLE IF EXISTS redis_stage_detail OWNER TO \"%s\";" % self.defaultuser
            self.redisInstanceForResCtrl.logger.debug("Sql command to alter the owner of table redis_stage_detail: %s"
                                                      % sql)

            # execute command to alter the owner of statistics table redis_stage_detail
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to alter the owner of table redis_stage_detail failed." )
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def dropResourcePool(self):
        """
        function: Drop the resource pool redisrespool
        input : NA
        output: NA
        """

        try:
            sql = "DROP RESOURCE POOL IF EXISTS redisrespool;"
            self.redisInstanceForResCtrl.logger.debug("Sql command to drop the resource pool: %s"
                                                      % sql)

            # execute command to drop a resource pool
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Drop resource pool for redistribution failed." )
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def alterResourcePoolChangeMemPercentToLevelF(self):
        """
        function: Alter memory percentage "mem_percent" to 100%
        input : NA
        output: NA
        """

        try:
            # level "f" means 100% resource
            sql = "ALTER RESOURCE POOL redisrespool WITH (mem_percent = 0);"
            self.redisInstanceForResCtrl.logger.debug("Sql command to alter the resource pool is: %s"
                                                      % sql)

            # execute command to alter the memory level of the resource pool
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to alter the resource pool failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def alterResourcePoolChangeMemPercentToLevelH(self):
        """
        function: Alter memory percentage "mem_percent" to 50%
        input : NA
        output: NA
        """

        try:
            # level "h" means 50% resource
            sql = "ALTER RESOURCE POOL redisrespool WITH (mem_percent = 50);"
            self.redisInstanceForResCtrl.logger.debug("Sql command to alter the resource pool is: %s"
                                                      % sql)

            # execute command to alter the memory level of the resource pool
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to alter the resource pool failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def alterResourcePoolChangeMemPercentToLevelM(self):
        """
        function: Alter memory percentage "mem_percent" to 30%
        input : NA
        output: NA
        """

        try:
            # level "m" means 30% resource
            sql = "ALTER RESOURCE POOL redisrespool WITH (mem_percent = 30);"
            self.redisInstanceForResCtrl.logger.debug("Sql command to alter the resource pool is: %s"
                                                      % sql)

            # execute command to alter the memory level of the resource pool
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to alter the resource pool failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def alterResourcePoolChangeMemPercentToLevelL(self):
        """
        function: Alter memory percentage "mem_percent" to 10%
        input : NA
        output: NA
        """

        try:
            # level "l" means 10% resource
            sql = "ALTER RESOURCE POOL redisrespool WITH (mem_percent = 10);"
            self.redisInstanceForResCtrl.logger.debug("Sql command to alter the resource pool is: %s"
                                                      % sql)

            # execute command to alter the memory level of the resource pool
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to alter the resource pool failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def createSubclassCgroup(self):
        """
        function: create a subclass resource group for redistribution:
                  1.get MpprcFile
                  2.get gausshome
                  3.form ssh command
                  4.execute the ssh command
        input : NA
        output: NA
        """

        try:
            # - the default class will occupy 20% resource,
            # - so the max percentage for redisclass is 80%, not 99%

            #get MpprcFile
            sourceFile = DefaultValue.getMpprcFile()

            #get gausshome
            gaussHome = DefaultValue.getEnv("GAUSSHOME")
            if not gaussHome:
                return

            for times in range(int(random.sample('23456789', 1)[0])):
                time.sleep(2)
                ret = self.isGroupClassExist("redisclass")
                if ret:
                    self.redisInstanceForResCtrl.logger.log("Warning! Command to get the cgroup redisclass is exist.")
                    return

            defaultClassConf = self.getGroupClassConf("DefaultClass")
            if (defaultClassConf < 0):
                self.redisInstanceForResCtrl.logger.log("Command to get the cgroup DefaultClass is not reasonable value. failed.")
                raise Exception(ErrorCode.GAUSS_517["GAUSS_51706"] + " Error:get DefaultClass")
            if (defaultClassConf < 2):
                self.redisInstanceForResCtrl.logger.log("Command to get the defaultClassConf information failed. is low 2.")
                raise Exception(ErrorCode.GAUSS_517["GAUSS_51706"] + " Error:get DefaultClass")
            redisClassConf = math.floor(defaultClassConf / 2)
            defaultClassConf = defaultClassConf - redisClassConf

            for times in range(int(random.sample('23456789', 1)[0])):
                time.sleep(2)
                ret = self.isGroupClassExist("redisclass")
                if ret:
                    self.redisInstanceForResCtrl.logger.log("Warning! Command to get the cgroup redisclass is exist.")
                    return

            #Virtual cluster mode
            if not self.isElasticGroup:
                # form ssh command
                sshcmd = "source %s; gs_ssh -c \"%s/bin/gs_cgroup -u -S DefaultClass -s %d\"" % (sourceFile, gaussHome, defaultClassConf)
            else:
                sshcmd = "source %s; gs_ssh -c \"%s/bin/gs_cgroup -u -S DefaultClass -s %d -N %s\"" % (sourceFile, gaussHome, defaultClassConf, self.vcGroupName)
            self.redisInstanceForResCtrl.logger.debug("Command to create subclass group defaultClassConf: %s"
                                                      % sshcmd)

            # execute the ssh command
            (status, output) = commands.getstatusoutput(sshcmd)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Command to create defaultClassConf group redisclass failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sshcmd + " Error:\n%s" % str(output))

            #Virtual cluster mode
            if not self.isElasticGroup:
                # form ssh command
                sshcmd = "source %s; gs_ssh -c \"%s/bin/gs_cgroup -c -S redisclass -s %d\"" % (sourceFile, gaussHome, redisClassConf)
            else:
                sshcmd = "source %s; gs_ssh -c \"%s/bin/gs_cgroup -c -S redisclass -s %d -N %s\"" % (sourceFile, gaussHome, redisClassConf, self.vcGroupName)
            self.redisInstanceForResCtrl.logger.debug("Command to create subclass group redisclass: %s"
                                                      % sshcmd)

            # execute the ssh command
            (status, output) = commands.getstatusoutput(sshcmd)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Command to create subclass group redisclass failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sshcmd + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def dropSubclassCgroup(self):
        """
        function: drop the subclass resource group for redistribution:
                  1.get MpprcFile
                  2.get gausshome
                  3.form ssh command
                  4.execute the ssh command
        input : NA
        output: NA
        """

        try:
            #get MpprcFile
            sourceFile = DefaultValue.getMpprcFile()

            #get gausshome
            gaussHome = DefaultValue.getEnv("GAUSSHOME")
            if not gaussHome:
                return

            ret = self.isGroupClassExist("redisclass")
            if not ret:
                self.redisInstanceForResCtrl.logger.log("Warning! Command to get the cgroup redisclass was cleared.")
                return

            redisClassConf = self.getGroupClassConf("redisclass")
            if (redisClassConf < 0):
                self.redisInstanceForResCtrl.logger.log("Command to get the cgroup redisclass is not reasonable value. failed.")
                raise Exception(ErrorCode.GAUSS_517["GAUSS_51706"] + " Error:get redisclass")

            defaultClassConf = self.getGroupClassConf("DefaultClass")
            if (defaultClassConf < 0):
                self.redisInstanceForResCtrl.logger.log("Command to get the cgroup DefaultClass is not reasonable value. failed.")
                raise Exception(ErrorCode.GAUSS_517["GAUSS_51706"] + " Error:get DefaultClass")

            defaultClassConf = defaultClassConf + redisClassConf

            #Virtual cluster mode
            if not self.isElasticGroup:
                # form ssh command
                sshcmd = "source %s; gs_ssh -c \"%s/bin/gs_cgroup -d -S redisclass\"" % (sourceFile, gaussHome)
            else:
                sshcmd = "source %s; gs_ssh -c \"%s/bin/gs_cgroup -d -S redisclass -N %s\"" % (sourceFile, gaussHome, self.vcGroupName)
            self.redisInstanceForResCtrl.logger.debug("Command to drop subclass group redisclass: %s"
                                                      % sshcmd)

            # execute the ssh command
            (status, output) = commands.getstatusoutput(sshcmd)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Command to drop group redisclass failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sshcmd + " Error:\n%s" % str(output))

            #Virtual cluster mode
            if not self.isElasticGroup:
                # form ssh command
                sshcmd = "source %s; gs_ssh -c \"%s/bin/gs_cgroup -u -S DefaultClass -s %d\"" % (sourceFile, gaussHome, defaultClassConf)
            else:
                sshcmd = "source %s; gs_ssh -c \"%s/bin/gs_cgroup -u -S DefaultClass -s %d -N %s\"" % (sourceFile, gaussHome, defaultClassConf, self.vcGroupName)
            self.redisInstanceForResCtrl.logger.debug("Command to change group DefaultClass : %s"
                                                      % sshcmd)

            # execute the ssh command
            (status, output) = commands.getstatusoutput(sshcmd)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Command to change group DefaultClass failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sshcmd + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def createWorkloadCgroup(self):
        """
        function: create the workload group:
                  1.get MpprcFile
                  2.get gausshome
                  3.form the ssh command
                  4.execute the ssh command
        input : NA
        output: NA
        """

        try:
            # - the max percentage for a workload group in its corresponding subclass group is 99%,
            #   which cannot reach 100%

            #get MpprcFile
            sourceFile = DefaultValue.getMpprcFile()

            #get gausshome
            gaussHome = DefaultValue.getEnv("GAUSSHOME")
            if not gaussHome:
                return

            # form the ssh command
            # Physical cluster mode
            if not self.isElasticGroup:
                sshcmd = "source %s; gs_ssh -c \"%s/bin/gs_cgroup -c -S redisclass -G redisgrp -g 99\"" % (sourceFile, gaussHome)
            else:
                sshcmd = "source %s; gs_ssh -c \"%s/bin/gs_cgroup -c -S redisclass -G redisgrp -g 99 -N %s\"" % (sourceFile, gaussHome, self.vcGroupName)
            self.redisInstanceForResCtrl.logger.debug("Command to create workload group: %s"
                                                      % sshcmd)

            # execute the ssh command
            (status, output) = commands.getstatusoutput(sshcmd)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Command to create workload group redisgrp failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sshcmd + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def dropWorkloadCgroup(self):
        """
        function: drop the workload group:
                  1.get MpprcFile
                  2.get gausshome
                  3.form the ssh command
                  4.execute the ssh command
        input : NA
        output: NA
        """

        try:
            #get MpprcFile
            sourceFile = DefaultValue.getMpprcFile()

            #get gausshome
            gaussHome = DefaultValue.getEnv("GAUSSHOME")
            if not gaussHome:
                return

            # form the ssh command
            # Physical cluster mode
            if not self.isElasticGroup:
                sshcmd = "source %s; gs_ssh -c \"%s/bin/gs_cgroup -d -S redisclass -G redisgrp\"" %(sourceFile, gaussHome)
            else:
                sshcmd = "source %s; gs_ssh -c \"%s/bin/gs_cgroup -d -S redisclass -G redisgrp -N %s\"" %(sourceFile, gaussHome, self.vcGroupName)
            self.redisInstanceForResCtrl.logger.debug("Command to drop workload group: %s"
                                                      % sshcmd)

            # execute the ssh command
            (status, output) = commands.getstatusoutput(sshcmd)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Command to drop workload group redisgrp failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sshcmd + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def alterWorkloadCgroupToLevelF(self):
        """
        function: Alter workload group to CPU resource level "f":
                  1.get MpprcFile
                  2.get gausshome
                  3.form the ssh command
                  4.execute the ssh command
        input : NA
        output: NA
        """

        try:
            #get MpprcFile
            sourceFile = DefaultValue.getMpprcFile()

            #get gausshome
            gaussHome = DefaultValue.getEnv("GAUSSHOME")
            if not gaussHome:
                return

            # Physical cluster mode
            if not self.isElasticGroup:
                sshcmd = "source %s; gs_ssh -c \"%s/bin/gs_cgroup -u -S redisclass -G redisgrp -g 99 --fixed\"" % (sourceFile, gaussHome)
            else:
                sshcmd = "source %s; gs_ssh -c \"%s/bin/gs_cgroup -u -S redisclass -G redisgrp -g 99 -N %s --fixed\"" % (sourceFile, gaussHome, self.vcGroupName)
            self.redisInstanceForResCtrl.logger.debug("Command to alter workload group level: %s"
                                                      % sshcmd)

            (status, output) = commands.getstatusoutput(sshcmd)
        except Exception, e:
            raise Exception(str(e))

    def alterWorkloadCgroupToLevelH(self):
        """
        function: Alter workload group to CPU resource level "h":
                  1.get MpprcFile
                  2.get gausshome
                  3.form the ssh command
                  4.execute the ssh command
        input : NA
        output: NA
        """

        try:
            #get MpprcFile
            sourceFile = DefaultValue.getMpprcFile()

            #get gausshome
            gaussHome = DefaultValue.getEnv("GAUSSHOME")
            if not gaussHome:
                return

            # Physical cluster mode
            if not self.isElasticGroup:
                sshcmd = "source %s; gs_ssh -c \"%s/bin/gs_cgroup -u -S redisclass -G redisgrp -g 50 --fixed\"" % (sourceFile, gaussHome)
            else:
                sshcmd = "source %s; gs_ssh -c \"%s/bin/gs_cgroup -u -S redisclass -G redisgrp -g 50 -N %s --fixed\"" % (sourceFile, gaussHome, self.vcGroupName)
            self.redisInstanceForResCtrl.logger.debug("Command to alter workload group level: %s"
                                                      % sshcmd)

            (status, output) = commands.getstatusoutput(sshcmd)
        except Exception, e:
            raise Exception(str(e))

    def alterWorkloadCgroupToLevelM(self):
        """
        function: Alter workload group to CPU resource level "m":
                  1.get MpprcFile
                  2.get gausshome
                  3.form the ssh command
                  4.execute the ssh command
        input : NA
        output: NA
        """

        try:
            #get MpprcFile
            sourceFile = DefaultValue.getMpprcFile()

            #get gausshome
            gaussHome = DefaultValue.getEnv("GAUSSHOME")
            if not gaussHome:
                return

            # Physical cluster mode
            if not self.isElasticGroup:
                sshcmd = "source %s; gs_ssh -c \"%s/bin/gs_cgroup -u -S redisclass -G redisgrp -g 30 --fixed\"" % (sourceFile, gaussHome)
            else:
                sshcmd = "source %s; gs_ssh -c \"%s/bin/gs_cgroup -u -S redisclass -G redisgrp -g 30 -N %s --fixed\"" % (sourceFile, gaussHome, self.vcGroupName)
            self.redisInstanceForResCtrl.logger.debug("Command to alter workload group level: %s"
                                                      % sshcmd)

            (status, output) = commands.getstatusoutput(sshcmd)
        except Exception, e:
            raise Exception(str(e))

    def alterWorkloadCgroupToLevelL(self):
        """
        function: Alter workload group to CPU resource level "l":
                  1.get MpprcFile
                  2.get gausshome
                  3.form the ssh command
                  4.execute the ssh command
        input : NA
        output: NA
        """

        try:
            #get MpprcFile
            sourceFile = DefaultValue.getMpprcFile()

            #get gausshome
            gaussHome = DefaultValue.getEnv("GAUSSHOME")
            if not gaussHome:
                return

            # Physical cluster mode
            if not self.isElasticGroup:
                sshcmd = "source %s; gs_ssh -c \"%s/bin/gs_cgroup -u -S redisclass -G redisgrp -g 10 --fixed\"" % (sourceFile, gaussHome)
            else:
                sshcmd = "source %s; gs_ssh -c \"%s/bin/gs_cgroup -u -S redisclass -G redisgrp -g 10 -N %s --fixed\"" % (sourceFile, gaussHome, self.vcGroupName)
            self.redisInstanceForResCtrl.logger.debug("Command to alter workload group level: %s"
                                                      % sshcmd)

            (status, output) = commands.getstatusoutput(sshcmd)
        except Exception, e:
            raise Exception(str(e))

    def alterResourcePoolChangeIOLimitToLevelF(self):
        """
        function: alter the resource pool to limit the io to level f:
                  1.form the ssh command
                  2.execute the ssh command
        input : NA
        output: NA
        """

        try:
            sql = "ALTER RESOURCE POOL redisrespool WITH (io_limits = 0);"
            self.redisInstanceForResCtrl.logger.debug("Sql command to alter iops_limit: %s"
                                                      % sql)

            # execute command to alter iops_limit
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to alter iops_limit failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def alterResourcePoolChangeIOLimitToLevelH(self):
        """
        function: alter the resource pool to limit the io to level h:
                  1.form the ssh command
                  2.execute the ssh command
        input : NA
        output: NA
        """

        try:
            sql = "ALTER RESOURCE POOL redisrespool WITH (iops_limit = 50);"
            self.redisInstanceForResCtrl.logger.debug("Sql command to alter iops_limit: %s"
                                                      % sql)

            # execute command to alter iops_limit
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to alter iops_limit failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def alterResourcePoolChangeIOLimitToLevelM(self):
        """
        function: alter the resource pool to limit the io to level m:
                  1.form the ssh command
                  2.execute the ssh command
        input : NA
        output: NA
        """

        try:
            sql = "ALTER RESOURCE POOL redisrespool WITH (iops_limit = 30);"
            self.redisInstanceForResCtrl.logger.debug("Sql command to alter iops_limit: %s"
                                                      % sql)

            # execute command to alter iops_limit
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to alter iops_limit failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def alterResourcePoolChangeIOLimitToLevelL(self):
        """
        function: alter the resource pool to limit the io to level l:
                  1.form the ssh command
                  2.execute the ssh command
        input : NA
        output: NA
        """

        try:
            sql = "ALTER RESOURCE POOL redisrespool WITH (io_limits = 10);"
            self.redisInstanceForResCtrl.logger.debug("Sql command to alter iops_limit: %s"
                                                      % sql)

            # execute command to alter iops_limit
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to alter iops_limit failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def alterResourcePoolChangeIOPriorityToLevelF(self):
        """
        function: alter the resource pool to limit the io_priority to level f:
                  1.form the ssh command
                  2.execute the ssh command
        input : NA
        output: NA
        """

        try:
            sql = "ALTER RESOURCE POOL redisrespool WITH (io_priority = 'None');"
            self.redisInstanceForResCtrl.logger.debug("Sql command to alter io_priority: %s"
                                                      % sql)

            # execute command to alter iops_limit
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to alter io_priority failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def alterResourcePoolChangeIOPriorityToLevelH(self):
        """
        function: alter the resource pool to limit the io_priority to level h:
                  1.form the ssh command
                  2.execute the ssh command
        input : NA
        output: NA
        """

        try:
            sql = "ALTER RESOURCE POOL redisrespool WITH (io_priority = 'High');"
            self.redisInstanceForResCtrl.logger.debug("Sql command to alter io_priority: %s"
                                                      % sql)

            # execute command to alter iops_limit
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to alter io_priority failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def alterResourcePoolChangeIOPriorityToLevelM(self):
        """
        function: alter the resource pool to limit the io_priority to level m:
                  1.form the ssh command
                  2.execute the ssh command
        input : NA
        output: NA
        """

        try:
            sql = "ALTER RESOURCE POOL redisrespool WITH (io_priority = 'Medium');"
            self.redisInstanceForResCtrl.logger.debug("Sql command to alter io_priority: %s"
                                                      % sql)

            # execute command to alter iops_limit
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to alter io_priority failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def alterResourcePoolChangeIOPriorityToLevelL(self):
        """
        function: alter the resource pool to limit the io_priority to level l:
                  1.form the ssh command
                  2.execute the ssh command
        input : NA
        output: NA
        """

        try:
            sql = "ALTER RESOURCE POOL redisrespool WITH (io_priority = 'Low');"
            self.redisInstanceForResCtrl.logger.debug("Sql command to alter io_priority: %s"
                                                      % sql)

            # execute command to alter iops_limit
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to alter io_priority failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))

        except Exception, e:
            raise Exception(str(e))

    def getResourcePoolLevel(self):
        """
        function: get the resource pool level of all setting:respool_name, control_group, io_limits, io_priority
                  1.form the ssh command
                  2.execute the ssh command
        input : NA
        output: NA
        """

        try:
            sql = "SELECT respool_name, control_group, io_limits, io_priority FROM PG_RESOURCE_POOL;"
            self.redisInstanceForResCtrl.logger.debug("Sql command to get resource pool setting level: %s"
                                                      % sql)

            # execute command to get the resource pool level of all setting
            (status, output) = ClusterCommand.remoteSQLCommand(sql,
                                                               self.defaultuser,
                                                               self.redisCN.hostname,
                                                               self.redisCN.port,
                                                               ignoreError = False)

            if (status != 0):
                self.redisInstanceForResCtrl.logger.log("Sql command to get resource pool setting level failed.")
                raise Exception(ErrorCode.GAUSS_514["GAUSS_51400"] % sql + " Error:\n%s" % str(output))
            else:
                self.redisInstanceForResCtrl.logger.debug("Redis resource pool level setting:\n%s\n" % str(output))

        except Exception, e:
            raise Exception(str(e))

